/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit;
import hiit.blackboard.BlackboardFactory;
import hiit.blackboard.Blackboard;
import hiit.blackboard.httpd.HttpServerFactory;
import hiit.connectorUnit.BTDeviceConnector;
import hiit.datagatherer.DataGatherer;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.microedition.midlet.MIDlet;
import javax.microedition.midlet.MIDletStateChangeException;
import javax.microedition.rms.RecordEnumeration;
import javax.microedition.rms.RecordStore;
import javax.microedition.rms.RecordStoreException;
import javax.microedition.rms.RecordStoreFullException;
import javax.microedition.rms.RecordStoreNotFoundException;
import javax.microedition.lcdui.*;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import hiit.reader.*;
import hiit.util.RemoteDeviceSorter;
import hiit.util.FileUtil;
import hiit.util.Settings;
import hiit.logger.*;
import hiit.mapper.BTMapper;
import hiit.mapper.MIDPRecordComparator;
import hiit.mapper.MIDPRecorder;
import hiit.mapper.Mapper;
import hiit.mapper.Properties;
import hiit.mapper.Recorder;
import hiit.midpgui.*;
import hiit.observer.ConnectorObserver;
import hiit.plugin.MIDPPlugin;

//import hiit.plugin.location.LocationPlugin;
import hiit.plugin.transmitter.BoardTransmitter;
import hiit.plugin.transmitter.LocalTransmitter;

/**
 * The Betelgeuse class extends MIDlet and thus it can be run on a mobile device
 * with MIDP 2.0. This class combines the GUI and core classes together.
 * @author Joonas Kukkonen and Eemil Lagerspetz
 *
 */

public class BetelgeuseMIDP extends MIDlet implements CommandListener, ConnectorObserver, Runnable{
	
	private Display display;
	private Stack currentStack = new Stack();
	
	//The main displayable and its commands
	private TabCanvas canvas;	
	private Command exitCommand;
	private Command mapperCommand;
	private Command devicesCommand;
	private Command readerTimeoutCommand;
	private Command optionsCommand;
	private Command showLog;
	private Command discoveryCommand;
	private Command aboutCommand;
	
	//The mapper rule screen and its commands
	private Mapper mapper = null;
	
	private Form aboutScreen = null;
	private boolean online = false; // using online transmitter?
	
	
	private List mapperScreen;
	private Command openNewRuleScreen;
	private Command deleteRule;
	private Hashtable rules;
	private Vector keys;
	
	private Form newRuleScreen;
	private Command addRule;
	private TextField friendlyName;
	private TextField btAddress;
	private ChoiceGroup deviceId;
	private Hashtable idName;
	private String[] nameIdArray;
	private ChoiceGroup friendlyOrAddress;
	
	//The reader timeout option screen and its commands
	private List timeoutScreen;
	private Vector readers;
	private Command readerSelect;
	private Form readerScreen;
	private TextField readDelay;
	private TextField startValue;
	private TextField stopValue;
	private ChoiceGroup giveUp;
	private Command saveReaderTimeouts;
	private Reader currentReader;
	
	//The general timeout option form and its commands
	private Form optionsScreen;
	private Command saveOptions;
	private ChoiceGroup discoveryOnOff;
	private TextField deviceInquiry;
	private ChoiceGroup ldrType;
	private int currentLdr = 0;
	
	//The found devices screen and its commands
	private Vector foundDevices = null;
	private Vector connectedDevices;
	private List deviceScreen;
	private Command deviceSelect;
	private Form deviceInfoScreen;
	private StringItem deviceFriendlyName;
	private StringItem deviceBTAddress;
	private StringItem deviceConnectionStatus;
	private Command deviceInfoAddRule;
	
	//Plugins
	private List pluginList;
	private Command pluginSelect;
	private Vector plugins;
	private Command showPluginScreen;
	
	//"Random" variables
	private final String titleString = "BeTelGeuse";	
	//private boolean isRunning = true;
	public static MIDPLogger logger;
	
	private DataGatherer dgat = null;
	
	//The readers that the GUI already displays
	private Vector currentReaders = new Vector();
	
	//Debug
	private BoardTransmitter transmitter = null;
	private LocalTransmitter localTransmitter = null;
	
	private Vector ldevs = new Vector();
	private Hashtable reader_io = new Hashtable();
	
	private boolean discoveryMode = true;
	private TextField frequency = null;
	
	//User information form
	private Form userInfo = null;
	private TextField userNameField = null;
	private TextField passwordField = null;
	private ChoiceGroup rememberUserInfo = null;
	private Command sendUserInfo = null;
	private Command userExit = null;
	
	private String userName = "";
	private String password = "";
	private boolean remember = false;
	
	// loading-screen
	private Gauge loadingBar = null;
	private Timer loadingTimer = null;
	private int loadingPerc = 0;
	private boolean initialized = false;
	
	private boolean isRunning = true;
	private GPSReader gps = null;
	
	private static int gpsRequestInterval = 60000;
	private static long lastRequestTime = -1;
	
	// 
	// Properties used in testing/event processing
	// 
	// Used in BoardTransmitter, measures throughput at each send
	public static boolean measureKbs = false;       
	
	// used in BordTransimitter, forces the socket-connection to stay open by sending pings when neccessary
	public static boolean pingSocketToServer = true;

	// when enabled, BoardTransmitter adjusts GPS's sample-interval according to the time elapsed since last 'distanceChange' -event
	public static boolean adjustGPSRequestInterval = false;
	
	// when enabled, GSMEvent emits 'BTDeviceChange' events.
	// disabled by default, since requires string comparison AND that BTDeviceListener sorts the list of discovered devices,
	// using RemoteDeviceSorter.quicksort
	public static boolean emitBTDevicesChangeEvents = true;
	
	private Blackboard b;
	
	
	public BetelgeuseMIDP() {
		super();
		this.display = Display.getDisplay(this);
		askUserInformation();
	}
	
	   public void run() {
	        while(isRunning){
	            if(this.gps == null){
	                Vector temp = this.dgat.getConnector().getConnectedDevices();
	                if(temp != null){
	                    for(int i=0; i<temp.size(); i++){
	                        Reader r = (Reader)temp.elementAt(i);
	                        if(r.getDeviceId() == Properties.GPS_READER){
	                        	try {
		                            this.gps = (GPSReader)r;
		                            this.gps.readOne();
	                        	} catch (Exception e){
	                        		 if (logger != null) logger.logMessage("Bet readGPS ERR: " + e);
	                        	}
	                        }
	                        /*
	                        if(r.getDeviceId() == Properties.GSM_READER){
	                            isRunning = false;
	                            return;
	                        }*/
	                    }
	                }
	            } else {
	            	try {
	            		BetelgeuseMIDP.logger.logMessage("Betelgeuse: read GPS, interval: " + gpsRequestInterval);
	            		lastRequestTime = System.currentTimeMillis();
	            		this.gps.readOne();
	            	} catch (Exception e){
               		 	if (logger != null) logger.logMessage("Bet readGPS ERR: " + e);
	            	}
	            }
	            synchronized(this){
	                try {
	                    this.wait(gpsRequestInterval);  // GPS read-wait, default (1 min): 60000
	                } catch (InterruptedException e) {
	                	
	                }
	            }
	           
	        }
		 
	   }

	   /*
	    * Initializes the plugins. Add the plugins you want to this method.
	    */

	   protected void initializePlugins(){

		   String transmitterType = Settings.getValue(Settings.TRANSMITTER);

		   if(transmitterType.equals("local")){


			   try{
				   online = false; // for about screen
				   localTransmitter = new LocalTransmitter("betelgeuse.dat", this, this.dgat, logger, false); //getPropertyAsBoolean("LocalTransmitterSyncTime"));
				   Thread t = new Thread(localTransmitter);
				   t.start();
			   }catch (Throwable t){
				   System.out.println("Error starting LocalTransmitter... " + t.getMessage());
				   if (this.logger != null) this.logger.logMessage("[GUI] " +"DataTransmitter crashed: " + t.toString());
			   }

		   }else if(transmitterType.equals("online")){
			   try{
				   online = true; // for about screen
				   transmitter = new BoardTransmitter(this, this.dgat, Settings.getValue(Settings.TRANSMITTER_ADDRESS), logger, false);
				   transmitter.switchOn();
				   Thread t = new Thread(transmitter);
				   t.start();
			   }catch (Throwable t){
				   System.out.println("[GUI] " +"DataTransmitter crashed: " + t.toString());
				   if (this.logger != null) this.logger.logMessage("[GUI] " +"DataTransmitter crashed: " + t.toString());
			   }
		   }


		   if(!((Boolean)(Properties.getIdMode().get(new Integer(Properties.GPS_READER)))).booleanValue()){
			   Thread t = new Thread(this);
			   t.start();
		   }

	}
	
	private void askUserInformation(){
		RecordStore rs = null;
		try {
			rs = RecordStore.openRecordStore("btg_user", true);
			if(rs.getNumRecords() > 0){
				RecordEnumeration recEn = rs.enumerateRecords(null, new MIDPRecordComparator(), false);
				String record = new String(recEn.nextRecord());
				int index = record.indexOf(';');
				if(index > 0)
					this.userName = record.substring(0, index);
				if(index + 1 < record.length())
					this.password = record.substring(index+1);
				this.remember = true;
			}
			rs.closeRecordStore();
		} catch (RecordStoreFullException e) {
		} catch (RecordStoreNotFoundException e) {
		} catch (RecordStoreException e) {
		}
		
		this.userInfo = new Form("Login");
		this.userNameField = new TextField("Username", this.userName, 16, TextField.ANY);
		this.passwordField = new TextField("Password", this.password, 16, TextField.PASSWORD);
		this.rememberUserInfo = new ChoiceGroup("", Choice.MULTIPLE);
		this.rememberUserInfo.append("Remember me", null);
		
		if(remember)
			this.rememberUserInfo.setSelectedFlags(new boolean[]{true});
		this.sendUserInfo = new Command("Ok", Command.ITEM, 1);
		this.userExit = new Command("Exit", Command.EXIT, 1);
		
		this.userInfo.append(userNameField);
		this.userInfo.append(passwordField);
		this.userInfo.append(rememberUserInfo);
		this.userInfo.addCommand(sendUserInfo);
		this.userInfo.addCommand(userExit);
		this.userInfo.setCommandListener(this);
		
		this.display.setCurrent(this.userInfo);
		
	} 
	
	private void initializeBTG(){
		long startTime = System.currentTimeMillis();
		try {
			//Logger for logging errors and other messages
			//this.logger = new FileLogger();
			this.logger = new MIDPLogger();
			// this.logger = new MIDPLiveLogger(30);
			
			FileUtil fileReader = new FileUtil(logger);
			Settings s = new Settings(logger);
			s.readSettings("betelgeuse_settings.xml");
			
			
			

			//Mapper which maps BT addresses and friendly names to ID numbers
			this.mapper = new BTMapper(new MIDPRecorder(this.logger));
			//Note: discovery is on by default!
			String savedDiscoveryMode = Settings.getValue(Settings.DISCOVERY_MODE);
			this.discoveryMode = savedDiscoveryMode != null? 
					Integer.parseInt(savedDiscoveryMode) == 1 : true; 

			
			updateLoadingBar();
			dgat = new DataGatherer(this.mapper, discoveryMode, logger);
			
			
			String savedWaitTime = Settings.getValue(Settings.DEVICE_INQUIRY);
			if (savedWaitTime != null)
				dgat.getConnector().setDiscoveryWaitTime(Integer.parseInt(savedWaitTime));
			
			
			//Blackboard b = BlackboardFactory.createBlackboard(this.logger, dgat.getConnector());
			b = BlackboardFactory.createBlackboard(this.logger, dgat.getConnector());
			
			updateLoadingBar();

			//Create a GSM reader...
			String savedLocalDeviceReaderType = Settings.getValue(Settings.LOCAL_READER);
			this.currentLdr = savedLocalDeviceReaderType != null? Integer.parseInt(savedLocalDeviceReaderType) : 1;
			if (!createLocalDeviceReader(this.currentLdr, true)){
				this.currentLdr = 0;
				createLocalDeviceReader(this.currentLdr, true);
			}

			/*
			this.currentLdr = 0;
			if (this.dgat.getConnector().createSpecialReader(null, Properties.LOCAL_GSM_READER, true)) {
			//if(this.dgat.getConnector().createSpecialReader("socket://127.0.0.1:4040", Properties.GSM_READER, true))
			//this.dgat.getConnector().createSpecialReader(null, Properties.LOCAL_READER, true);
			//Set the type of the LDR used...
				this.currentLdr = 1; // This is WRONG. But I guess it does not matter...
            } else {
				this.dgat.getConnector().createSpecialReader("", Properties.LOCAL_READER, true);
				this.currentLdr = 0;
			}
			System.out.println("LOCAL reader " + this.currentLdr);
*/
			
			
			//Main screen. Header, reader buttons, and the area where
			//reader information is shown
			this.canvas = new TabCanvas(this.titleString);
			String savedRedrawSetting = Settings.getValue(Settings.GUI_REDRAW);
			int defaultRedraw = 500;
			Settings.setValue(Settings.GUI_REDRAW, new Integer(defaultRedraw));
			canvas.setRedrawFrequency(savedRedrawSetting != null? Integer.parseInt(savedRedrawSetting) : defaultRedraw );
			
			//Initialize the gui
			initializeGUI();
			
			// and show the GUI.
			long endTime = System.currentTimeMillis();
			long timeTaken = (endTime - startTime)/1000;
			if (logger != null) logger.logMessage("[GUI] " + "Starting BeTelGeuse took "+ timeTaken + " s."); 
		
			updateLoadingBar();

			//Start the DataGatherer, sensor discovery starts
			dgat.start();
			
			updateLoadingBar();
			
			BTDeviceConnector c = dgat.getConnector();
			
			updateLoadingBar();
			
			c.subscribe(this);
			
			updateLoadingBar();
			
			//If you want to start any plugins, put them in initializePlugins() method.
			initializePlugins();
			
			// Subscribe to the Connector... and tell it that everything is ready
			synchronized(c){
				c.notifyAll();
			}
			
			
			updateLoadingBar();
			
			initialized = true;
		} catch(Exception ex) {
			if (logger != null) logger.logMessage("[GUI] " +"Error in Betelgeuse constructor:" + ex.toString());
		}
	}
	
	private void initializeGUI() {
		
		this.exitCommand = new Command("Exit", Command.EXIT, 4);
		this.mapperCommand = new Command("Device Mappings", Command.ITEM, 1);
		this.devicesCommand = new Command("Discovered devices", Command.ITEM, 1);
		this.readerTimeoutCommand = new Command("Reader timeouts", Command.ITEM, 1);
		this.optionsCommand = new Command("Options", Command.ITEM, 1);
		this.showPluginScreen = new Command("Plugins", Command.ITEM, 1);
		this.showLog = new Command("Show log", Command.ITEM, 2);
		this.discoveryCommand = new Command("Search devices", Command.ITEM,1);
		this.aboutCommand = new Command("About", Command.ITEM,1);
		
		this.canvas.addCommand(this.mapperCommand);
		this.canvas.addCommand(this.devicesCommand);
		this.canvas.addCommand(this.readerTimeoutCommand);
		this.canvas.addCommand(this.optionsCommand);
		this.canvas.addCommand(this.showPluginScreen);
		this.canvas.addCommand(this.showLog);	
		this.canvas.addCommand(this.exitCommand);
		this.canvas.addCommand(this.discoveryCommand);
		this.canvas.addCommand(this.aboutCommand);
		this.canvas.setCommandListener(this);
		
		createReaderTimeoutScreen();
		createOptionsScreen();
		createDeviceInfoScreen();
		createMapperScreen();
		createPluginScreen();
		createAboutScreen();
		//Show the main screen at the start	
		//display.setCurrent(this.canvas);
		
		
		/*
		//Start the reader checking
		Thread t = new Thread(this);
		t.start();
		*/
	}

	protected void startApp() throws MIDletStateChangeException {
		synchronized(this) {
			this.notifyAll();
		}
	}

	protected void pauseApp() {		
		/*synchronized(this) {
			try {
				this.wait();
			}catch(Exception ex) {
				if (logger != null) logger.logMessage("[GUI] " +ex.toString());
			}
		}*/
	}

	protected synchronized void destroyApp(boolean arg0) throws MIDletStateChangeException {
		if (initialized == true) {
			for(int i = 0; i < this.pluginList.size(); i++){
				((MIDPPlugin)plugins.elementAt(i)).close();
			}
			if (this.canvas != null)
				this.canvas.close();
			if (this.dgat != null)
				this.dgat.closeAll();

			BlackboardFactory.getBlackboard(logger).kill();
			if(this.transmitter != null)
				this.transmitter.close();
			if(this.localTransmitter != null)
				this.localTransmitter.close();
		}
			isRunning = false;
		
		/*if(this.locationPlugin != null)
			this.locationPlugin.close();*/
		//this.transmitter.close();
	}
	
	/*
	 * Creates the mapper rule screen and a skeleton screen for adding
	 * new rules.
	 */
	
	private void createMapperScreen() {
		
		//Create the rule screen
		
		this.mapperScreen = new List("Device Mappings", Choice.IMPLICIT);
		this.mapperScreen.addCommand(new Command("Back", Command.BACK, 1));
		
		this.openNewRuleScreen = new Command("Add new", Command.ITEM, 2);
		this.deleteRule = new Command("Delete", Command.ITEM, 2);
		
		this.mapperScreen.addCommand(this.openNewRuleScreen);
		this.mapperScreen.addCommand(this.deleteRule);
		this.mapperScreen.setCommandListener(this);
		
		//Create the screen for adding a new rule
		
		this.newRuleScreen = new Form("New device mapping");
		this.addRule = new Command("Add", Command.SCREEN, 2);
		this.newRuleScreen.addCommand(new Command("Back", Command.BACK, 1));
		this.newRuleScreen.addCommand(this.addRule);
		this.newRuleScreen.setCommandListener(this);
	
		this.friendlyName = new TextField("", "", 16, TextField.ANY);
		this.friendlyName.setLayout(Item.LAYOUT_NEWLINE_BEFORE);
		this.btAddress = new TextField("", "", 12, TextField.ANY);
		this.btAddress.setLayout(Item.LAYOUT_NEWLINE_BEFORE);
		
		this.idName = Properties.getNameId();
		
		Enumeration enumer = this.idName.keys();
		Vector names = new Vector();
		while(enumer.hasMoreElements())
			names.addElement(enumer.nextElement());
		
		this.nameIdArray = new String[names.size()];
		for (int i = 0; i < names.size(); i++)
			nameIdArray[i] = (String)names.elementAt(i);
		
		this.deviceId = new ChoiceGroup("", Choice.POPUP, nameIdArray, null);
		this.deviceId.setLayout(Item.LAYOUT_NEWLINE_BEFORE);
	
		this.friendlyOrAddress = new ChoiceGroup("", Choice.EXCLUSIVE);
		this.friendlyOrAddress.append("Use partial friendly name", null);
		this.friendlyOrAddress.append("Use BT address", null);
		this.friendlyOrAddress.setSelectedIndex(1, true);

		this.newRuleScreen.append("Parser");
		this.newRuleScreen.append(this.deviceId);
		this.newRuleScreen.append("Partial friendly name");
		this.newRuleScreen.append(this.friendlyName);
		this.newRuleScreen.append("BT address");
		this.newRuleScreen.append(this.btAddress);
		this.newRuleScreen.append("Which to use?");
		this.newRuleScreen.append(this.friendlyOrAddress);
	}

	/*
	 * Generates a skeleton screen for showing all the discovered devices
	 * and a skeleton screen for showing information of one device.
	 */
	
	private void createDeviceInfoScreen() {
		this.deviceScreen = new List("Found Devices", Choice.IMPLICIT);		

		this.deviceSelect  = new Command("Select", Command.ITEM, 2);	
		
		this.deviceScreen.setSelectCommand(this.deviceSelect);
		this.deviceScreen.addCommand(new Command("Back", Command.BACK, 1));
		this.deviceScreen.setCommandListener(this);
		
		// and then the screen that shows a device information form 
		this.deviceInfoScreen = new Form("Device Info");
		
		this.deviceFriendlyName = new StringItem("Friendly Name", "");
		this.deviceInfoScreen.append(this.deviceFriendlyName);
		
		this.deviceBTAddress = new StringItem("\nBluetooth Address", "");
		this.deviceInfoScreen.append(this.deviceBTAddress);
		this.deviceConnectionStatus = new StringItem("\nConnection Status", "");
		this.deviceInfoScreen.append(this.deviceConnectionStatus);
		
		this.deviceInfoAddRule = new Command("Add rule", Command.ITEM, 2);
		this.deviceInfoScreen.addCommand(new Command("Back", Command.BACK, 1));	
		this.deviceInfoScreen.addCommand(this.deviceInfoAddRule);
		this.deviceInfoScreen.setCommandListener(this);
	}

	/*
	 * Generates a screen for adjusting the reader timeouts.
	 */
	
	private void createReaderTimeoutScreen() {
		
		//List of readers
		
		this.timeoutScreen = new List("Parsers", Choice.IMPLICIT);
		this.readerSelect = new Command("Select", Command.ITEM, 2);
		this.timeoutScreen.setSelectCommand(this.readerSelect);
		this.timeoutScreen.addCommand(new Command("Back", Command.BACK, 1));
		this.timeoutScreen.setCommandListener(this);
		
		//Single reader screen
		
		this.readerScreen = new Form("Timeouts");
		this.readDelay = new TextField("Read interval (ms):", "", 9, TextField.NUMERIC);
		this.startValue = new TextField("Reconnect timeout (ms):", "", 9, TextField.NUMERIC);
		this.stopValue = new TextField("Max reconnect timeout (ms):", "", 9, TextField.NUMERIC);
		this.giveUp = new ChoiceGroup("Stop trying to reconnect:", Choice.MULTIPLE);
		
		this.giveUp.append("Give up", null);
		
		this.readerScreen.append(this.readDelay);
		this.readerScreen.append(this.startValue);
		this.readerScreen.append(this.stopValue);
		this.readerScreen.append(this.giveUp);
		
		this.saveReaderTimeouts = new Command("Save", Command.SCREEN, 2);
		this.readerScreen.addCommand(this.saveReaderTimeouts);
		this.readerScreen.addCommand(new Command("Back", Command.BACK, 1));
		this.readerScreen.setCommandListener(this);
	}
	
	/*
	 * Generates a screen for adjusting general timeouts.
	 */
	
	private void createOptionsScreen() {
		this.optionsScreen = new Form("Options");
		
		this.optionsScreen.append("Discovery mode:");
		
		this.discoveryOnOff = new ChoiceGroup("", Choice.EXCLUSIVE);
		this.discoveryOnOff.append("Automatic", null);
		this.discoveryOnOff.append("Request", null);
		if(this.discoveryMode)
			this.discoveryOnOff.setSelectedIndex(0, true);
		else
			this.discoveryOnOff.setSelectedIndex(1, true);
		
		this.optionsScreen.append(this.discoveryOnOff);
		
		this.deviceInquiry = new TextField("\nDevice inquiry:", "", 9, TextField.NUMERIC);
		
		this.optionsScreen.append(this.deviceInquiry);
		
		this.frequency  = new TextField("GUI redraw frequency (ms):", "" + this.canvas.getRedrawFrequency(), 4, TextField.NUMERIC);
		this.optionsScreen.append(frequency);
		
		this.optionsScreen.append("\nLocal device reader type:");
		this.ldrType = new ChoiceGroup("", Choice.EXCLUSIVE);
		this.ldrType.append("Local device reader", null);
		this.ldrType.append("GSM reader", null);
		this.ldrType.append("JNI reader", null);
		this.ldrType.setSelectedIndex(this.currentLdr,true);
		
		this.optionsScreen.append(this.ldrType);
				
		this.saveOptions = new Command("Save", Command.SCREEN, 2);
		this.optionsScreen.addCommand(new Command("Back", Command.BACK, 1));
		this.optionsScreen.addCommand(this.saveOptions);
		this.optionsScreen.setCommandListener(this);
	}
	
	/*
	 * Creates a screen for plugins. Note that this screen is not accessible
	 * if no plugins have been added to the GUI.
	 */
	
	private void createPluginScreen(){
		this.pluginList = new List("Plugins", Choice.IMPLICIT);
		this.pluginSelect = new Command("Select", Command.ITEM, 2);
		this.pluginList.setSelectCommand(this.pluginSelect);
		this.pluginList.addCommand(new Command("Back", Command.BACK, 1));
		this.pluginList.setCommandListener(this);
		
		this.plugins = new Vector();
	}
	
	private void createAboutScreen(){
	    aboutScreen = new Form("About Betelgeuse");
        StringItem s = new StringItem("Version", Properties.major + "."+ Properties.minor + " " + Properties.flavor);
        aboutScreen.append(s);
        aboutScreen.append(new StringItem("Transmitter type", (online ? "Online" : "Offline")));
        aboutScreen.addCommand(new Command("Back", Command.BACK, 1));
        aboutScreen.setCommandListener(this);
    }

	public void commandAction(Command arg0, Displayable arg1) {
		if (arg1 == this.canvas){
			if (arg0 == this.exitCommand){
				try {
					destroyApp(true);
					notifyDestroyed();
				} catch (MIDletStateChangeException e) {
				}				
			}else if (arg0 == this.devicesCommand){
				this.currentStack.push(this.canvas);
				updateDeviceScreen();
				this.display.setCurrent(this.deviceScreen);
			}else if (arg0 == this.readerTimeoutCommand){
				this.currentStack.push(this.canvas);
				updateReaderTimeoutScreen();
				this.display.setCurrent(this.timeoutScreen);
			}else if (arg0 == this.mapperCommand){
				this.currentStack.push(this.canvas);
				updateMapperScreen();
				this.display.setCurrent(this.mapperScreen);
			}else if (arg0 == this.aboutCommand){
			    this.currentStack.push(this.canvas);
			    this.display.setCurrent(this.aboutScreen);   
			}else if (arg0 == this.optionsCommand){
				this.currentStack.push(this.canvas);
				this.deviceInquiry.setString("" + this.dgat.getConnector().getDiscoveryWaitTime());
				this.display.setCurrent(this.optionsScreen);	
			}else if (arg0 == this.showLog){
				Form temp = this.logger.getLogScreen();
				temp.addCommand(new Command("Back", Command.BACK, 1));
				temp.setCommandListener(this);
				this.currentStack.push(this.canvas);
				this.display.setCurrent(temp);	
			} else if (arg0 == this.showPluginScreen){
				this.currentStack.push(this.canvas);
				this.display.setCurrent(this.pluginList);				
			} else if (arg0 == this.discoveryCommand){
				if(this.discoveryMode==false){
					synchronized(this.dgat.getConnector()){
						this.dgat.getConnector().notify();
					}
				}
			}
		}else if (arg1 == deviceScreen){
			if (arg0 == deviceSelect){
				if(updateDeviceInfoScreen()){
					this.currentStack.push(this.deviceScreen);
					this.display.setCurrent(this.deviceInfoScreen);
				}
			}
		}else if (arg1 == this.deviceInfoScreen){	
			if (arg0 == this.deviceInfoAddRule){
				this.currentStack.push(this.deviceInfoScreen);
				this.display.setCurrent(this.newRuleScreen);
				
			}
		}else if (arg1 == this.optionsScreen){
			if(arg0 == this.saveOptions){
				saveOptions();
				this.display.setCurrent((Displayable)this.currentStack.pop());
			}
			
		}else if (arg1 == this.timeoutScreen){
			if (arg0 == readerSelect){
				if(updateTimeoutScreenForReader()){
					this.currentStack.push(this.timeoutScreen);
					this.display.setCurrent(this.readerScreen);
				}			
			}
		}else if (arg1 == this.readerScreen){
			if (arg0 == this.saveReaderTimeouts){
				saveReaderTimeouts();
				this.display.setCurrent((Displayable)this.currentStack.pop());				
			}
		}else if (arg1 == this.mapperScreen){
			if(arg0 == this.openNewRuleScreen){
				this.currentStack.push(this.mapperScreen);
				this.friendlyName.setString("");
				this.btAddress.setString("");	
				this.display.setCurrent(this.newRuleScreen);
			}else if (arg0 == this.deleteRule){
				int i = mapperScreen.getSelectedIndex();
				
				String key = (String)keys.elementAt(i);
				int id = ((Integer)rules.get(key)).intValue();
				
				if(!this.mapper.deleteRecord(id, key)){
					Alert alert = new Alert("Syntax error");
		            alert.setTimeout(5000);
		            alert.setType(AlertType.ERROR);
		            alert.setString("Deleting the rule did not succeed. Please check the log for more details.");
		            display.setCurrent(alert, this.display.getCurrent());
		            if(this.logger != null) this.logger.logMessage("[GUI] " +"Deleting a rule did not succeed.");
				}else
					if (this.logger != null) this.logger.logMessage("[GUI] " +"Rule deleted: " + id + " - " + key);
				updateMapperScreen();
			}
		}else if (arg1 == this.newRuleScreen){
			if(arg0 == this.addRule){
				if(addRule()){
					updateMapperScreen();
					this.display.setCurrent((Displayable)this.currentStack.pop());
				}
			}
		}else if (arg1 == this.pluginList){
			if (arg0 == this.pluginSelect){
				int i = this.pluginList.getSelectedIndex();
				if (i > -1 && i < this.plugins.size())
					((MIDPPlugin)this.plugins.elementAt(i)).showPluginGui(this.display, this.display.getCurrent());
			}
		}else if(arg1 == this.userInfo){
			if(arg0 == this.userExit){
				try {
					destroyApp(true);
					notifyDestroyed();
				} catch (MIDletStateChangeException e) {
				}
			}else if (arg0 == this.sendUserInfo){
				if(this.userNameField.getString().indexOf(';') != -1){
					Alert alert = new Alert("Invalid username.");
		            alert.setTimeout(5000);
		            alert.setType(AlertType.ERROR);
		            alert.setString("Username may not contain character ';'");
		            display.setCurrent(alert, this.display.getCurrent());
				}else{
					this.showLoadingBar();
				}
			}
		}
		if (arg0.getCommandType() == Command.BACK){
			this.display.setCurrent((Displayable)this.currentStack.pop());
		}
		
	}
	
	/**
	 * Allows showing an Alert on top of the BeTelGeuse GUI.
	 * @param a The Alert to be shown
	 */

	public void showAlert(Alert a){
		if (this.currentStack.contains(this.display.getCurrent())){
			// Ok we don't stack alerts, so we aren't showing an alert now. safe to show one.
			this.display.setCurrent(a, this.display.getCurrent());
			//NOTE: An exception will be thrown if the current displayable shown is already an alert.
		}
	}
	
	
	/*
	 * Updates the screen showing all discovered devices.
	 */
	
	private boolean updateDeviceInfoScreen() {
		int i = deviceScreen.getSelectedIndex();
		
		if(foundDevices == null)
			return false;
		if (i < this.foundDevices.size()){
		RemoteDevice r = (RemoteDevice)this.foundDevices.elementAt(i);
		if (r == null)
			return false;
		try {
			String friendlyName = r.getFriendlyName(false);
			deviceFriendlyName.setText(friendlyName);
			this.friendlyName.setString(friendlyName);
		} catch (IOException e) {
			// Fails if fails...
			deviceFriendlyName.setText("Not available");
		}
		String address = r.getBluetoothAddress();
		deviceBTAddress.setText("\n" + address);
		this.btAddress.setString(address);

		
		this.connectedDevices = this.dgat.getConnector().getConnectedRemoteDevices();
		if (connectedDevices.contains(r))
			deviceConnectionStatus.setText("\nConnected");
		else
			deviceConnectionStatus.setText("\nNot connected");
		
		
		return true;
		}
		// a local device, or some skew.
		i -= foundDevices.size();
		if (i < ldevs.size()){
			Reader r = (Reader)ldevs.elementAt(i);
			deviceFriendlyName.setText(r.getDeviceName());
			this.friendlyName.setString("");
			deviceBTAddress.setText("\n" + r.getBluetoothAddress());
			btAddress.setString("");
			deviceConnectionStatus.setText("\nConnected");
			return true;
		}
		return false;
	}
	
	/*
	 * Updates the device info screen.
	 */

	private void updateDeviceScreen() {
		this.deviceScreen.deleteAll();
		
		this.foundDevices = new Vector();
		//These are remote devices
		Vector allRemoteDevices = dgat.getConnector().getLastFound();
		Vector temp = dgat.getConnector().getConnectedRemoteDevices();
		
		for(int i=0; i<temp.size();i++){
			if(!allRemoteDevices.contains(temp.elementAt(i)))
				allRemoteDevices.addElement(temp.elementAt(i));
		}
		//These are Readers
		Vector allReaders = dgat.getConnector().getConnectedDevices();
		Reader r;	
		RemoteDevice rd;
		int size = allRemoteDevices.size();
		
		//Loop readers
		outer: for (int i = 0; i < allReaders.size(); i++){
			r = (Reader)allReaders.elementAt(i);
			if (r.getBluetoothAddress().equalsIgnoreCase(this.dgat.getBluetoothAddress())){
				if (!ldevs.contains(r)){
					this.ldevs.addElement(r);
				}
				continue outer;
			}
			for (int j = 0; j < size; j++){
				rd = (RemoteDevice)allRemoteDevices.elementAt(j);
				//If the Reader is connected to this device...
				if (r.getBluetoothAddress().equalsIgnoreCase(rd.getBluetoothAddress())){
					//Append the device to the list
					try {
						this.deviceScreen.append("Connected: " + rd.getFriendlyName(false), null);
					} catch (IOException e) {
						this.deviceScreen.append("Connected: " + rd.getBluetoothAddress(), null);
					}
					//Add it to a Vector from where it can be found again
					this.foundDevices.addElement(rd);
					//Delete it from the other one so that we don't run into it again
					allRemoteDevices.removeElementAt(j);
					j--;
					size--;
					continue outer;
					
				}
			}
		}
		//There are nonconnected devices still in the other Vector... at least we hope so
		for (int i = 0; i < allRemoteDevices.size(); ++i){
			rd = (RemoteDevice)allRemoteDevices.elementAt(i);
			this.foundDevices.addElement(rd);
			try {
				this.deviceScreen.append(rd.getFriendlyName(false), null);
			} catch (IOException e) {
				this.deviceScreen.append(rd.getBluetoothAddress(), null);
			}
		}
		for (int i = 0; i < ldevs.size(); ++i)
			this.deviceScreen.append("Connected: " + ((Reader)ldevs.elementAt(i)).getDeviceName(), null);
		}

	/*
	 * Updates the screen showing the readers.
	 */
	
	private void updateReaderTimeoutScreen() {
		this.readers = dgat.getConnector().getConnectedDevices();
		
		if (this.readers == null || this.readers.size() == 0)
			return;
		
		this.timeoutScreen.deleteAll();
		for (int i = 0; i< this.readers.size(); i++)
			this.timeoutScreen.append(((Reader)this.readers.elementAt(i)).getDeviceName() + " - " + ((Reader)this.readers.elementAt(i)).getBluetoothAddress(), null);
	}
	
	/*
	 * Updates reader timeout screen.
	 */
	
	private boolean updateTimeoutScreenForReader(){
		
		int index = this.timeoutScreen.getSelectedIndex();
		if (index < 0 || index > this.readers.size())
			return false;
		
		this.currentReader = (Reader)this.readers.elementAt(index);
		
		this.readDelay.setString("" + this.currentReader.getReadDelay());
		Vector timeouts = this.currentReader.getTimeouts();
		
		this.startValue.setString("" + ((Integer)timeouts.elementAt(0)).intValue());
		this.stopValue.setString("" + ((Integer)timeouts.elementAt(1)).intValue());
		
		if(((Boolean)timeouts.elementAt(2)).booleanValue())
			this.giveUp.setSelectedIndex(0, true);
		else
			this.giveUp.setSelectedIndex(0, false);
			
		
		return true;
	}

	/*
	 * Saves new reader timeouts.
	 */
	
	private void saveReaderTimeouts(){
		char[] c1 = new char[this.readDelay.size()];
		char[] c2 = new char[this.startValue.size()];
		char[] c3 = new char[this.stopValue.size()];
		
		int i, read, start, stop, temp;
		
		i = readDelay.getChars(c1);
		read = 0;
		for(int j = 0; j < i; j++){
			temp = c1[j];
			read = read * 10;
			read = read + temp - 48;
		}
		
		i = startValue.getChars(c2);
		start = 0;
		for(int j = 0; j < i; j++){
			temp = c2[j];
			start = start * 10;
			start = start + temp - 48;
		}
		
		i = stopValue.getChars(c3);
		stop = 0;
		for(int j = 0; j < i; j++){
			temp = c3[j];
			stop = stop * 10;
			stop = stop + temp - 48;
		}
		
		boolean[] give = new boolean[this.giveUp.size()];
		
		
		this.currentReader.setReadDelay(read);
		this.currentReader.setTimeouts(start, stop, give[0]);
	}
	
	/*
	 * Saves general timeouts.
	 */
	
	private void saveOptions(){
		int guiRedraw = 100;
		try{
			guiRedraw = Integer.parseInt(this.frequency.getString());
		} catch (NumberFormatException e){
			// using 100 if this cannot be parsed
		}
		if (guiRedraw < 100){
			guiRedraw = 100;
		} else if (guiRedraw > 10000){
			guiRedraw = 10000;
		}
		
		this.canvas.setRedrawFrequency(guiRedraw);
		//String s = String.valueOf(guiRedraw);
		Settings.setValue(Settings.GUI_REDRAW, new Integer(guiRedraw));
		
		if(this.discoveryMode == true && this.discoveryOnOff.getSelectedIndex() == 1){
			this.dgat.getConnector().discoveryOff();
			this.discoveryMode = false;
		}else if(this.discoveryMode == false && this.discoveryOnOff.getSelectedIndex() == 0){
			this.dgat.getConnector().discoveryOn();
			this.discoveryMode=true;
		}
		// TODO: if discovery-mode is "request" at startup, GUI hangs... so we'll keep it on "automatic" for now
		// Settings.setValue(Settings.DISCOVERY_MODE, new Integer(this.discoveryMode == true? 1 : 0));
		
		
		char[] c2 = new char[this.deviceInquiry.size()];
		
		int i, temp, result = 0;
		
		i = this.deviceInquiry.getChars(c2);
			
		for(int j = 0; j < i; j++){
			temp = c2[j];
			result = result * 10;
			result = result + temp - 48;
		}
		
		this.dgat.getConnector().setDiscoveryWaitTime(result);
		Settings.setValue(Settings.DEVICE_INQUIRY, new Integer(result));
		
		
		int type = 0;
		if(this.currentLdr != (type = this.ldrType.getSelectedIndex())){
			boolean success = createLocalDeviceReader(type, false);
			if (success)
				Settings.setValue(Settings.LOCAL_READER, new Integer(type));
		}

		
		Settings.save();
	}
	
	private boolean createLocalDeviceReader(int type, boolean firstTime){
		boolean success = false;
		if(type == 0){
			success = this.dgat.getConnector().createSpecialReader(null, Properties.LOCAL_READER, true);
		}else if(type == 1){
			success = this.dgat.getConnector().createSpecialReader(null, Properties.LOCAL_GSM_READER, true);
			//success = this.dgat.getConnector().createSpecialReader("socket://127.0.0.1:4040", Properties.GSM_READER, true); 
		}else if (type == 2){
			success = this.dgat.getConnector().createSpecialReader("socket://127.0.0.1:1234", Properties.JNI_READER, true);
		}
		
		if (!firstTime && success) {
			Reader r = null;
			
			if(success){
				if(this.dgat.getConnector().getLDR().size() > 0){
					r = (Reader)this.dgat.getConnector().getLDR().elementAt(0);
					this.dgat.getConnector().getLDR().removeElementAt(0);
				}
				if(r != null)
					r.close();
				this.currentLdr = type;
			}else{
				this.ldrType.setSelectedIndex(currentLdr, true);
				this.dgat.getConnector().getLDR().removeElementAt(1);
			}
		}
		return success;
	}
	/*
	 * Updates the mapper rule screen.
	 */
	
	private void updateMapperScreen() {
		this.rules = this.mapper.getRecorder().retrieve(0);
		
		this.mapperScreen.deleteAll();
		this.keys = new Vector();
		String temp1, temp2;
		
		if (this.rules != null && this.rules.size() > 0){
			
			Enumeration sensorId = rules.elements();
			Enumeration names = rules.keys();
			while (names.hasMoreElements() && sensorId.hasMoreElements()){
				
				temp1 = Integer.toHexString(((Integer)sensorId.nextElement()).intValue());
				temp2 = (String)names.nextElement();
				this.keys.addElement(temp2);
				
				this.mapperScreen.append(temp1 + " " + temp2, null);
			}
		}else{
			if(logger != null) logger.logMessage("[GUI] " +"No rule records found");
		}
	}
	
	/*
	 * Adds a new mapper rule to the mapper object.
	 */
	
	private boolean addRule(){
		
		boolean[] flags = new boolean[2];
		this.friendlyOrAddress.getSelectedFlags(flags);
		String nameOrAddress;
		
		if(flags[0])
			nameOrAddress = this.friendlyName.getString();
		else
			nameOrAddress = this.btAddress.getString();
		
		int id = ((Integer)this.idName.get(this.nameIdArray[this.deviceId.getSelectedIndex()])).intValue();
		int type = -1;
		String toSend;
		
		if (flags[0]){
			type = Recorder.BT_FRIENDLY_NAME;
			toSend = nameOrAddress;
		}
		
		else{ 
			type = Recorder.BT_ADDRESS;
			// check if it's a valid BT address..
			if (nameOrAddress.length() != 12){
				Alert alert = new Alert("Syntax error");
	            alert.setTimeout(5000);
	            alert.setType(AlertType.ERROR);
	            alert.setString("The given bluetooth address is not valid.");
	            display.setCurrent(alert, this.display.getCurrent());
	            if(this.logger != null) this.logger.logMessage("[GUI] " +"Invalid bluetooth address input.");
				return false;
			}
			char[] validChars = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};
			
			nameOrAddress = nameOrAddress.toLowerCase();
			for (int i = 0; i < nameOrAddress.length(); ++i){
				boolean valid = false;
				for (int j = 0; j < validChars.length; ++j){
					if (nameOrAddress.charAt(i) == validChars[j])
						valid = true;
				}
				if (valid == false){
					Alert alert = new Alert("Syntax error");
		            alert.setTimeout(5000);
		            alert.setType(AlertType.ERROR);
		            alert.setString("Character " + i +" is invalid in the given Bluetooth address!");
		            display.setCurrent(alert, this.display.getCurrent());
		            if(this.logger != null) this.logger.logMessage("[GUI] " +"Character " + i +" is invalid in the given Bluetooth address!");
					return false;
				}				
			}
			toSend = nameOrAddress;
		}
		
		boolean success = this.mapper.saveRecord(type, id, toSend);
		if (!success){
			Alert alert = new Alert("Syntax error");
            alert.setTimeout(5000);
            alert.setType(AlertType.ERROR);
            alert.setString("Saving the new rule did not succeed. Please check the log for more details.");
            display.setCurrent(alert, this.display.getCurrent());
            if(this.logger != null) this.logger.logMessage("[GUI] " +"Saving the new rule did not succeed.");
			return false;
		}
		//if (this.logger != null) this.logger.logMessage("[GUI] " +"New rule added.");
		return true;
	}
	
	private void showLoadingBar(){
		this.loadingBar = new Gauge("Please wait...", false, 100, Gauge.CONTINUOUS_RUNNING);
		this.userInfo.append(this.loadingBar);
		this.userInfo.removeCommand(this.sendUserInfo);
		
		loadingTimer = new Timer();
	    InitUITask task = new InitUITask(this);
	    loadingTimer.schedule(task, 100);  
	}
	private void updateLoadingBar(){
		this.loadingBar.setValue(loadingPerc);
		loadingPerc += 10;
	}
	public void completeUIInit(){
		updateLoadingBar();
		
		this.userName = this.userNameField.getString();
		this.password = this.passwordField.getString();
		
		initializeBTG();
		
		boolean[] tempChoices = new boolean[1];
		this.rememberUserInfo.getSelectedFlags(tempChoices);
		this.remember = tempChoices[0];
		
		try {
			RecordStore.deleteRecordStore("btg_user");
			if(this.remember){
				RecordStore rs = RecordStore.openRecordStore("btg_user", true);
				byte[] temp = (this.userName + ";" + this.password).getBytes();
				rs.addRecord(temp, 0, temp.length);
				rs.closeRecordStore();
			}
		} catch (RecordStoreNotFoundException e) {
		} catch (RecordStoreException e) {
		}

		updateLoadingBar();

		b.setUsername(this.userName);
		LocalDevice ld;
		try {
			ld = LocalDevice.getLocalDevice();
			b.setBTAddress(ld.getBluetoothAddress());
		} catch (BluetoothStateException e) {
			e.printStackTrace();
		}

		updateLoadingBar();
		
		
		this.display.setCurrent(this.canvas);
		
	}
//		


	/**
	 * Adds a new plugin to the GUI. The plugin will be added
	 * the plugin screen which can be accessed using the options menu.
	 * @param plugin the plugin to be added.
	 */
	
	public void addPlugin(MIDPPlugin plugin){
		if (plugin != null){
			this.pluginList.append(plugin.getName(), null);
			
			
			
			//If this is the 1st plugin to be added, enable the access
			//to the plugin screen in the options menu.
			//if (this.plugins.size() == 0)
			//	this.canvas.addCommand(this.pluginSelect);
			this.plugins.addElement(plugin);
			
		}
	}
	
	/**
	 * Returns the DataGatherer object.
	 */
	
	public DataGatherer getDataGatherer(){
		return this.dgat;
	}

	public void update(Vector readers, Object connector) {
		Reader r = null;
		if (readers != null){
			for(int i = 0; i < readers.size(); i++){
				r = (Reader)readers.elementAt(i);
				if (!this.currentReaders.contains(r)){
					this.currentReaders.addElement(r);
					ReaderInformationObject rio = new ReaderInformationObject(r);
					reader_io.put(r, rio);
					this.canvas.addInformationObject(rio);
				}	
			}
			for (int i = 0; i < this.currentReaders.size(); i++){
				r = (Reader)this.currentReaders.elementAt(i);
				if (!readers.contains(r)){
					//This means that the reader has been disconnected... 
					this.canvas.removeInformationObject((ReaderInformationObject)reader_io.get(r));
				}
			}
			this.currentReaders = readers;
		}
	}
	
	public String getUserName(){
		return this.userName;
	}
	
	public String getPassword(){
		return this.password;
	}

	public void setGPSSampleInterval(int millis) {
		BetelgeuseMIDP.gpsRequestInterval = millis;
		// wakeup immediately if enough time has elapsed since last request
		if ((System.currentTimeMillis()-lastRequestTime) > millis)
			this.notifyAll();
	}
	public int getGPSSampleInterval() {
		return BetelgeuseMIDP.gpsRequestInterval;
	}
}

class InitUITask extends TimerTask {
	private BetelgeuseMIDP parent;
	private boolean completed = false;
	public InitUITask(BetelgeuseMIDP parent){
		this.parent = parent;
	}
	public void run() {
		if (!completed){
			this.parent.completeUIInit();
			this.cancel();
			this.completed = true;
		}
	}
}

