package hiit.blackboard.httpd;

import java.util.Date;

public class HttpResponse {
    private String version;
    private int statusCode;
    private String reasonPhrase;

    private String server;
    private boolean connectionPersist;

    private Date lastModified;
    private String contentType;
    private String charset;
    private int contentLength;

    private String entity;

    private static final String DEFAULT_SERVER = "Betelgeuse";
    private static final String DEFAULT_CONTENT_TYPE = "text/html";
    private static final String DEFAULT_CHARSET = "iso-8859-1";

    public static final int OK = 200;
    public static final int BAD_REQUEST = 400;
    public static final int NOT_FOUND = 404;
    public static final int METHOD_NOT_ALLOWED = 405;
    public static final int METHOD_NOT_IMPLEMENTED = 501;

    public HttpResponse(String entityBody) {
        this.version = HttpServer.HTTP_VERSION;
        this.statusCode = 200;
        this.reasonPhrase = "OK";

        this.server = DEFAULT_SERVER;
        this.connectionPersist = false;

        this.lastModified = new Date();
        this.contentType = DEFAULT_CONTENT_TYPE;
        this.charset = DEFAULT_CHARSET;
        this.contentLength = entityBody.length();

        this.entity = entityBody;
    }

    public static HttpResponse notFound(String path, String server) {
        String html = "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">"
                + "<html><head>"
                + "<title>404 Not Found</title>"
                + "</head><body>"
                + "<h1>Not Found</h1>"
                + "<p>The requested URL "
                + path
                + " was not found on this server.</p>"
                + "<hr>"
                + "<address>"
                + server + "</address>" + "</body></html>";
        HttpResponse r = new HttpResponse(html);
        r.setStatusCode(NOT_FOUND);
        r.setReasonPhrase("Not Found");
        return r;
    }
    
    public static HttpResponse badRequest(String explanation, String server) {
        String html = "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">"
            + "<HTML><HEAD>"
            + "<TITLE>400 Bad Request</TITLE>"
            + "</HEAD><BODY>"
            + "<H1>Bad Request</H1>"
            + "<P>Your browser sent a request that this server could not understand.</p>"
            + "<P>" + explanation+ "</P>"
            + "<HR>"
            + "<ADDRESS>" + server + "</ADDRESS>"
            + "</BODY></HTML>";
        HttpResponse r = new HttpResponse(html);
        r.setStatusCode(BAD_REQUEST);
        r.setReasonPhrase("Bad Request");
        return r;
    }
    
    public static HttpResponse methodNotAllowed(String method, String path, String server) {
        String html = "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">" 
            + "<HTML><HEAD>"
            + "<TITLE>405 Method Not Allowed</TITLE>"
            + "</HEAD><BODY>"
            + "<H1>Method Not Allowed</H1>"
            + "<P>The requested method " + method + " is not allowed for the URL " + path + ".</P>"
            + "<HR>"
            + "<ADDRESS>" + server + "</ADDRESS>"
            + "</BODY></HTML>";
        HttpResponse r = new HttpResponse(html);
        r.setStatusCode(METHOD_NOT_ALLOWED);
        r.setReasonPhrase("Method Not Allowed");
        return r;
    }
    
    public static HttpResponse methodNotImplemented(String method, String path, String server) {
        String html = "<!DOCTYPE HTML PUBLIC \"-//IETF//DTD HTML 2.0//EN\">" 
            + "<HTML><HEAD>"
            + "<TITLE>501 Method Not Implemented</TITLE>"
            + "</HEAD><BODY>"
            + "<H1>Method Not Implemented</H1>"
            + "<P> Method " + method + " not implemented.</P>"
            + "<HR>"
            + "<ADDRESS>" + server + "</ADDRESS>"
            + "</BODY></HTML>";
        HttpResponse r = new HttpResponse(html);
        r.setStatusCode(METHOD_NOT_IMPLEMENTED);
        r.setReasonPhrase("Method Not Implemented");
        return r;
    }

    public int getStatusCode() {
        return statusCode;
    }

    public void setStatusCode(int statusCode) {
        this.statusCode = statusCode;
    }

    public String getReasonPhrase() {
        return reasonPhrase;
    }

    public void setReasonPhrase(String reasonPhrase) {
        this.reasonPhrase = reasonPhrase;
    }

    public String getServer() {
        return server;
    }

    public void setServer(String server) {
        this.server = server;
    }

    public boolean isConnectionPersist() {
        return connectionPersist;
    }

    public void setConnectionPersist(boolean connectionPersist) {
        this.connectionPersist = connectionPersist;
    }

    public Date getLastModified() {
        return lastModified;
    }

    public void setLastModified(Date lastModified) {
        this.lastModified = lastModified;
    }

    public String getContentType() {
        return contentType;
    }

    public void setContentType(String contentType) {
        if (contentType.trim().equals("") || contentType.equals(null))
            this.contentType = this.DEFAULT_CONTENT_TYPE;
        else
            this.contentType = contentType;
    }

    public String getEntity() {
        return entity;
    }

    public void setEntity(String entity) {
        this.entity = entity;
        if (entity == null) {
        	this.contentLength = 0;
        } else {
        	this.contentLength = entity.length();
        }
    }

    public int getContentLength() {
        return this.contentLength;
    }

    public String getCharset() {
        return charset;
    }

    public void setCharset(String charset) {
        this.charset = charset;
    }

    public String toString() {
        String response = "";

        response += version + " " + statusCode + " " + reasonPhrase + "\r\n";
        response += "Server: " + server + "\r\n";
        if (connectionPersist) {
            response += "Connection: persist" + "\r\n";
        } else {
            response += "Connection: close" + "\r\n";
        }
        response += "Last-Modified: " + lastModified + "\r\n";
        response += "Content-Type: " + contentType;
        if (!(this.charset.equals("") || this.charset.equals(null))) {
            response += "; charset=" + charset;
        }
        response += "\r\n";
        response += "Content-Length: " + contentLength;
        
        if (entity != null && !entity.equals("")) {
        	response += "\r\n\r\n";
        	response += entity;
        }

        return response;
    }
}
