package hiit.blackboard.protocol;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;

/**
 * A class to ease communication with and within the Blackboard components.
 * Allows construction of the Blackboard protocol messages. Most messages are in
 * the following format:
 * 
 * <pre>
 *  Generic message format:
 *  +---------------------------+
 *  |c|type|user_id|component_id|
 *  +---------------------------+
 *   1   4    4         4
 * </pre>
 * 
 * Where <code>c</code> is a character, the message command, identifying the
 * purpose of the message, type is an integer, the type of the component
 * identified by the message, or Integer.MAX_VALUE to signify all types, user_id
 * is an integer that identifies a user within the Blackboard, and component_id
 * is an integer identifying a specific instance of a component type; 0 here
 * means any component of the type will do, Integer.MAX_VALUE means that all ids
 * are to be chosen. For example, a subscribe message could look like this: 's'
 * 0x00000001 0x00000003 0xFFFFFFFF and this message would result in the
 * receiving component subscribing to all components of component type 1 owned
 * by user 3. To read a generic message, just call <code>genericRead()</code>
 * with the InputStream containing the message (The command is assumed to have
 * been read already). Specific read methods for other messages will be
 * implemented here in the future. To send a generic message, build a message
 * with one of the methods in this class, giving Your parameters. The message
 * will be written to the <code>ByteArrayOutputStream</code> provided. Then
 * convert the stream to a bytearray and send that through the connection.
 */

// TODO: Verify new command message
// TODO: Verify added BB protocol header works

public class Protocol {
    
    public static final char PROTOCOL_ID = 'b';
    public static final int PROTOCOL_MAJOR = 1;
    public static final int PROTOCOL_MINOR = 0;
    
    /**
     * Writes the header common to all messages:
     * <pre></pre> 
     * @param bos the stream to write to.
     */
    public static void writeHeader(ByteArrayOutputStream bos){
        /*bos.write('b');
        bos.write(1);
        bos.write(0);*/
    }
    
    public static boolean readHeader(InputStream in){
      /*  try {
            char protocolId = readNextChar(in);
            int major = readNextChar(in);
            int minor = readNextChar(in);
            return protocolId == PROTOCOL_ID && major == PROTOCOL_MAJOR && minor >= PROTOCOL_MINOR;
        } catch (IOException e){
            return false;
        }*/
        return true;
    }
    
    public static void writeCreate(ByteArrayOutputStream bos, int type,
            int userid, String username) {
        if (bos == null)
            return;
        writeHeader(bos);
        bos.write('c');
        writeIntToBytes(type, 4, bos);
        writeIntToBytes(userid, 4, bos);
        byte[] ub = username.getBytes();
        writeGenericData(bos, ub);
    }
    
    public static Object[] readCreate(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        try {
            //System.out.println("reading create message.");
            int type = readNextInt(in);
            //System.out.println("Read type=" + type);
            int user = readNextInt(in);
            //System.out.println("Read user=" + user);
            byte[] uname = readGenericData(in);
            //System.out.println("Read username=" + new String(uname));
            if (type == -1 || user == -1 || uname == null)
                return null;
            Object[] ob = new Object[3];
            ob[0] = new Integer(type);
            ob[1] = new Integer(user);
            ob[2] = uname;
            return ob;
        } catch (IOException e) {
            System.out
                    .println("Could not read create message: " + e.toString());
        }
        return null;
    }
    
    public static void writeCreated(ByteArrayOutputStream bos, int type,
            int userid, int compid) {
        writeGenericMessage(bos, 'd', type, userid, compid);
    }
    
    public static int[] readCreated(InputStream is) {
        if (!readHeader(is)) // Not a valid BB Protocol message.
            return null;
        return readGeneric(is);
    }
    
    public static void writeSubscribe(ByteArrayOutputStream bos, int type,
            int userid, int compid) {
        writeGenericMessage(bos, 's', type, userid, compid);
    }
    
    public static int[] readSubscribe(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        return readGeneric(in);
    }
    
    public static void writePushData(ByteArrayOutputStream bos, int compId,
            byte[] data) {
        writeHeader(bos);
        if (bos == null || data == null)
            return;
        bos.write('p');
        writeIntToBytes(compId, 4, bos);
        writeGenericData(bos, data);
    }
    
    public static Object[] readPushData(InputStream in) {
    	
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        int compid;
        try {
            compid = readNextInt(in);
            return new Object[] { new Integer(compid), readGenericData(in) };
        } catch (IOException e) {
            System.out.println("Could not read pushData: " + e.toString());
        }
    	return null;
        
    }
    
    /**
     * Also known as the UpdateClient message.
     * 
     * @param bos
     * @param type
     * @param userid
     * @param compid
     * @param data
     */
    public static void writeReturnData(ByteArrayOutputStream bos, int type,
            int userid, int compid, byte[] data) {
        writeGenericMessage(bos, 'u', type, userid, compid);
        writeGenericData(bos, data);
    }
    
    /**
     * 
     * @param in
     * @return An Object[], with three integers: type, userid and componentid,
     *         followed by a byte[] containing the data.
     */
    
    public static Object[] readReturnData(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        int[] ints = readGeneric(in);
        byte[] b = readGenericData(in);
        if (ints == null || b == null)
            return null;
        Object[] ob = new Object[4];
        ob[0] = new Integer(ints[0]);
        ob[1] = new Integer(ints[1]);
        ob[2] = new Integer(ints[2]);
        ob[3] = b;
        return ob;
    }
    
    public static void writeRequestData(ByteArrayOutputStream bos, int type,
            int userid, int compid) {
        writeGenericMessage(bos, 'r', type, userid, compid);
    }
    
    public static int[] readRequestData(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        return readGeneric(in);
    }
    
    public static void writeListComponents(ByteArrayOutputStream bos, int type,
            int userid, int compid) {
        writeGenericMessage(bos, 'l', type, userid, compid);
    }
    
    public static int[] readListComponents(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        return readGeneric(in);
    }
    
    public static void writeComponentList(ByteArrayOutputStream bos,
            Vector comps) {
        if (bos == null || comps == null)
            return;
        writeHeader(bos);
        bos.write('m');
        writeIntToBytes(comps.size(), 4, bos);
        for (int i = 0; i < comps.size(); ++i) {
            ComponentInfo comp = (ComponentInfo) comps.elementAt(i);
            int type = comp.getComponentType();
            int userid = comp.getUserId();
            int compid = comp.getComponentId();
            writeTypeIdUserIdComponentId(bos, type, userid, compid);
        }
    }
    
    public static void writeComponentList(ByteArrayOutputStream bos,
            ComponentInfo[] comps) {
        if (bos == null || comps == null)
            return;
        writeHeader(bos);
        bos.write('m');
        writeIntToBytes(comps.length, 4, bos);
        for (int i = 0; i < comps.length; ++i) {
            int type = comps[i].getComponentType();
            int userid = comps[i].getUserId();
            int compid = comps[i].getComponentId();
            writeTypeIdUserIdComponentId(bos, type, userid, compid);
        }
    }
    
    public static ComponentInfo[] readComponentList(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        try {
            int numc = readNextInt(in);
            ComponentInfo[] comps = new ComponentInfo[numc];
            for (int i = 0; i < numc; ++i) {
                int type = readNextInt(in);
                int userId = readNextInt(in);
                int compId = readNextInt(in);
                comps[i] = new BasicComponentInfo(type, userId, compId);
            }
            return comps;
        } catch (IOException e) {
            System.out.println("Could not read component list: " + e.toString());
        }
        return null;
    }
    
    /**
     * Uses generic message format. Command: 'u'. Reply will be the UserList message, 'v'.
     * @param bos
     * @param type
     * @param userid
     * @param compid
     */
    public static void writeListUsers(ByteArrayOutputStream bos, int type,
            int userid, int compid) {
        writeGenericMessage(bos, 'u', type, userid, compid);
    }

    /**
     * Uses generic message format. Command: 'u'. Reply will be the UserList message, 'v'.
     * @param in
     */
    
    public static int[] readListUsers(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
       return readGeneric(in);
    }
    /**
     * 
     * Writes the UserList message, command: 'v'. NOTE: Message may contain no users at all.
     * In that case unum will be zero.
     * Message structure:
     * <pre>
     * cmd unum userid unamelen username ... userid username
     *  1   4      4      4        n            4       n
     * </pre>
     * @param bos The stream to write to
     * @param userIds The ids of matching connected users
     * @param userNames The names of matching connected users
     * 
     */
    public static void writeUserList(ByteArrayOutputStream bos, int[] userIds, String[] userNames) {
        if (userIds == null || userIds.length < 0 || bos == null)
            return;
        writeHeader(bos);
        bos.write('v');
        writeIntToBytes(userIds.length, 4, bos);
        for (int i = 0; i < userIds.length; ++i){
            writeIntToBytes(userIds[i], 4, bos);
            writeGenericData(bos, userNames[i].getBytes());
        }
    }
    
    /**
     * Reads the UserList message, command: 'v'. If the message contains no users,
     * will return an empty Object[].
     * @param in the stream to read from
     * @return And Object[] containing all the user ids and names, in sequence:
     * <pre>
     * Integer String Integer String ... </pre>
     */
    public static Object[] readUserList(InputStream in) throws IOException{
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        int num = readNextInt(in);
        Object[] ob = new Object[num*2];
        for (int i = 0; i < num; i = i+2){
            int uid = readNextInt(in);
            String uname = new String(readGenericData(in));
            ob[i] = new Integer(uid);
            ob[i + 1] = uname;
        }
        return ob;
    }
    
    
    public static void writeAckOrError(ByteArrayOutputStream bos,
            int ackedCommand, int status, String message) {
        if (bos == null)
            return;
        writeHeader(bos);
        bos.write('a');
        bos.write(ackedCommand);
        bos.write(status);
        byte[] b = message.getBytes();
        writeGenericData(bos, b);
        System.out.println("Wrote an ack message with command="
                + (char) ackedCommand + " status=" + status + " message="
                + message);
    }
    
    /**
     * 
     * @param in
     * @return an Object[], containing two integers that represent the command
     *         character and the status code, and a byte[], the error message.
     *         The byte[] is of length 0 in acks for successful execution.
     */
    public static Object[] readAckOrError(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        try {
            int ackCommand = in.read();
            int status = in.read();
            byte[] msg = readGenericData(in);
            if (status == -1 || ackCommand == -1 || msg == null)
                return null;
            Object[] ob = new Object[3];
            ob[0] = new Integer(ackCommand);
            ob[1] = new Integer(status);
            ob[2] = msg;
            return ob;
        } catch (IOException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        return null;
    }
    
    public static void writeUnsubscribe(ByteArrayOutputStream bos, int type,
            int userid, int compid) {
        writeGenericMessage(bos, 'z', type, userid, compid);
    }
    
    public static int[] readUnsubscribe(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        return readGeneric(in);
    }
    
    public static void writeDelete(ByteArrayOutputStream bos, int type,
            int userid, int compid) {
        writeGenericMessage(bos, 'x', type, userid, compid);
    }
    
    public static int[] readDelete(InputStream in) {
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        return readGeneric(in);
    }
    /**
     * Reads the Command Message:
     *  * <pre>
 *  +---------------------------+----------+------+----+----+------+
 *  |c|type|user_id|component_id|sender cid|signal|nump|plen|param |
 *  +---------------------------+----------+------+----+----+------+
 *   1   4    4         4            4         1     1    1     n
 * </pre>
 * Where plen and param are repeated nump times, and c is always 'k'.
     * @param in the stream to read the message from, command (k) already read.
     * @return An Object[] with type, userid, component_id, sender_cid as Integers,
     * signal as char, and parameters as Strings.
     */
    public static Object[] readCommandMessage(InputStream in){
        if (!readHeader(in)) // Not a valid BB Protocol message.
            return null;
        Vector stuff = new Vector();
        int[] params = readGeneric(in);
        
        try {
            int senderid = readNextInt(in);
            char signal = readNextChar(in);
            int nump = readNextInt(in);
            for (int i = 0; i < nump; ++i) {
                stuff.addElement(new String(readGenericData(in)));
            }
            Object[] obs = new Object[stuff.size()];
            for (int i = 0; i < params.length; ++i)
                obs[i] = new Integer(params[i]);
            obs[3] = new Integer(senderid);
            obs[4] = new Character(signal);
            for (int i = 5; i < stuff.size(); ++i){
                obs[i] = stuff.elementAt(i);
            }
            return obs;
        } catch (IOException e) {
            System.out
                    .println("Could not read component list: " + e.toString());
        }
        return null;
    }
    /**
     * Writes the command message
     * @param bos The ByteOutputStream to write to.
     * @param type The component type this command is sent to.
     * @param userid The user id of the target component.
     * @param compid The component id of the target component.
     * @param sendercid The component id of the sender id.
     * @param signal The signal or command sent to the component.
     * @param params The parameters for the signal.
     */
    public static void writeCommandMessage(ByteArrayOutputStream bos,
            int type, int userid, int compid, int sendercid, char signal, String[] params){
        if (bos == null)
            return;
        writeGenericMessage(bos, 'k', type, userid, compid);
        writeIntToBytes(sendercid, 4, bos);
        bos.write(signal);
        for (int i = 0; i < params.length; ++i)
            writeGenericData(bos, params[i].getBytes());
    }
    
    /**
     * Writes the common format of messages. Messages such as created, delete,
     * listcomponents, requestdata, subscribe and unsubscribe can be written
     * with this.
     * 
     * @param bos
     *            the stream to write in
     * @return
     */
    private static void writeGenericMessage(ByteArrayOutputStream bos, int cmd,
            int type, int userid, int compid) {
        if (bos == null)
            return;
        writeHeader(bos);
        bos.write(cmd);
        writeTypeIdUserIdComponentId(bos, type, userid, compid);
    }
    
    private static void writeTypeIdUserIdComponentId(ByteArrayOutputStream bos,
            int type, int userid, int compid) {
        if (bos == null)
            return;
        writeIntToBytes(type, 4, bos);
        writeIntToBytes(userid, 4, bos);
        writeIntToBytes(compid, 4, bos);
    }
    
    private static void writeGenericData(ByteArrayOutputStream bos, byte[] data) {
        if (bos == null)
            return;
        writeIntToBytes(data.length, 4, bos);
        for (int i = 0; i < data.length; ++i)
            bos.write(data[i]);
    }
    
    private static byte[] readGenericData(InputStream in) {
        if (in == null)
            return null;
        try {
            int datalen = readNextInt(in);
            if (datalen == -1)
                return null;
            byte[] ret = new byte[datalen];
            in.read(ret);
            return ret;
        } catch (IOException e) {
            System.out.println("Could not read data: " + e.toString());
        }
        return null;
    }
    
    /**
     * Reads the common format of messages. Messages such as
     * created, delete, listcomponents, requestdata, 
     * subscribe and unsubscribe can be read with this.
     * @param is
     * @return result
     */
    public static int[] readGeneric(InputStream is) {
        int[] resp = new int[3];
        try {
            resp[0] = readNextInt(is);
            resp[1] = readNextInt(is);
            resp[2] = readNextInt(is);
            return resp;
        } catch (IOException ioe) {
            System.out.println("Could not read the message from is.");
        }
        return null;
    }
    
    private static void writeIntToBytes(int number, int size,
            ByteArrayOutputStream out) {
        byte[] array = new byte[size];
        
        for (int i = size - 1; i >= 0; i--) {
            array[i] = (byte) (number % 256);
            number = number / 256;
        }
        try {
            out.write(array);
        } catch (IOException e) {
            // TODO
        }
    }
    
    private static int readNextInt(InputStream in) throws IOException {
        int temp = in.read() * 256 * 256 * 256;
        temp += in.read() * 256 * 256;
        temp += in.read() * 256;
        temp += in.read();
        return temp;
    }
    
    private static char readNextChar(InputStream in) throws IOException {
        return (char) in.read();
    }
}
