package hiit.blackboard.protocol;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * This class aims to implement the current Blackboard protocol as used by Capricorn,
 * for seamless transition to using this Blackboard.
 * This class should translate String messages into BTG Blackboard protocol messages,
 * so that a class using this protocol will be able to send and receive messages
 * conforming to the old Blackboard specification that are translated to messages
 * understood by the BTG Blackboard. 
 * TODO: Class not functional.
 * @author Eemil Lagerspetz
 *
 */
public class StringProtocolAdapter {
    /**

        4 Protocol format
        Clients communicate with the blackboard via a TCP/IP connection and by passing ASCII 
        messages to it. The format of the messages is as follows: cmd arg arg arg ... 
        Arguments are separated by spaces and each message ends with a newline. 
        If the arguments themselves contain whitespace they must be quoted either 
        with double quotes (") or enclosed with braces ({ and }). 
        The data contained within the arguments can use any encoding. 
        Each successful message to the blackboard will receive a response in return: 
        response <response-data> The contents of response-data depend on the 
        command sent to the blackboard. If no response is specified then an 
        empty string will be returned inside braces ({}). If an error occurs 
        an error message will be sent instead: error <error-code> <error-message> 
        The error codes are documented in section 5. The error-message part contains 
        a human-readable description of the error condition. A module must send a 
        register message before any other messages. Once registered any data sent 
        by the module will be stored until either the blackboard is brought down, 
        the data is removed or the module sends an unregister message.
        A module may not register more than once during a single connection and 
        two clients cannot register with the same module identifier. 
        If a module disconnects it may reregister itself with the same module 
        identifier upon reconnection.

        5 Error codes
        The client has already been registered. 
        Generic error contained for a condition caused by the command.

        clientRegistered cmdError
        
        nameRegistered notRegistered
        The module name has already been registered. 
        The client has not registered as a module. 
        No other commands can be sent before the register command.

        6 Commands

        6.1 direct <moduleID> <type> <msg>
        Sends a direct message through the blackboard to the module 
        identified by moduleID. type is the type of the message sent. 
        The understood types depend on the module in question.
         msg is the actual message sent to the module and depends on type. 
         If the type of the message is not understood, 
         the module should respond with a direct message and the type unknownType, 
         with msg containing information possibly useful to the caller.

        6.2 get <moduleID> <item> [<subItem>]
        Gets data stored by a module identified by moduleID. item specifies 
        the identifier of the item to retreive. The return value is a list 
        with two elements: type data Where type is the type of the value, 
        as set by replace and data is the actual data, quoted as appropriate 
        with either braces or quotes. F.ex.
        name {Kristoffer Lawson} birthyear 1977 company "Fishpool Creations" 
        This list has three keys: name, birthyear and company. It has three 
        values attached to those keys: Kristoffer Lawson, 1977 and Fishpool 
        Creations. The type of the returned value is simply specified as value. 
        Both the identifiers for the item and the sub-item are case-sensitive.

        6.3 getAll <moduleID>
        Returns all data set by the module identified by moduleID. 
        It is returned as a keyvalue list with alternating keys and
         values separated by whitespace. Each value contains first the 
         type of data and then the actual data. The values may be quoted
          in braces ({ and }) if the values themselves contain whitespace. 
          For example: 
          today {
            calendar {
              { location {Tktl} startTime {1184238000.0} endTime {1184241600.0}
                content {Stepwise}
              }
              { location {} startTime {1184274000.0} endTime {1184360400.0}
                content {Friday!}
              }
            }
          } lastUpdate { time 1184236190.0 } 
          The first entry above is today with data of the type calendar. 
          The next entry is called lastUpdate and its data is of type time.

        6.4 isConnected <moduleID>
        Returns true if there is a connected client which has registered 
        as the given module or false if the module has been registered 
        but there is no client currently connected with that registration.

        6.5 listen <moduleID or *>
        Listen to any updates performed by the module 
        identified by moduleID or all modules if * is given. 
        When changes are made the calling module will be informed 
        with a dataUpdate message (see section 7.1). The argument 
        moduleID can point to a module which has not yet registered. 
        When it does register the calling module will be 
        made a listener of that module.

        6.7 notifyGet <itemID>
        Specifies that the client wishes to be notified when another module reads, 
        with get, which has been set by this module. Currently getAll does not 
        trigger this notification. See section 7.2.

        6.8 register [-force] <userID> <moduleID>
        Register this client with the given moduleID for further purposes. 
        The moduleID can be any alphanumeric string. userID is the user 
        identifier and can also be any alphanumeric character. Case matters 
        with both arguments. A userID of * can be used to create a module 
        which is between all users. Apart from this, each user has their own
        unique set of modules registered. Module data is not shared between 
        users unless * is used as the user ID. Registration must be done 
        before any other activities. It is an error to register a module 
        with an ID that has already been registered with that user and 
        which has a client currently connected to it. If the client that 
        registered the module disconnects then the module data is not 
        lost and when a client connects and registers as that module 
        it can continue to operate from the point at which the disconnection 
        took place. The -force option can be used to force a currently 
        connected client to be disconnected if it is using the 
        moduleID specified in the sent command. This client can 
        then continue to act as that module. Returns old if the 
        module had been previously registered and the client 
        disconnected or new if this is a new module.

        6.9 replace <type> <item> <data>
        Store or replace data that exists for this module identified by item. 
        The type defines the kind of data. Both type and item can be any 
        string and should be agreed on by the module developers. 
        The item ID is case sensitive.

        6.10
        rm <item>
        Remove the data associated with the identifier item.

        6.11 unregister
        Removes all information about the module the client is 
        registered at and destroy all data.

        6.12 version
        Returns the version of the blackboard.

        7 Server notifications

        7.1 dataUpdate
        When a module uses the listen command to hear about updates the 
        blackboard will inform it with the following message:
        dataUpdate <userID> <moduleID> <type> <item> <data> moduleID 
        specifies the identifier of the module making the change and 
        userID the user identifier behind thatmodule. The other 
        arguments are the same as for the equivalent replace message. 
        When a module performs an update all listening modules 
        will be sent a dataUpdate before any more updates are allowed by any module.

        7.2 itemRead
        This message is sent when the client has used the getNotification 
        command to specify it wished to be notified every time another module 
        uses one of its items. The format of the message is: 
        itemRead <id> [sub-id] value 
        The id and sub-id fields match those of the equivalent get command. 
        The value field contains the value of the item on the blackboard 
        at the time it was passed to the requesting module.
     */
    
    /*    
     * Problem: BTG BB has no concept of clients and a limited one of users (no regging).
     * Therefore the term client will be a synonym for module in this protocol.
     * listClients will do the same as listModules. listUsers will return 0 (the local user)
     * and maybe other user ids that are currently connected. 
     */
    
    /**
     * Forcibly disconnects a client connected to the Blackboard. Uses a BTG BB
     * Command message with a quit command. 
     * @param clientId The module/client id for the target module to forcibly disconnect.
     * 
     */
    public static void writeDisconnect(ByteArrayOutputStream bos, String clientId, String sendercid) throws NumberFormatException{
        Protocol.writeCommandMessage(bos, 0, 0, Integer.parseInt(clientId), Integer.parseInt(sendercid), 'q', new String[0]);
    }
    
    /**
     * Reads the AckOrError message, returned as a reply to f.ex. Disconnect.
     * @param in the stream to read from.
     * @return the status message. Null if everything went ok.
     */
    public static String readAckOrError(InputStream in){
        String msg = new String((byte[])Protocol.readAckOrError(in)[2]);
        return msg.length() == 0 ? null: msg;
    }
    
    /** Command message format: 
     * An Object[] with type, userid, component_id, sender_cid as Integers,
     * signal as char, and parameters as Strings.
     * @return int[] with the module id to disconnect and the sender id.
     */
    public static int[] readDisconnect(InputStream in){
        Object[] commands = Protocol.readCommandMessage(in);
        int[] ret = new int[]{((Integer)commands[2]).intValue(),((Integer)commands[3]).intValue()};
        return ret; 
    }
    
    /**
     * See below.
     */
    public static void writeListClients(ByteArrayOutputStream bos){
        writeListModules(bos);
    }
    /**
     * Lists modules currently connected to the BTG BB. 
     */
    public static void writeListModules(ByteArrayOutputStream bos){
        Protocol.writeListComponents(bos, Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);
    }
    /**
     * Reads the resulting modules list.
     * @param in Stream to read from.
     * @return A ComponentInfo[] that contains the information of all connected modules (components).
     */
    public static ComponentInfo[] readModulesList(InputStream in){
        return Protocol.readComponentList(in); 
    }
    
    /**
     * Lists the user ids owning currently connected modules.
     */
    public static void writeListUsers(ByteArrayOutputStream bos){
        Protocol.writeListUsers(bos, Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MAX_VALUE);
    }
    /**
     * reads the user id list response. Returns an Object[] with userid and username for each user in sequence. 
     * @param in the stream to read from.
     * @return An Object[] with userid and username for each user in sequence.
     */
    public static Object[] readUsersList(InputStream in){
        try {
            return Protocol.readUserList(in);
        } catch (IOException e) {
            return null;
        }
    }
    
    public static void writeGetAll(ByteArrayOutputStream bos, String moduleId, String sendercid){
        writeGet(bos, moduleId, sendercid, null, null);
    }
    
    public static void writeGet(ByteArrayOutputStream bos, String moduleId, String sendercid, String item){
        writeGet(bos, moduleId, sendercid, item, null);    
    }
    
    /**
     * Writes a GET message to the specified module. Item and subitem may be specified. Sender's component (module) id must be specified.
     * @param bos Stream to write to.
     * @param moduleId Module to get data from.
     * @param sendercid Component requesting data.
     * @param item The data item name to get.
     * @param subitem The data subitem name to get.
     * @throws NumberFormatException If module or sender ids are not proper integers.
     */
    public static void writeGet(ByteArrayOutputStream bos, String moduleId, String sendercid, String item, String subitem) throws NumberFormatException{
        String[] params = null;
        if (item != null && subitem != null){
        params = new String[]{item, subitem};   
        }else if (item != null){
            params = new String[]{item};
            
        } else { // If item is null and subitem is not, we don't care.
            params = new String[0];
        }
        Protocol.writeCommandMessage(bos, 0, 0, Integer.parseInt(moduleId), Integer.parseInt(sendercid), 'g', params);
    }
    
    /**
     * Returns data given as response to Get. TODO: The data is in the format used by the module.
     * Parsing might be required. (Use Joonas's BTG Server to SW Blackboard ruleset?)
     *
     * @param in The stream to read from.
     * @return A String containing the data requested.
     */
    public static String readGet(InputStream in){
        return new String((byte[])Protocol.readReturnData(in)[3]);
    }
    
}
