/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.connectorUnit;

import hiit.BetelgeuseMIDP;
import hiit.mapper.*;
import hiit.observer.ConnectorObserver;
import hiit.reader.ProducerReaderSkeleton;
import hiit.reader.Reader;
import hiit.datagatherer.DataGatherer;
import hiit.logger.*;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.Vector;

import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.microedition.io.Connection;

/**
 * <p>BTDeviceConnector class performs the device and service discovery by using
 * BTDeviceListener class. Finally it creates readers for new devices with
 * BTReaderFactory class. A Thread of this class performs device discovery
 * periodically. The device discovery period can be adjusted with 
 * setDiscoveryWaitTime(..) method and it can be turned on/off with discoveryOn()
 * and discoveryOff() methods. If the the automatic device discovery is off 
 * (getDiscoveryMode() == false), calling performDiscovery() will cause the class
 * to run a single Bluetooth device discovery.</p>
 * 
 * <p>One discovery includes:
 * <ul>
 * <li>Search for Bluetooth devices</li>
 * <li>Creation of Readers for supported devices</li>
 * <li>Check of connected devices and Readers</li>
 * <li>Update of objects that have subscribed to BTDeviceConnector</li>
 * </ul></p>
 * 
 * <p>Note: If you do not use friendly-names in the Mapper rules, BTDeviceConnector
 * will not inquiry for friendly-names of nearby devices when it tries to determine
 * the devices' types. Thus not using the friendly-names in Mapper rules may
 * improve performance.</p>
 * @author Joonas Kukkonen and Eemil Lagerspetz
 *
 */


public class BTDeviceConnector implements Runnable{
	
	/* Mapper for devices types */
	private Mapper mapper;
	private Logger logger;
	private BTReaderFactory factory;
	/* Listener to find devices */
	private BTDeviceListener listener;
	private Vector processed = new Vector(); // RemoteDevice
	private Vector connectedRemote = new Vector(); // RemoteDevice
	private Vector allReaders = new Vector(); // READER
	private Hashtable readers_remote = new Hashtable();
	
	private DataGatherer gatherer;
	private int waitTime;
	private boolean isRunning = true;
	private Vector subscribers = new Vector();
	private boolean init;
	private int initDelay = 1000;
	private Vector lastFound = null;
	
	private long lastDiscoveryTime = 0;
	private int discoveryNumber = 0;

	private Vector ldr = new Vector();

	
	private boolean discoveryOn = false;
	
	/**
	 * Returns the DataGatherer.
	 * @return The DataGatherer object
	 */
	
	public DataGatherer getGatherer(){
		return this.gatherer;
	}
	
	/**
	 * Main constructor.
	 * @param listener the listener that searches for the devices
	 * @param mapper A mapper that knows about the device types that are compatible with DataGatherer
	 * @param logger Logs error and status messages, for mobile device compatibility
	 * @param waitTime The time to wait between subsequent searches for new devices
	 * @param dg the DataGatherer instance to give the new Readers to
	 * @param discoveryOn should automatic discovery be on when run() is called
	 */
	
	public BTDeviceConnector (BTDeviceListener listener, Mapper mapper, Logger logger, int waitTime, DataGatherer dg, boolean discoveryOn){
		this.gatherer = dg;
		this.listener = listener;
		this.mapper = mapper;
		this.logger = logger;
		this.waitTime = waitTime;
		this.discoveryOn = discoveryOn;
		this.factory = new BTReaderFactory(this);
		init = true;
	}
	
	/**
	 * This method can be used to create special Readers, which for example
	 * use local ckets to read data from the device.
	 * @param url the connection url (handling depends on the Reader in question)
	 * @param id the type of Reader defined in the Properties class
	 * @return true if the creation of the Reader was succesfull, false otherwise
	 */
	
	public boolean createSpecialReader(String url, int id, boolean ldr){
		Thread t;
		try{
			Reader reader = factory.createReaderForId(null, url, id, this.logger);
			if(reader != null && !reader.isDead()){
				this.allReaders.addElement(reader);
				t = new Thread(reader);
				t.start();
				if(ldr)
					this.ldr.addElement(reader);
				return true;
			}
		} catch (IllegalAccessException e){
			// Not complaining if not created, then we just don't get a local reader
		} catch (InstantiationException e){
			// Not complaining if not created, then we just don't get a local reader
		}
		return false;
	}
	
	public ProducerReaderSkeleton createProducerReader(Connection conn, String url, int deviceId, InputStream in, OutputStream out, int userId){
	    Thread t;
	    try{

	        ProducerReaderSkeleton reader = factory.createProducerReaderForId(conn, url, deviceId, logger, in, out, userId);
	        if(reader != null && !reader.isDead()){
	            this.allReaders.addElement(reader);
	            t = new Thread(reader);
	            t.start();
	            return reader;
	        }
	        
	    } catch (IllegalAccessException e){
	        // Not complaining if not created, then we just don't get a reader
	    } catch (InstantiationException e){
	        // Not complaining if not created, then we just don't get a reader
	    }
	    return null;
	}
	
	/**
	 * Starts the BTDeviceConnector. Device inquiry, reader creating and checking of old Readers
	 * and devices are done in this method.
	 */
	public synchronized void run(){
		Vector newDevices = null;
		Vector newReaders = null;
		Vector foundDevices = null;
		Vector temp = new Vector();
		Thread t = null;
		
		try {
			while(isRunning){
				boolean checkFriendlyName = this.mapper.friendlyNameUsed();
				//If the discovery is not turned on, wait for notification.
				if (!init && this.discoveryOn){
					try {
						this.wait(waitTime);
					} catch (InterruptedException e) {
						// Do nothing, we were woken up, that's fine.
					}
				}
				
				newReaders = null;
				foundDevices = null;
				newDevices = new Vector();	
		
				long startTime = System.currentTimeMillis();
				//Get devices from BTDeviceListener
				
				foundDevices = listener.getRemoteDevices();
				
				// This is used for example by LDR, to get the devices nearby at the moment.
				this.lastFound = foundDevices;
				if (foundDevices == null) return;
				//Compare the devices to previously processed ones. Also check if there is a mapper rule for the device.
				//Note that unknow devices do not get added to the processed vector, so those
				//devices will be discovered again... Inquiring for the mapper rule is fast.
				
				for (int i = 0; i < foundDevices.size(); i++){
					try {
						RemoteDevice device = (RemoteDevice)foundDevices.elementAt(i);
						// if (logger != null) logger.logMessage("[BTDC] " +"Device discovered: " + device.getBluetoothAddress());
										
						String name = null;
						
					        if(checkFriendlyName){
							try {
								name = device.getFriendlyName(false);
						     	} catch (IOException e) {
								// Nothing to do...
						     	}
						}
						int deviceID = this.mapper.getReaderIdentifier(device.getBluetoothAddress(), name);
					
						if (!this.processed.contains(device) && deviceID != Properties.UNKNOWN_DEVICE){
							newDevices.addElement(device);
						}
					} catch (Exception e){
						System.out.println("BTDeviceConnecor, run foundDevices: " + e);
						BetelgeuseMIDP.logger.logMessage("BTDeviceConnecor, run foundDevices: " + e);
					}
				}
				
				//Loop all the "canditates for readers"
				outer: for(int i = 0; i<newDevices.size(); i++){
					RemoteDevice device = (RemoteDevice)newDevices.elementAt(i);
					//Get services...
					Vector services = listener.getServices(device);
					if(services==null || services.size() == 0){
						continue outer;
					}
					//Check the responce code.
					int responceCode = ((Integer)services.elementAt(0)).intValue();
					switch(responceCode){
					case DiscoveryListener.SERVICE_SEARCH_DEVICE_NOT_REACHABLE:
						if(logger != null)
							logger.logMessage("[BTDC] " +"Device " + device.getBluetoothAddress() + " not in range.");
						continue outer;
					case DiscoveryListener.SERVICE_SEARCH_ERROR:
						if(logger != null)
							logger.logMessage("[BTDC] " +"Search error, " + device.getBluetoothAddress() + " not in range.");					
						continue outer;
					case DiscoveryListener.SERVICE_SEARCH_NO_RECORDS:
						if(logger != null)
							logger.logMessage("[BTDC] " +"Device " + device.getBluetoothAddress() + " does not support RFCOMM.");
						processed.addElement(device);
						continue outer;
					case DiscoveryListener.SERVICE_SEARCH_TERMINATED:
						if(logger != null)
							logger.logMessage("[BTDC] " +"Connectivity check terminated on new device: " + device.getBluetoothAddress());
						processed.addElement(device);
						continue outer;
					case DiscoveryListener.SERVICE_SEARCH_COMPLETED:
						//Nothing to do... at least for now
						//All the possible cases have been listed.
					}
					
					String name = null;
					if(checkFriendlyName){
						try {
							name = device.getFriendlyName(false);
						} catch (IOException e) {
							// Nothing to do...
						}
					}
					int deviceID = this.mapper.getReaderIdentifier(device.getBluetoothAddress(), name);
					
					if(services.size()<2)
						continue outer;			
					ServiceRecord srec = (ServiceRecord)services.elementAt(1);
					
					//Now we should have everything we need for creation of a Reader...
					
					Reader re = this.factory.createReaderForDevice(deviceID, device, srec, logger);
					if(re!=null){
						//Add the RemoteDevice object to the processed vector
						this.processed.addElement(device);
						//Map the reader and remote device
						readers_remote.put(re, device);
						//Add it to allReaders vector
						allReaders.addElement(re);
						//Add the remote device to "connected remote devices" vector
						connectedRemote.addElement(device);
						t = new Thread(re);
						t.start();
					}
				}
				
				if (foundDevices.size() == 0){
					if (logger != null)
						logger.logMessage("[BTDC] No devices discovered!");
				}
				
				//check them...
				temp = new Vector();
				newReaders = new Vector();
				int size = allReaders.size();
				check: for(int i = 0; i < size; i++){
					//First step is to check that the devices have been mapped to correct readers
					//The user might have changed the reader...
					Reader re = (Reader)allReaders.elementAt(i);
					RemoteDevice rd = (RemoteDevice)readers_remote.get(re);
					//The rd is null if re is the LocalDeviceReader
					if (rd == null && !re.isDead()){
						temp.addElement(re);
						continue check;
					}else if (rd == null && re.isDead()){
						this.allReaders.removeElement(re);
						size--;
						i--;
						continue check;
					}
					
					String bta = rd.getBluetoothAddress();
					String name = null;
					if(checkFriendlyName){
					try {
						name = rd.getFriendlyName(false);
					} catch (IOException e) {
						// Nothing to do
					}
					}
					int id = mapper.getReaderIdentifier(bta, name);
					if (re.getDeviceId() != id){
						if (logger != null) logger.logMessage("[BTDC] " +"Device ID has changed for device " + re.getBluetoothAddress() + ". Trying to create a new Reader");
						//Ouch! The user has been playing arround with the mapper rules!
						re.close();
						//Remove any references to the old reader...
						allReaders.removeElement(re);
						size--;
						i--;
						readers_remote.remove(re);
						//Double ouch! Somebody removed the rule and did not add a new one...
						if(id == Properties.UNKNOWN_DEVICE){
							this.processed.removeElement(rd);
							connectedRemote.removeElement(rd);
							//Don't even try to create a new Reader...
							continue check;
						}
						
						//To create a new reader, we need the services again...
						Vector services = listener.getServices(rd);
						//Did we get any services?
						if(services == null || services.size()<2)
							continue check;
						ServiceRecord srec = (ServiceRecord)services.elementAt(1);
						//Create a new reader for it...
						re = this.factory.createReaderForDevice(id, rd, srec, logger);
						if (re != null){
							if(logger != null) logger.logMessage("[BTDC] " +"New " + re.getClass().getName() + " created for " + re.getBluetoothAddress());
							newReaders.addElement(re);
							readers_remote.put(re, rd);
							t = new Thread(re);
							t.start();
							continue check;
						}
						//For some reason we failed to create a new reader for the object
						processed.removeElement(rd);
						connectedRemote.removeElement(rd);
						continue check;
						
					}
					
					if (!re.isConnected()){
						if (logger!=null) logger.logMessage("[BTDC] " +"Device " + re.getBluetoothAddress() + " has been disconnected from a Reader");
					}else{
						//The reader is connected, all is ok...
						temp.addElement(re);
						continue;
					}
					
					if (re.isDead()){
						if (logger != null) logger.logMessage("[BTDC] " +"Device " + re.getBluetoothAddress() + " has been disconnected.");
						//Problems! The reader is not connected and it is dead... Remove it from the vector...
						allReaders.removeElement(re);
						connectedRemote.removeElement(rd);
						processed.removeElement(rd);
						readers_remote.remove(re);
						continue;
					}
				}
				
				for (int i=0; i<newReaders.size();i++){
					temp.addElement(newReaders.elementAt(i));
					allReaders.addElement(newReaders.elementAt(i));
				}
				
				if (init){
					try {
						this.wait(initDelay);
					} catch (InterruptedException e) {
					}
					init = false;
				}
				
				long endTime = System.currentTimeMillis();
				discoveryNumber++;
				lastDiscoveryTime = endTime - startTime;
				notifySubscribers(temp);
				
				if(!discoveryOn){
					try{
						this.wait();
					}catch(InterruptedException e){
						//Do nothing
					}
				}
				
			}
			if(this.logger!=null) {
				this.logger.flushLog();
				this.logger.clearLog();
			}
		} catch (Exception e){
			System.out.println("BTDeviceConnector, run ERR: " + e);
			BetelgeuseMIDP.logger.logMessage("BTDeviceConnector, run ERR: " + e);
		}
	}
	
	/**
	 * Returns the discovery mode.
	 * @return true if discovery is done periodically, false if discovery
	 * can be performed by calling performDiscovery() method.
	 */
	
	public boolean getDiscoveryMode(){
		return this.discoveryOn;
	}
	
	/**
	 * Turns the automatic discovery of devices on. Discovery is performed periodically
	 * according to the discovery timeout.
	 */
	
	public void discoveryOn(){
		this.discoveryOn = true;
		synchronized(this){
			this.notify();
		}
	}
	
	/**
	 * Turns the automatic device discovery off. Note that single device discovery can
	 * be performed by calling notify() for this object.
	 */
	
	public void discoveryOff(){
		if(this.discoveryOn == true){
			synchronized(this){
				this.discoveryOn = false;
				this.notify();
			}
		}
	}
	
	/**
	 * Stops the Connector.
	 */
	public void close(){
		this.isRunning = false;
		if(this.allReaders==null) {
			return;
		}
		for(int i=0,j=this.allReaders.size();i<j;i++) {
			((Reader)this.allReaders.elementAt(i)).close();
		}
	}
	
	/**
	 * Sets the time to wait between subsequent searches for new devices
	 * @param waitTime The time to wait between subsequent searches for new devices
	 */
	
	public void setDiscoveryWaitTime (int waitTime){
		this.waitTime = waitTime;
	}
	
	/**
	 * Returns the time to wait between subsequent searches for new devices.
	 * @return the wait time in milliseconds
	 */
	
	public int getDiscoveryWaitTime (){
		return this.waitTime;
	}
	
	/**
	 * Returns a Vector of devices that have been found during the last query.
	 * @return a Vector of RemoteDevice objects
	 */
	
	public Vector getLastFound() {
		Vector temp = new Vector();
		if (lastFound == null) return null;
		for (int i=0; i<lastFound.size(); i++)
			temp.addElement(lastFound.elementAt(i));
		
		return temp;			
	}
	
	/**
	 * Returns a Vector of all the Readers for devices that have been connected.
	 * @return a Vector of Readers.
	 */
	
	public Vector getConnectedDevices() {
		Vector temp = new Vector();
		
		for (int i=0; i<allReaders.size(); i++)
			if(((Reader)allReaders.elementAt(i)).isConnected())
				temp.addElement(allReaders.elementAt(i));
		
		return temp;			
	}
	
	/**
	 * Returns a Vector of all the devices that have been connected.
	 * @return a Vector of RemoteDevices objects that are connected to BeTelGeuse
	 */
	
	public Vector getConnectedRemoteDevices() {
		Vector temp = new Vector();
		
		for (int i=0; i<connectedRemote.size(); i++)
			temp.addElement(connectedRemote.elementAt(i));
		
		return temp;			
	}
	
	/**
	 * BTReaderFactory handles connection restoration for bluetooth Readers.
	 * If connection restoration fails, the Reader needs to be removed from the connected
	 * devices vector.
	 * @param r Reader that was disconnected.
	 */
	
	public void disconnected(Reader r){
		if (this.allReaders.contains(r))
			this.allReaders.removeElement(r);
	}
	
	/**
	 * This method allows BTReaderFactory to announce a Reader whose connection was restored.
	 * @param r Reader that was reconnected.
	 */
	
	public void reconnected(Reader r){
		if (!this.allReaders.contains(r))
			this.allReaders.addElement(r);
	}
	
	/**
	 * Subscribes the object to device list updates.
	 * @param obs The ConnectorObserver that subscribes to events.
	 */
	public void subscribe(ConnectorObserver obs) {
		synchronized(this) {
			this.subscribers.addElement(obs);
		}
	}
	/**
	 * Unsubscribes the object from receiving device list updates.
	 * @param obs The ConnectorObserver that unsubscribes.
	 */
	public void unsubscribe(ConnectorObserver obs) {
		synchronized(this) {
			this.subscribers.removeElement(obs);
		}
	}
	/**
	 * Notifies all subscribers that new devices have been found.
	 */
	protected void notifySubscribers(Vector temp) {
		synchronized(this){
			for(int i=0,j=this.subscribers.size();i<j;i++) {
				
				//Send readers to the observer. Observer decides what it does with the data.
				try {
					((ConnectorObserver)this.subscribers.elementAt(i)).update(temp, this);
				} catch (Exception e){
					System.out.println("BTDeviceConnector: " + e);
					BetelgeuseMIDP.logger.logMessage("BTDeviceConnector, notifySubscribers: " + e);
				}
				//Wake the observer
				//synchronized(this.subscribers.elementAt(i)) {
				//	this.subscribers.elementAt(i).notify();
				//}
			}
		}
	}
	
	/**
	 * Returns the length of last device discovery in milliseconds.
	 * @return the length of last device discovery
	 */
	
	public long getLastDiscoveryTime(){
		return lastDiscoveryTime;
	}
	
	/**
	 * Returns the discovery number of last discovery.
	 * @return the discovery number
	 */
	
	public long getDiscoveryNumber(){
		return discoveryNumber;
	}
	

	/**
	 * Returns a Vector of the LocalDeviceReaders
	 * @return a Vector of LocalDeviceReaders
	 */
	
	public Vector getLDR(){
		return this.ldr;
	}
	
	/**
	 * Calling this method will cause BTDeviceConnector to run a 
	 * device discovery, if the automatic discovery is off.
	 */
	
	public synchronized void performDiscovery(){
		if (!this.discoveryOn)
			this.notify();
	}
}
