/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.connectorUnit;

import javax.bluetooth.*;
import javax.microedition.rms.RecordComparator;

import java.io.IOException;
import java.util.Vector;

import hiit.BetelgeuseMIDP;
import hiit.logger.*;
import hiit.util.RemoteDeviceSorter;

/**
 * This class offers services for discovering Bluetooth devices and services.
 * @author Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 *
 */

public class BTDeviceListener implements DiscoveryListener{
	
	private DiscoveryAgent dAgent;
	private Vector rdevices = new Vector();
	//private Vector srecords = new Vector();
	private Vector trecords = new Vector();
	private Logger logger;
	
	private final UUID[] uuidList = {
			new UUID(0x0003)
	};
	
	private int activeTransactionID = 0;
	
	/**
	 * Creates a new BTDeviceListener.
	 * @param ldevice The local device object.
	 * @param logger The Logger object used to log error messages.
	 */
	
	public BTDeviceListener(LocalDevice ldevice, Logger logger) {
		
		// Set
		try {
			ldevice.setDiscoverable(DiscoveryAgent.GIAC);
		}catch(BluetoothStateException b) {
			if(logger!=null)
				logger.logMessage("[BTDL] " +"Line 28: " + b.toString());
		}
		// Create the discovery agent that is used to discover devices and services
		this.dAgent = ldevice.getDiscoveryAgent();
		
		// 
		this.logger = logger;
	}
	
	/**
	 * This method is called each time a new device is discovered during
	 * device discovery.
	 */
	
	public void deviceDiscovered(RemoteDevice rdev, DeviceClass code) {
		if(rdev!=null & this.rdevices!=null)
			this.rdevices.addElement(rdev);
	}
	
	/**
	 * This method is called when device discovery has been completed.
	 */
	
	public void inquiryCompleted(int discoveryType) {
		synchronized(this) {
			this.notifyAll();
		}
	}
	
	/**
	 * This method is called each time a service is discovered on a remote device.
	 */
	
	public void servicesDiscovered(int transactionID, ServiceRecord[] serviceRecord) {
		/*		if(serviceRecord!=null)
		 logger.logMessage("[BTDL] " +"Services discovered " + serviceRecord.length);
		 */
		for(int i=0,j=serviceRecord.length;i<j;i++)
			this.trecords.addElement(serviceRecord[i]);
		
		/*
		try{
			for(int i=0,j=serviceRecord.length;i<j;i++)
				logger.logMessage("[BTDL] " +serviceRecord[i].getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT,false));
		}catch (Throwable e){
			if (logger != null) logger.logMessage("[BTDL] " +"Debug: 1");
			if (logger != null) logger.logMessage("[BTDL] " +e.toString());
		}
		*/
	}
	
	/**
	 * This method is called when service discovery on remote device has
	 * been completed.
	 */
	
	public void serviceSearchCompleted(int transactionID, int responseCode) {
		trecords.insertElementAt(new Integer(responseCode), 0);
		synchronized(this) {
			this.notifyAll();
		}
	}
	
	/**
	 * A method for checking if a known device is currently in range.
	 * @param rDevice The device we are interested in
	 * @return DiscoveryListener's response code or -1 if an error occured
	 */
	public synchronized int deviceInRange(RemoteDevice rDevice){
		if(this.dAgent==null)return -1;
		
		this.trecords.removeAllElements();
		
		try {											
			this.activeTransactionID = this.dAgent.searchServices(null, this.uuidList, rDevice,this);	
			if(this.activeTransactionID>=0) {
				//Wait for the inquiry to complete
				this.wait();
		
			}
		} catch(Exception ex) {
			if(logger!=null)
				logger.logMessage("[BTDL] " +"Line 99: " + ex.toString());
		}

		return ((Integer)trecords.elementAt(0)).intValue();
	}
	
	
	/**
	 * This method searches and returns all the services on a bluetooth device.
	 * @param rDevice The remote bluetooth device
	 * @return A Vector of services.
	 */
	
	public synchronized Vector getServices(RemoteDevice rDevice) {
		
		// If the discovery agent is null, initialization has failed. In this case return null.
		if(this.dAgent==null)return null;
		
		// Clear the list of service records
		this.trecords.removeAllElements();
		
		try {											
			this.activeTransactionID = this.dAgent.searchServices(null,this.uuidList,rDevice,this);
			if(this.activeTransactionID>=0) {	
				//Wait for service discovery to complete
				this.wait();
				
			}else {
				if(logger!=null)
					logger.logMessage("[BTDL] " +"Service discovery failed");
				return null;
			}			
			
		} catch(Exception ex) {
			if(logger!=null)
				logger.logMessage("[BTDL] " +"Line 99: " + ex.toString());
		}
		
		int responseCode = ((Integer)trecords.elementAt(0)).intValue();
		
		if(logger!=null){
			if(DiscoveryListener.INQUIRY_COMPLETED == responseCode)
				logger.logMessage("[BTDL] " +"Response code: " + responseCode + " - Inquiry completed!");
			else if(DiscoveryListener.INQUIRY_ERROR == responseCode)
				logger.logMessage("[BTDL] " +"Response code: " + responseCode + " - Inquiry failed to compelete normally!");
			else if(DiscoveryListener.INQUIRY_TERMINATED == responseCode)
				logger.logMessage("[BTDL] " +"Response code: " + responseCode + " - Inquiry canceled by application!");
			else if(DiscoveryListener.SERVICE_SEARCH_COMPLETED == responseCode);
			else if(DiscoveryListener.SERVICE_SEARCH_DEVICE_NOT_REACHABLE == responseCode)
				logger.logMessage("[BTDL] " +"Response code: " + responseCode + " - Device is not reachable!");
			else if(DiscoveryListener.SERVICE_SEARCH_ERROR == responseCode){
				logger.logMessage("[BTDL] " +"Response code: " + responseCode + " - Service search terminated due to an error!");
			}
			else if(DiscoveryListener.SERVICE_SEARCH_NO_RECORDS == responseCode){
				logger.logMessage("[BTDL] " +"Response code: " + responseCode + " - No service records found on the device!");				
			}
			else if(DiscoveryListener.SERVICE_SEARCH_TERMINATED == responseCode)
				logger.logMessage("[BTDL] " +"Response code: " + responseCode + " - Service search canceled by the application!");
		}
		
		if(logger!=null)
			logger.logMessage("[BTDL] " + (trecords.size() - 1) + " service(s) discovered on remote device " + rDevice.getBluetoothAddress());
		return this.trecords;
	}
	
	/**
	 * This method searches for remote bluetooth devices.
	 * @return A Vector of remote devices.
	 */
	public synchronized Vector getRemoteDevices() {
		
		// If the discovery agent is null, initialization has failed. In this case return null.
		if(this.dAgent==null)return null;
		
		// Clear the list of devices
		this.rdevices.removeAllElements();
		
		// Perform device discovery, use the general/unlimited access code (GIAC)
		try {
			// The startInquiry method returns true if the discovery was successfully started. 
			// If there is an error starting the discovery, output an error message and quit. 
			if(this.dAgent.startInquiry(DiscoveryAgent.GIAC,this)) { 
				if(logger!=null)
					logger.logMessage("[BTDL] " +"Inquiring for new devices");
			}else {
				if(logger!=null)
					logger.logMessage("[BTDL] " +"Error starting inquiry");
				return null;
			}
		}catch(BluetoothStateException bex) {
			// In case of a Bluetooth exception, return null.
			if(logger!=null) {
				logger.logMessage("[BTDL] " +"Error performing bluetooth discovery");
				logger.logMessage("[BTDL] " +"Line 133: " + bex.toString());
			}
			return null;
		}
		
		// Wait for inquiry to finish
		try{
			this.wait();
			
		}catch(Exception ex) {
			if(logger!=null) {
				logger.logMessage("[BTDL] " +"Synchronization error, application will quit");
				logger.logMessage("[BTDL] " +"Line 146: " + ex.toString());
			}
			return null;
        }
		RemoteDevice[] rdev = this.dAgent.retrieveDevices(DiscoveryAgent.PREKNOWN);	
		
		if(rdev!=null) {
			if(logger!=null)
				logger.logMessage("[BTDL] " +"Adding " + rdev.length + " preknown devices.");
			for(int i=0,j=rdev.length;i<j;i++) {				
				if(!this.rdevices.contains(rdev[i])) {
					try {
						if(logger!=null)
							logger.logMessage("[BTDL] " +(i+1) + "\t" + rdev[i].getBluetoothAddress() + "\t" + rdev[i].getFriendlyName(false));
					}catch(Exception ex) {
						if(logger!=null)
							logger.logMessage("[BTDL] " +"Line 161: " + ex.toString());
					}
					this.rdevices.addElement(rdev[i]);
				}
			}
		}
		if(logger!=null)
			logger.logMessage("[BTDL] " +"Discovery completed...");
		// Return the discovered devices
		
		
		
		if (BetelgeuseMIDP.emitBTDevicesChangeEvents == true) {
			// Device-list should be sorted, when 'BTDeviceChange' events are emitted
			// Otherwise, unnecessary events could be sent, since the devices appear in a random order on the list.
			// 
			try {
				int cnt = this.rdevices.size();
				RemoteDevice deviceArr[] = new RemoteDevice[cnt];
				for (int i=0; i<cnt; i++ ){
					deviceArr[i] = (RemoteDevice)this.rdevices.elementAt(i);
				}
				
				RemoteDeviceSorter.quicksort(deviceArr);
				
				this.rdevices = new Vector();
				for (int i=0; i<deviceArr.length; i++){
					this.rdevices.addElement(deviceArr[i]);
				}
								
			} catch (Exception e){
				BetelgeuseMIDP.logger.logMessage("BTDeviceListener, sort devices ERR: " + e);
			}
			// 
		}		
		
		return this.rdevices;
    }
	
}
