/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package hiit.connectorUnit;

import javax.bluetooth.*;
import javax.microedition.io.Connection;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import hiit.reader.*;
import hiit.logger.Logger;
import hiit.mapper.*;

/**
 * This class offers services for creating Reader objects for bluetooth devices.
 * @author Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 *
 */

public class BTReaderFactory {
	
	private BTDeviceConnector connector = null;
	
	public BTDeviceConnector getConnector(){
		return connector;
	}
	
	/**
	 * Creates a new BTReaderFactory.
	 * @param connector the BTDeviceConnector object that handles the device discovery and reader creation
	 */
	
	public BTReaderFactory(BTDeviceConnector connector){
		this.connector = connector;
	}
	
	/**
	 * This method creates a Reader object for a given device. If no Reader object can be created,
	 * null is returned.
	 * @param deviceID the device identification number
	 * @param rDev the RemoteDevice object of the bluetooth sensor
	 * @param srec the service record to be used in the creating of the Reader
	 * @param logger the Logger used to log error messages
	 * @return a Reader
	 */
	
	public Reader createReaderForDevice(int deviceID, RemoteDevice rDev, ServiceRecord srec, Logger logger) {
		try {
			String url = null;
			try{
				//if (logger != null) logger.logMessage("[BTRF] " +"Trying to get the connection url.");
				url = srec.getConnectionURL(ServiceRecord.NOAUTHENTICATE_NOENCRYPT,false);
			}catch (Throwable e){
				//Bad programming style, but what can you do.
				//getConnectionURL() method may hang up and only way to
				//to detect it is to catch a throwable
				if (logger != null) logger.logMessage("[BTRF] " +e.toString());
				e.printStackTrace();
				return null;
			}
			
			boolean mode = ((Boolean)(Properties.getIdMode().get(new Integer(deviceID)))).booleanValue();
			if(!mode){
				return createReaderForId(null, url, deviceID, logger);
			}
			
			StreamConnection conn = null;				
			if(url.indexOf("btspp://")!=-1) {					
				try {
					if (logger != null) logger.logMessage("[BTRF] " +"Trying to open a connection to " + rDev.getBluetoothAddress());
					conn = (StreamConnection)Connector.open(url);		
				}catch(BluetoothConnectionException e) {
					int status = e.getStatus();
					switch(status) {
					case BluetoothConnectionException.FAILED_NOINFO:
						logger.logMessage("[BTRF] " +"FAILURE - The connection to the server failed due to unknown reasons.");
						break;
					case BluetoothConnectionException.NO_RESOURCES:
						logger.logMessage("[BTRF] " +"FAILURE - A lack of resources either on the local device or on the remote device.");
						break;
					case BluetoothConnectionException.SECURITY_BLOCK:
						logger.logMessage("[BTRF] " +"FAILURE - The security settings on the local device or the remote device were incompatible with the request.");
						break;
					case BluetoothConnectionException.TIMEOUT:
						logger.logMessage("[BTRF] " +"FAILURE - Timed out.");
						break;
					case BluetoothConnectionException.UNACCEPTABLE_PARAMS:
						logger.logMessage("[BTRF] " +"FAILURE - The configuration parameters provided were not acceptable to either the remote device or the local device.");
						break;
					case BluetoothConnectionException.UNKNOWN_PSM:
						logger.logMessage("[BTRF] " +"FAILURE - No service for the given PSM was registered.");
						break;
					}
				}catch (IOException e){
					if (logger != null) logger.logMessage("[BTRF] " +e.toString());
				}
				
			}
			//Create right reader for the device
			if(conn != null){
				return createReaderForId(conn, url, deviceID, logger);
			}
		}catch(Exception ex) {
			if (logger != null) logger.logMessage("[BTRF] " +ex.toString());
		}
		return null;
		
	}
	
	/**
	 * This method can be used to create a Reader for specific Reader ID number and url.
	 * @param conn the connection object the Reader should use to read data (note: this can be null if the Reader in question does not use normal connections)
	 * @param url the connection url
	 * @param id the Reader ID (determines which Reader is created, see Properties class)
	 * @param logger the Logger object used to log error messages
	 * @return a Reader, null if the creation failed
	 * @throws InstantiationException if the Reader's class is missing, or invalid class name was defined in Properties class
	 * @throws IllegalAccessException
	 */
	
	public Reader createReaderForId(Connection conn, String url, int id, Logger logger) throws InstantiationException, IllegalAccessException{
		Hashtable table = Properties.getIdClassName();
		String name = (String)table.get(new Integer(id));
		
		Reader r = null;
		Class c = null;
		try{
			c = Class.forName(name);
		} catch (ClassNotFoundException cnfe){
			if (logger != null) logger.logMessage("[BTRF] " +"Cannot find class "+ c + " for name=" + name + " and id=" + id);
		}
		
		Object o = c.newInstance();
		try{
			r = (Reader)o;
		}catch(Exception e){
			if (logger != null) logger.logMessage("[BTRF] " +"ERROR ERROR ERROR");
			e.printStackTrace();
			return null;
		}
		
		if (logger != null) logger.logMessage("[BTRF] " +"Creating a " + r.getDeviceName() + " reader.");
		boolean mode = ((Boolean)(Properties.getIdMode().get(new Integer(id)))).booleanValue();
		r.initialize(conn, url, logger, this, id, mode);
		
		return r;
	}
	
	   /**
     * This method can be used to create a Reader for specific Reader ID number and url. This method can be used to pass the stream objects to the new Reader.
     * However, this only works for Readers that extend ProducerReaderSkeleton.
     * @param conn the connection object the Reader should use to read data (note: this can be null if the Reader in question does not use normal connections)
     * @param url the connection url
     * @param id the Reader ID (determines which Reader is created, see Properties class)
     * @param logger the Logger object used to log error messages
     * @param in the InputStream of the connection
     * @param out the OutputStream of the connection
     * @return a Reader, null if the creation failed
     * @throws InstantiationException if the Reader's class is missing, or invalid class name was defined in Properties class
     * @throws IllegalAccessException
     */
    
    public ProducerReaderSkeleton createProducerReaderForId(Connection conn, String url, int id, Logger logger, InputStream in, OutputStream out, int userId) throws InstantiationException, IllegalAccessException{
        Hashtable table = Properties.getIdClassName();
        String name = (String)table.get(new Integer(id));
        
        ProducerReaderSkeleton r = null;
        Class c = null;
        try{
            c = Class.forName(name);
        } catch (ClassNotFoundException cnfe){
            if (logger != null) logger.logMessage("[BTRF] " +"Cannot find class "+ c + " for name=" + name + " and id=" + id);
        }
        
        Object o = c.newInstance();
        try{
            r = (ProducerReaderSkeleton)o;
        }catch(Exception e){
            if (logger != null) logger.logMessage("[BTRF] " +"ERROR ERROR ERROR");
            e.printStackTrace();
            return null;
        }
        
        if (logger != null) logger.logMessage("[BTRF] " +"Creating a " + r.getDeviceName() + " reader.");
        boolean mode = ((Boolean)(Properties.getIdMode().get(new Integer(id)))).booleanValue();
        r.initialize(conn, url, logger, this, id, mode, in, out, userId);
        
        return r;
    }
	
	
	
	/**
	 * Reader objects use this method to restore their connections, should the
	 * connection be lost to the sensor.
	 * @param r the reader object
	 * @param url the device URL address
	 * @param logger logger object for error logging
	 * @return a new connection object to the sensor
	 */
	
	public javax.microedition.io.Connection tryRestore(Reader r, String url, Logger logger){
		StreamConnection conn = null;
		
		try {
			//Open a connection to the device
			conn = (StreamConnection)Connector.open(url);
			this.connector.reconnected(r);
			return conn;
		} catch(java.io.IOException ex) {
			// Occurs every time the connection cannot be restored. Readers inform when they are reconnected 
			// or the connection has been hopelessly lost.
		}
		this.connector.disconnected(r);
		return null;
	}
}
