/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.datagatherer;

import javax.bluetooth.*;

import hiit.connectorUnit.*;
import hiit.mapper.*;
import hiit.logger.*;

/**
 * The main initialization logic of BeTelGeuse.
 * @author Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 *
 */
public class DataGatherer{
	
	protected Logger logger;
	protected BTDeviceConnector connector;
	protected Mapper mapper;
	protected LocalDevice ldev;
	
	/*
	public DataGatherer() {
		this("hiit.mapper.DefaultMapper", new MIDPLogger());
	}
	*/
	
	/**
	 * Creates a new DataGatherer object. Performs device discovery and gives the ReaderFactory information about the available
	 * devices. The ReaderFactory is then responsible for creating the appropriate readers. 
	 * 
	 * @param mapperClassName class name for the device mapper
	 * @param discoveryOn toggles device discovery on or off 
	 * @param logger object used to log messages
	 * @throws BluetoothStateException 
	 */
	public DataGatherer(String mapperClassName, boolean discoveryOn, Logger logger, int waitTime) {
		
		// Set the logging device
		this.logger = logger;
		
		// Initialize the used bluetooth stack. For instructions on how to 
		// do this, please refer to the documentation / examples of the 
		// used stack / library.
		
		
			try {
				initializeStack();
			} catch (Exception e) {
				if (logger != null)
					logger.logMessage(e.getMessage());
			}
		
		
		// Create the bluetooth listener which is responsible for the discovery and the processing needed in the discovery.
		// Takes as an input parameter a reference to the local device which is the high level manager of the bluetooth stack  
		
			BTDeviceListener btlistener = null;
			try {
				this.ldev = LocalDevice.getLocalDevice();
				btlistener = new BTDeviceListener(ldev,logger);
			} catch (BluetoothStateException e) {
				if (logger != null)
					logger.logMessage("[DG] Cannot retrieve LocalDevice from the bluetooth stack: " + e.getMessage());
			}
			
			this.mapper = null;
			try {
				mapper = (Mapper)(Class.forName(mapperClassName)).newInstance();
			} catch (InstantiationException e) {
				if (logger != null)
					logger.logMessage(e.getMessage());
			} catch (IllegalAccessException e) {
				if (logger != null)
					logger.logMessage(e.getMessage());
			} catch (ClassNotFoundException e) {
				if (logger != null)
					logger.logMessage(e.getMessage());
			}
			
			//Create a new BTDeviceConnector object, which handles device
			//and service discoveries.
			connector = new BTDeviceConnector(btlistener, mapper, logger, 60000, this, discoveryOn);				
	}


	/**
	 * Creates a new DataGatherer object. Performs device discovery and gives the ReaderFactory information about the available
	 * devices. The ReaderFactory is then responsible for creating the appropriate readers. 
	 * 
	 * @param mapper mapper object
	 * @param discoveryOn toggles device discovery on or off 
	 * @param logger object used to log messages
	 * @throws BluetoothStateException 
	 */
	public DataGatherer(Mapper mapper, boolean discoveryOn, Logger logger) {
		
		// Set the logging device
		this.logger = logger;
		
		// Initialize the used bluetooth stack. For instructions on how to 
		// do this, please refer to the documentation / examples of the 
		// used stack / library.
		
		
			try {
				initializeStack();
			} catch (Exception e) {
				if (logger != null)
					logger.logMessage(e.getMessage());
			}
		
		
		// Create the bluetooth listener which is responsible for the discovery and the processing needed in the discovery.
		// Takes as an input parameter a reference to the local device which is the high level manager of the bluetooth stack  
		
			BTDeviceListener btlistener = null;
			try {
				this.ldev = LocalDevice.getLocalDevice();
				btlistener = new BTDeviceListener(ldev,logger);
			} catch (BluetoothStateException e) {
				if (logger != null)
					logger.logMessage("[DG] Cannot retrieve LocalDevice from the bluetooth stack: " + e.getMessage());
			}
			
			this.mapper = mapper;
			
			connector = new BTDeviceConnector(btlistener, mapper, logger, 60000, this, discoveryOn);		
				
	}
	
	/**
	 * Starts a BTDeviceConnector thread. This method should be called before
	 * any data consumers are started. A Thread of BTDeviceConnector handles
	 * the device and service discoveries and reader creation. Calling this
	 * method starts the device discovery.
	 *
	 */
	
	
	public void start() {
		Thread t = null;
		t = new Thread(this.connector);
		t.start();
		
	}
	
	/**
	 * 	This method closes all of the readers.
	 *
	 */
	public void closeAll() {
		this.connector.close();
		if(logger!=null)
			logger.logMessage("[DG] Application finished.");
	}
	
	
	/**
	 * This method is used to initialize the used bluetooth stack. 
	 * Nowadays most stacks handle the initialization automatically, but 
	 * because some stacks still need initialization, this workaround is needed.  	 
	 */
	protected void initializeStack() throws Exception{}
	
	/**
	 * Returns the BTDeviceConnector.
	 * @return the BTDeviceConnector object
	 */
	
	public BTDeviceConnector getConnector(){
		return this.connector;
	}
	
	/**
	 * Returns the Mapper object.
	 * @return the mapper object
	 */

	public Mapper getMapper(){
		return this.mapper;
	}
	
	/**
	 * Returns the bluetooth address of the device.
	 * @return the bluetooth address
	 */
	
	public String getBluetoothAddress(){
		if (this.ldev != null){
			return this.ldev.getBluetoothAddress();
		}
		return null;
	}
	
	/**
	 * Returns the friendly name of the device.
	 * @return the friendly name
	 */
	
	public String getFriendlyName(){
		if (this.ldev != null){
			return ldev.getFriendlyName();
		}
		return null;
	}
}
