/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.mapper;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This is a Mapper implementation for Bluetooth devices.
 * @author Joonas Kukkonen and Eemil Lagerspetz
 *
 */

public class BTMapper implements Mapper {

	private Recorder rec;
	private Hashtable friendly;
	private Hashtable btAddressTable;
	private boolean friendlyNameUsed = false;

	/**
	 * Constructs a new BTMapper object. Sensor identification rules are retrieved from the Recorder object.
	 * @param rec The Recorder object use to save and retrieve sensor rules.
	 */
	
	public BTMapper(Recorder rec){
		this.rec = rec;
		if (rec != null) this.friendly = this.rec.retrieve(Recorder.BT_FRIENDLY_NAME);
		if (rec != null) this.btAddressTable = this.rec.retrieve(Recorder.BT_ADDRESS);
	}
	
	/**
	 * Returns the identifier number for a bluetooth address or friendly name. Mapper.UNKNOWN_DEVICE is returned if
	 * the bluetooth address or friendly name can not me mapped to an identification number.
	 * @param btAddress the bluetooth address of the device
	 * @param friendlyName the user friendly name of the bluetooth device
	 * @return the sensor identifier number
	 */
	
	public int getReaderIdentifier(String btAddress, String friendlyName) {
		int id = -1;		
		if (btAddress != null && this.btAddressTable != null){
			Object o = this.btAddressTable.get(btAddress.toLowerCase());
			if(o != null){
				id = ((Integer)o).intValue();
				return id;
			}
			
		}
		
		if (friendlyName != null && this.friendly != null){
			friendlyName = friendlyName.toLowerCase();
			String str = null, str2 = null;
			Enumeration enumeration = friendly.keys();
			while(enumeration.hasMoreElements()){
				str = (String)enumeration.nextElement();
				str2 = str.toLowerCase();
				
				if(str2.indexOf(friendlyName) != -1)
					return ((Integer)friendly.get(str)).intValue();
			}
		}
		return Properties.UNKNOWN_DEVICE;
	}
	
	/**
	 * Updates the records in the mapper. Call this method if you delete or add a new record.
	 */
	
	public void updateRecords(){
		if (rec != null){
			this.friendly = this.rec.retrieve(Recorder.BT_FRIENDLY_NAME);
			if(this.friendly.isEmpty())
				this.friendlyNameUsed = false;
			else
				this.friendlyNameUsed = true;
			this.btAddressTable = this.rec.retrieve(Recorder.BT_ADDRESS);
		}
	}
	
	/**
	 * Calls Recorder's deleteRecord method.
	 * @return true if deletion was succesfull, false otherwise
	 */
	
	public boolean deleteRecord(int deviceID, String address){
		if (rec != null){
			boolean result = this.rec.deleteRecord(deviceID, address);
			updateRecords();
			return result;
		}
		return false;
	}

	/**
	 * Returns the Recorder.
	 * @return the Recorder object.
	 */
	
	public Recorder getRecorder() {
		return this.rec;
	}
	
	/**
	 * Saves a rule record using the Recorder.
	 * @param id the format the rule is to be saved in
	 * @param sensorType the sensor identification number
	 * @param sensorIdentifier the friendly name or bluetooth address
	 * @return true if saving the record was succesful, false otherwise
	 */
	
	public boolean saveRecord(int id, int sensorType, String sensorIdentifier) {
		if (rec != null){
			boolean result = this.rec.saveRecord(id, sensorType, sensorIdentifier);
			updateRecords();
			return result;
		}
		return false;
	}

	public boolean friendlyNameUsed() {
		return this.friendlyNameUsed;
	}
	
	
}
