/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.mapper;

import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.rms.RecordEnumeration;
import javax.microedition.rms.RecordStore;
import javax.microedition.rms.RecordStoreException;
import javax.microedition.rms.RecordStoreFullException;
import javax.microedition.rms.RecordStoreNotFoundException;
import javax.microedition.rms.RecordStoreNotOpenException;
import hiit.logger.Logger;

/**
 * MIDPRecorder is a Recorder implementation for MIDP programs. It uses RecordStore
 * to save the mapper rules to, for example, phone's memory. 
 * @author Joonas Kukkonen and Eemil Lagerspetz
 *
 */


public class MIDPRecorder implements Recorder {
	
	private RecordStore store = null;
	private Logger logger;
	
	private final String recordStoreName = "Sensor records";
	
	/**
	 * Creates a new MIDPRecorder.
	 * @param logger The logger use to log error messages.
	 */
	
	public MIDPRecorder(Logger logger){
		
		this.logger = logger;
		
		try {
			this.store = RecordStore.openRecordStore(this.recordStoreName, true);
		} catch (RecordStoreFullException e) {
			//Throw some exception here...
		} catch (RecordStoreNotFoundException e) {
			if (this.logger != null) this.logger.logMessage("RecordStore not found!");
			//Shouldn't happen. A new Record store is created if there is none.
		} catch (RecordStoreException e) {
			if (this.logger != null) this.logger.logMessage("RecordStore exception:");
			if (this.logger != null) this.logger.logMessage(e.toString());
			
		}

	}

	/**
	 * Saves a record in the format specified by the id number. See Recorder interface for the id numbers.
	 * @param id The identification number of the saving format.
	 * @param sensorType The sensor type (an integer from 0 to 65535)
	 * @param sensorIdentifier The sensor identifier used to mach the sensor to the sensor type 
	 * @return true if the saving of the record succeeded, false otherwise
	 */
	
	public boolean saveRecord(int id, int sensorType, String sensorIdentifier) {
		if(id < 0 || id > 255){
			if (this.logger != null) this.logger.logMessage("Invalid format ID given to the Recorder");
			return false;
		}
		if(sensorType < 0 || sensorType > 65535){
			if (this.logger != null) this.logger.logMessage("Invalid sensor ID given to the Recorder");
			return false;
		}
		
		String temp1 = "", temp2 = "";
		
		if(id > 15)
			temp1 = Integer.toHexString(id);
		else
			temp1 = "0" + Integer.toHexString(id);
		
		if(sensorType < 16)
			temp2 = "000" + Integer.toHexString(sensorType);
		else if(sensorType < 256)
			temp2 = "00" + Integer.toHexString(sensorType);
		else if(sensorType < 4096)
			temp2 = "0" + Integer.toHexString(sensorType);
		else 
			temp2 = Integer.toHexString(sensorType);
		
		sensorIdentifier = sensorIdentifier.toLowerCase();
		
		byte[] message = (temp1 + temp2 + sensorIdentifier).getBytes();
		
		try {
			store.addRecord(message, 0, message.length);
			return true;
		} catch (RecordStoreNotOpenException e) {
			if(logger != null) logger.logMessage("The record store has not been opened. Adding a new record failed.");
			return false;
		} catch (RecordStoreFullException e) {
			if(logger != null) logger.logMessage("The record store is full. Adding a new record failed.");
			return false;
		} catch (RecordStoreException e) {
			if(logger != null) logger.logMessage("The record store is broken. Adding a new record failed.");
			return false;
		}		
	}
	
	/**
	 * Closes the MIDPRecorder.
	 */
	
	public void close(){
		try {
			store.closeRecordStore();
		} catch (RecordStoreNotOpenException e) {
			//We are trying to close it and it couldn't be opened... Hmm...
		} catch (RecordStoreException e) {
			if (logger != null) logger.logMessage("Closing the record store failed.");
		}
	}
	
	/**
	 * Destroys the RecordStore the MIDPRecorder is using.
	 * @return true if RecordStore was destroyed, false otherwise
	 */
	
	public boolean destroy(){
		try {
			this.store.closeRecordStore();
		} catch (RecordStoreNotOpenException e) {
			if(logger != null) logger.logMessage("The record store has not been opened. Destroying the RecordStore failed.");
			return false;
		} catch (RecordStoreException e) {
			if(logger != null) logger.logMessage("The record store is broken. Destroying the RecordStore failed.");
			return false;
		}
		
		return true;
	}

	/**
	 * This method returns all the device identifier - address / name pairs which are in the "id" format. For example
	 * if id is 1, the method returns all bluetooth sensors which are defined by bluetooth friendly name. Check the Recorder
	 * interface for possible id numbers. The maximum number id can have is 255 and the minimum 0. If the id is 0, all devices  identifier -
	 * address / name pairs in the repository are returned. 
	 * @param id the identification number of the format of which the rules are saved in
	 * @return the found rules in a Hastable object
	 */	
	
	public Hashtable retrieve(int id) {
		
		RecordEnumeration recEn = null;
		Hashtable table = new Hashtable();
		try {
			recEn = store.enumerateRecords(null, new MIDPRecordComparator(), false);
		} catch (RecordStoreNotOpenException e) {
			if (logger != null) logger.logMessage("Record store could not be opened when attempting to retrieve records.");
			return null;
		}
		
		String tempID, message;
		int deviceID;
		
		if(id < 0 || id > 255){
			if (this.logger != null) this.logger.logMessage("Invalid format ID given to the record store.");
			return null;
		}
		
		if(id > 15)
			tempID = Integer.toHexString(id);
		else
			tempID = "0" + Integer.toHexString(id);
		
		byte[] idBytes = new byte[2];
		idBytes[0] = (byte) tempID.charAt(0);
		idBytes[1] = (byte) tempID.charAt(1);
		
		byte[] record = null;
		boolean found = false;
		
		if (id > 0){
			while(recEn.hasNextElement()){
				//Keep looking until you find the right format id
				try {
					record = recEn.nextRecord();
				} catch (RecordStoreNotOpenException e) {
					if(logger != null) logger.logMessage("The record store has not been opened. Retrieving records failed.");
					return null;
				} catch (RecordStoreFullException e) {
					if(logger != null) logger.logMessage("The record store is full. Retrieving records failed.");
					return null;
				} catch (RecordStoreException e) {
					if(logger != null) logger.logMessage("The record store is broken. Retrieving records failed.");
					return null;
				}
				
				//Error check maybe?
				if (record[0] == idBytes[0] && record[1] == idBytes[1]){
					found = true;
					break;
				}
				
			}
		}else{
			if(recEn.hasNextElement()){
				try {
					record = recEn.nextRecord();
				} catch (RecordStoreNotOpenException e) {
					if(logger != null) logger.logMessage("The record store has not been opened. Adding a new record failed.");
					return null;
				} catch (RecordStoreFullException e) {
					if(logger != null) logger.logMessage("The record store is full. Adding a new record failed.");
					return null;
				} catch (RecordStoreException e) {
					if(logger != null) logger.logMessage("The record store is broken. Adding a new record failed.");
					return null;
				}
				found = true;
			}
		}
		
		while(found){
			//Handle the record
			if (record.length < 6){
				if (logger != null) logger.logMessage("Invalid record in the RecordStore.");
				continue;
			}
			
			String temp1;
			deviceID = 0;
			temp1 = new String(record, 2, 4);
			deviceID = Integer.parseInt(temp1, 16);
			
			message = new String(record, 6, record.length - 6);
			
			table.put(message.toLowerCase(), new Integer(deviceID));
			
			if(recEn.hasNextElement()){
				try {
					record = recEn.nextRecord();
				} catch (RecordStoreNotOpenException e) {
					if(logger != null) logger.logMessage("The record store has not been opened. Adding a new record failed.");
					return null;
				} catch (RecordStoreFullException e) {
					if(logger != null) logger.logMessage("The record store is full. Adding a new record failed.");
					return null;
				} catch (RecordStoreException e) {
					if(logger != null) logger.logMessage("The record store is broken. Adding a new record failed.");
					return null;
				}
				if (id > 0){
					//Check the ID
					if (!(record[0] == idBytes[0] && record[1] == idBytes[1])){
						found = false;
					}
				}
			}else
				found = false;
			
		}
		//if(logger != null) logger.logMessage("Retrieve completed with id: " + id);
		//if(logger != null) logger.logMessage(table.size() + " elements in the hastable.");
		return table;
	}
	

	
	/**
	 * Deletes the record with the sensorType and sensorIdentifier.
	 * @param sensorType the sensor type of the record to be deleted
	 * @param sensorIdentifier the sensor identifier of the record to be deleted
	 * @return true if the deletion of record succeeded, false otherwise
	 */
	
	public boolean deleteRecord(int sensorType, String sensorIdentifier) {
		if(sensorType < 0 || sensorType > 65535){
			if (this.logger != null) this.logger.logMessage("Invalid sensor ID given to the Recorder");
			return false;
		}
		
		String temp = "";
		
		if(sensorType < 16)
			temp = "000" + Integer.toHexString(sensorType);
		else if(sensorType < 256)
			temp = "00" + Integer.toHexString(sensorType);
		else if(sensorType < 4096)
			temp = "0" + Integer.toHexString(sensorType);
		else 
			temp = Integer.toHexString(sensorType);
		
		sensorIdentifier = sensorIdentifier.toLowerCase();
		
		byte[] message = ("00" + temp + sensorIdentifier).getBytes();
		
		/*
		 * Instead of a buggy implementation, deletion of records 
		 * is done in a much cruder way. Read all records to a
		 * Vector, delete the correct records from the vector, destroy
		 * the record store and save the old records in the new Store.
		 * This is time consuming and crude, but it works...
		 * 
		 * PS. The "right" way does not work...
		 */
		
		try {
			RecordEnumeration recEnum = this.store.enumerateRecords(null, null, false);
			Vector records = new Vector();
			outer: while (recEnum.hasNextElement()){
				byte[] record = recEnum.nextRecord();
				for (int i = 2; i < record.length && i < message.length; i++){
					if (message[i] != record[i]){
						records.addElement(record);
						continue outer;
					}
				}
			}
			
			if (records.size() == recEnum.numRecords())
				return false;
			
			//if (logger != null) logger.logMessage("Closing and reopening the record store.");
			this.store.closeRecordStore();
			RecordStore.deleteRecordStore(this.recordStoreName);
			this.store = RecordStore.openRecordStore(this.recordStoreName, true);
			
			for (int i = 0; i < records.size(); i++){
				byte[] save = (byte[])records.elementAt(i);
				this.store.addRecord(save, 0, save.length);
			}
			
			
		} catch (RecordStoreNotOpenException e) {
			if(logger != null) logger.logMessage("The record store has not been opened. Deleting a record failed.");
			return false;
		} catch (RecordStoreFullException e) {
			if(logger != null) logger.logMessage("The record store is full. Deleting a record failed.");
			return false;
		} catch (RecordStoreException e) {
			if(logger != null) logger.logMessage("The record store is broken. Deleting a record failed.");
			return false;
		}		
		
		return true;
	}

}
