/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
package hiit.mapper;

import java.util.Hashtable;

/**
 * Reader ID numbers are stored in the Properties class. It also offers
 * methods for retreiving (description -> ID) and (ID -> class name) mappings.
 * @author Joonas Kukkonen
 *
 */

public class Properties {
    public static final int major = 1;
    public static final int minor = 3;
    public static final char flavor = 'a';
    
    // Data type definitions for Readers:
    public static final int DATATYPE_STRING = 0;
    public static final int DATATYPE_INT = 1;
    public static final int DATATYPE_DOUBLE = 2;
    
    // Blackboard listening port on local device
    public static final int LISTEN_PORT = 10004;
    
	// Unknown device constant
	public final static int UNKNOWN_DEVICE = 0x0000;
	// Low level parsers
	public final static int SERIAL_PORT_READER = 0x0001;
	public final static int BT_PORT_READER = 0x0002;
	
	// Local device parser variants
	public final static int LOCAL_READER = 0x0003;
	public final static int JNI_READER = 0x0004;
	public final static int GSM_READER = 0x0005;
	public final static int MUSIC_LOCAL_READER = 0x0006;
    public final static int LOCAL_GSM_READER = 0x0007;
    
    //Does not implement a local device reader class, but works in similar maner
    
    public final static int CALL_READER = 0x0008;
    
    //Bogus stuff
    public final static int COUNTER_READER = 0x0300;
	
	// Remote device parsers
	public final static int GPS_READER = 0x0100;
	public final static int ACCELEROMETER_READER = 0x0103;
	public final static int ECG_ACC_READER = 0x0104;
	public final static int ECG_ACC_BYTE_READER = 0x0105;
	public final static int ICUBEX_READER = 0x0106;
	public final static int NOKIA_ACC_READER = 0x0107;
	
    public final static int BARCODE_CLIENT = 0x201;
    public final static int BARCODE_SERVER = 0x202;
    public final static int GSM_SERVER = 0x204;
    public final static int CALL_SERVER = 0x205;
	
	private static Hashtable nameId = new Hashtable();
	private static Hashtable idClassName = new Hashtable();
	private static Hashtable idMode = new Hashtable();
	private static Hashtable serverReaderMap = new Hashtable();
	
	static{
		nameId.put("GPS device", new Integer(GPS_READER));
		nameId.put("Accelerometer", new Integer(ACCELEROMETER_READER));
		nameId.put("ECG - Accelerometer", new Integer(ECG_ACC_READER));
		nameId.put("ECG/Acc with HR", new Integer(ECG_ACC_BYTE_READER));
		nameId.put("I-CubeX system", new Integer(ICUBEX_READER));
		nameId.put("Nokia Accelerometer", new Integer(NOKIA_ACC_READER));
		
		idClassName.put(new Integer(UNKNOWN_DEVICE), "");
		idClassName.put(new Integer(SERIAL_PORT_READER), "");
		idClassName.put(new Integer(BT_PORT_READER), "");
		
		idClassName.put(new Integer(LOCAL_READER), "hiit.reader.LocalDeviceReader");
		idClassName.put(new Integer(JNI_READER), "hiit.reader.JNIServerReader");
		idClassName.put(new Integer(GSM_READER), "hiit.reader.GSMReader");
		
		idClassName.put(new Integer(GPS_READER), "hiit.reader.GPSReader");
		idClassName.put(new Integer(ACCELEROMETER_READER), "");
		idClassName.put(new Integer(ECG_ACC_READER), "hiit.reader.ECGReader");
		idClassName.put(new Integer(ECG_ACC_BYTE_READER), "hiit.reader.ECGByteReader");
		idClassName.put(new Integer(ICUBEX_READER), "hiit.reader.ICubexWiMiniDigParser");
		idClassName.put(new Integer(MUSIC_LOCAL_READER), "hiit.reader.MusicLocalDeviceReader");
		idClassName.put(new Integer(LOCAL_GSM_READER), "hiit.reader.LocalGSMReader");
		idClassName.put(new Integer(CALL_READER), "hiit.reader.CallReader");
		idClassName.put(new Integer(NOKIA_ACC_READER), "hiit.reader.NokiaAccReader");
		idClassName.put(new Integer(COUNTER_READER), "hiit.reader.CounterReader");
		
		idMode.put(new Integer(UNKNOWN_DEVICE), new Boolean(true));
		idMode.put(new Integer(SERIAL_PORT_READER), new Boolean(true));
		idMode.put(new Integer(BT_PORT_READER), new Boolean(true));
		
		idMode.put(new Integer(LOCAL_READER), new Boolean(true));
		idMode.put(new Integer(JNI_READER), new Boolean(true));
		idMode.put(new Integer(GSM_READER), new Boolean(true));
		
		idMode.put(new Integer(GPS_READER), new Boolean(false)); // true for streaming GPS, false for request mode
		idMode.put(new Integer(ACCELEROMETER_READER), new Boolean(true));
		idMode.put(new Integer(ECG_ACC_READER), new Boolean(true));
		idMode.put(new Integer(ECG_ACC_BYTE_READER), new Boolean(true));
		idMode.put(new Integer(ICUBEX_READER), new Boolean(true));
		idMode.put(new Integer(MUSIC_LOCAL_READER), new Boolean(true));
        idMode.put(new Integer(LOCAL_GSM_READER), new Boolean(true));
        idMode.put(new Integer(CALL_READER), new Boolean(true));
		idMode.put(new Integer(NOKIA_ACC_READER), new Boolean(true));
		idMode.put(new Integer(COUNTER_READER), new Boolean(true));

		//Uncomment this when call gsm reader has been separated from the local device reader...
		//serverReaderMap.put(new Integer(GSM_SERVER), new Integer(LOCAL_GSM_READER));
		serverReaderMap.put(new Integer(CALL_SERVER), new Integer(CALL_READER));
	}
	
	/**
	 * Returns a Hashtable that maps the descriptions of Readers to ID numbers.
	 * @return a Hashtable that maps the descriptions of Readers to ID numbers
	 */
	
	public static Hashtable getNameId(){
		return nameId;
	}
	
	/**
	 * Returns a Hashtable that maps the ID numbers to class names of Readers.
	 * @return a Hashtable that maps the ID numbers to class names of Readers
	 */
	
	public static Hashtable getIdClassName(){
		return idClassName;
	}
	
	/**
	 * Returns a Hashtable that maps the ID numbers to default mode of Readers.
	 * @return a Hashtable that maps the ID numbers to default mode of Readers.
	 */
	
	public static Hashtable getIdMode(){
		return idMode;
	}
	
	/**
	 * Maps a Blackboard Customer Type identifier to its descriptive name.
	 * @param id The Blackboard Customer Type.
	 * @return A String with a descriptive name for the specified customer type.
	 */
	public static String getNameForCustomerType(int id){
	    switch (id){
	    case BARCODE_CLIENT: return "Barcode Client";
	    case BARCODE_SERVER: return "Barcode Server";
	    case GSM_SERVER: return "GSM Server";
	    case CALL_SERVER: return "Call Server";
	    default: return null;
	    }
	}
	
	/**
	 * Returns a mapping of server id numbers to reader id numbers.
	 * @return a mapping of server id numbers to reader id numbers.
	 */
	public static Hashtable getServerMapping(){
	    return serverReaderMap;
	}
}
