/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.midpgui;

import javax.microedition.lcdui.Canvas;
import javax.microedition.lcdui.Font;
import javax.microedition.lcdui.Graphics;
import javax.microedition.lcdui.Image;

/**
 * A customized UI component. Motivated by the Tabs in Nokia phones, the TabCanvas 
 * offers a three level GUI containing a Header, TabPanel and a container for 
 * visualizing application information. 
 * @author Petteri Nurmi
 */
public class TabCanvas extends Canvas implements Runnable {

	/**
	 * The main application class is kept running constantly. This flag 
	 * controls that behaviour.
	 */
	private boolean isRunning = true;
	
	/**
	 * The width of the UI component.
	 */
	private int width = -1;
	
	/**
	 * The height of the UI component.
	 */
	private int height = 1;
	
	/**
	 * A UI component that prints a header on top of the screen. 
	 * The header is used for displaying relevant application info, 
	 * logos etc.  
	 */
	private TabHeader _tabHeader = null;
	
	/**
	 * A container for tabs. The tabs are the core components, which control 
	 * what is seen in the main text area.
	 */
	private TabPane _tabPanel = null;
	
	/**
	 * Backbuffer image for double buffering purposes. Used to remove flickering. 
	 */
	private Image _backbufferImage = null;
	
	/**
	 * Container for the objects that are responsible for generating the contents 
	 * of the main text area.
	 */
	private InformationContainer _container = null;
	
	/**
	 * Flag indicating that upward scrolling has taken place. 
	 */
	static final int SCROLL_UP = 1;
	
	/**
	 * Flag indicating that downward scrolling has taken place.
	 */
	static final int SCROLL_DOWN = -1;
	
	/**
	 * Flag indicating that the tabs have moved to the left.
	 */
	static final int TAB_LEFT = -1;
	
	/**
	 * Flag indicating that the tabs have moved to the right.
	 */
	static final int TAB_RIGHT = 1;

	/**
	 * Indicates if the canvas is in full screen mode
	 */
	
	private boolean fullScreen = false;
	
	/**
	 * The redraw frequency, how many milliseconds to wait between subsequent redraws of the GUI.
	 */
	
	private int redrawFrequency = 500;
	
	
	/**
	 * Constructor for the TabCanvas. Initialized UI components and starts the main 
	 * thread.
	 * @param _title The title of the application / the text that is used in the 
	 * header.
	 */

	
	public TabCanvas(String _title) {
		// Create the UI components
		this.setFullScreenMode(true);
		this.fullScreen = true;
		this._tabHeader = new TabHeader(_title);
		this._tabPanel = new TabPane();
		this._container = new InformationContainer(_tabPanel);
		// Runme
		Thread t = new Thread(this);
		t.start();
	}

	/**
	 * Paints the UI
	 * @param arg0 A graphics object on which the drawing takes place.
	 */
	protected void paint(Graphics arg0) {		
		// Get width and height
		int w = getWidth();
		int h = getHeight();
		
		// If the values are different than our previously stored values, 
		// update storage
		if(w!=width)width = w;
		if(h!=height)height = h;
		
		// If nothing on which to draw, quit
		if(width==-1 || height==-1)return;
		
		//arg0.setFont(Font.getFont(arg0.getFont().getFace(),arg0.getFont().getStyle(),Font.SIZE_LARGE));
		
		// Calculate the appropriate regions of the screen that are used for drawing 
		// the UI components.
		int h10 = (int)(height/10);
		if (h10 < arg0.getFont().getSize())
			h10 = arg0.getFont().getSize();
		
		//This was height*0.15
		int h15 = (int)(height / 7);
		if (h15 < (int)((Font.SIZE_SMALL +5) + (Font.SIZE_SMALL +5)/10))
			h15 = (int)((Font.SIZE_SMALL +5) + (Font.SIZE_SMALL +5)/10);
		
		// Create offscreen buffer for drawing
		if(this._backbufferImage==null)
			this._backbufferImage = Image.createImage(width,height);
		
		// Get the backbuffer
		Graphics _backBuffer = this._backbufferImage.getGraphics();
				
		// Reserve 10% of screen space for header
		_tabHeader.draw(_backBuffer,1,1,width,h10);
				
		// Reserve 15% of screen space for tabs
		_tabPanel.draw(_backBuffer,1,h10,width, h15);
		
		// Use the rest for drawing the container
		_container.draw(_backBuffer,0,h10+h15,width,height - (h10+h15));
		
		// Paint the buffer to screen
		arg0.drawImage(this._backbufferImage,0,0,Graphics.TOP|Graphics.LEFT);
	}
	
	/**
	 * Closes the application
	 */
	public void close(){
		this.isRunning = false;
	}

	/**
	 * The program logic of the application. Simply calls repaint to ensure that 
	 * screen is updated regularly enough.
	 */
	public void run() {
		while (isRunning) {
			synchronized (this) {
				try {
					wait(redrawFrequency);
				} catch (InterruptedException e) {

				}
				repaint();				
			}
		}
	}
	
	/**
	 * Adds an information object to the canvas. The information objects are 
	 * components that are responsible for generating contents for different tabs. 
	 * @param _io The information object to add.
	 */
	public void addInformationObject(InformationObject _io) {
		this._container.addInformationObject(_io);
	}
	
	
	public void removeInformationObject(InformationObject _io){
		this._container.removeInformationObject(_io);
	}
	
	
	/**
	 * Eventhandler for keypresses. 
	 * @param keyCode The code of the key that was pressed.
	 */
	protected void keyPressed(int keyCode) {
		
		/**
		 * At least the emulator gives keyCodes that are not specified in the API.
		 * Here we map them to proper ones.
		 */
		if(keyCode == -1)keyCode = 1;
		if(keyCode == -2)keyCode = 6;
		if(keyCode == -3)keyCode = 2;
		if(keyCode == -4)keyCode = 5;
		
		/**
		 * Select correct event.
		 */
		switch(keyCode) {
			case Canvas.UP:
			case Canvas.KEY_NUM2:
				scroll(TabCanvas.SCROLL_UP);
				this.repaint();
				break;
			case Canvas.DOWN:
			case Canvas.KEY_NUM8:
				scroll(TabCanvas.SCROLL_DOWN);
				this.repaint();
				break;
			case Canvas.LEFT:
			case Canvas.KEY_NUM4:
				changeTab(TabCanvas.TAB_LEFT);
				this.repaint();
				break;
			case Canvas.RIGHT:
			case Canvas.KEY_NUM6:
				changeTab(TabCanvas.TAB_RIGHT);
				this.repaint();
				break;
			case Canvas.KEY_STAR:
				if (this.fullScreen == false){
					this.setFullScreenMode(true);
					this.fullScreen = true;
					this.repaint();
				}else{
					this.setFullScreenMode(false);
					this.fullScreen = false;
					this.repaint();
				}
				break;
		}
	}
	
	/**
	 * EventHandler for scroll-related events.
	 * @param code The Event code for scrolling.
	 */
	private void scroll(int code) {
		switch(code) {
			case TabCanvas.SCROLL_UP:
				this._container.scrollUp();
				break;
			case TabCanvas.SCROLL_DOWN:
				this._container.scrollDown();
				break;
		}
	}
	
	/**
	 * EventHandler for tab switch related events.
	 * @param code The event code for tab switching
	 */
	private void changeTab(int code) {
		switch(code) {
			case TabCanvas.TAB_LEFT:
				this._tabPanel.tabLeft();
				break;
			case TabCanvas.TAB_RIGHT:
				this._tabPanel.tabRight();
				break;
		}
	}
	
	public void setRedrawFrequency(int redrawFrequency){
		this.redrawFrequency = redrawFrequency;
	}
	
	public int getRedrawFrequency(){
		return this.redrawFrequency;
	}
}