package hiit.plugin.activity.filter;

import hiit.plugin.activity.data.Instance;
import hiit.plugin.activity.utils.MathUtils;

public class MedianFilter implements Filter {
	
	private int _windowSize = 3;
	
	public MedianFilter() {
		setWindowSize(3);
	}
	
	public MedianFilter(int _windowSize) {
		setWindowSize(_windowSize);
	}
	
	public void setWindowSize(int _windowSize) {
		if(_windowSize<=0)return;
		this._windowSize = _windowSize;
	}

	public Instance[] applyFilter(Instance[] _signals) {
		
		// Check that the signals are valid
		if(_signals == null)return null;
		
		// Ensure we have actually data points
		final int _n = _signals.length;
		if(_n <= 0)return null;
		
		// And that the data has dimensionality greater than zero
		final int _d = _signals[0].getDimensionality();
		if(_d<=0)return null;
		
		// Create storage for results
		Instance[] _newInstances = new Instance[_signals.length];
		
		// Temporary storage for values
		double[][] _valueMatrix = new double[_n][_d];
		
		// Check whether the window size is even
		boolean _even = (_windowSize % 2 == 0 ? true : false);
		
		// Determine the appropriate shifts
		final int _lowshift = (_even == true ? _windowSize / 2 - 1: _windowSize / 2);
		final int _highshift = _windowSize / 2;
		
		// Loop through the data points
		for(int _i = 0;_i<_n;_i++) {
			
			// Resolve the lower and upper indices, i.e., the points that are considered for the median  
			int _lowerIndex = Math.max(0, _i - _lowshift);			
			int _upperIndex = Math.min(_n - 1, _i + _highshift);
			
			// Insert the current value into temporary storage
			_valueMatrix[_i] = _signals[_i].getValues();
			
			// Count maintains the number of datapoints that are considered for taking the median
			int _count = 1 + (_i - _lowerIndex);
			
			// Ensure that we have the other values in temporary storage
			for(int _q=(_i+1);_q<=_upperIndex;_q++) {
				_valueMatrix[_q] = _signals[_q].getValues();
				_count++;
			}
			
			// Create temporary storage for the window that is used to compute the median
			double[][] _medianWindow = new double[_count][_d];
						
			// Create a copy of the values, otherwise sorting will screw up the order of datapoints
			for(int _q = 0;_q<_count;_q++) {				
				System.arraycopy(_valueMatrix[_lowerIndex + _q],0,_medianWindow[_q],0,_d);				
			}			
			
			// Sort the data
			_medianWindow = MathUtils._HEAPSORT(_medianWindow);
			
			// Temporary variable that is used to return results
			double[] _tmpMatrix = new double[_d];
			
			// Perform filtering, i.e., replace the current point with the median. Note that
			// we loop over dimensions, not datapoints.
			for(int _q=0;_q<_d;_q++) {
				
				if(_medianWindow.length % 2 == 0) {
					
					final int _lowPoint = (int)((_medianWindow.length + 1)/2 - 0.5);
					final int _highPoint = (int)((_medianWindow.length + 1)/2 + 0.5);
					_tmpMatrix[_q] = (_medianWindow[_lowPoint][_q] +  _medianWindow[_highPoint][_q])/ 2.0;  
						
				}else {
					_tmpMatrix[_q] = _medianWindow[(_medianWindow.length + 1) / 2 - 1][_q];
				}
				
			}
			
			// Create filtered results
			_newInstances[_i] = new Instance(_tmpMatrix);
						
		}
				
		// Return results
		return _newInstances;
		
	}

}
