/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.plugin.sampler;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Vector;

import javax.microedition.lcdui.Choice;
import javax.microedition.lcdui.ChoiceGroup;
import javax.microedition.lcdui.Command;
import javax.microedition.lcdui.Form;
import javax.microedition.lcdui.TextField;

public class QuestionSet {
	
	private int index = -1;
	private int numberOfQuestions = 0;
	//All questions have a text
	private Vector questionTexts = new Vector();
	//All questions have a type
	private Vector questionTypes = new Vector();
	//Only multiple choice questions have different choices
	private Vector questionChoices = new Vector();
	//Question numbers, just in case...
	private Vector questionNumbers = new Vector();
	
	private ExperienceSampler es = null;
	private int setId = 0;
	
	private Vector questionElements = new Vector();
	private Form questionForm = null;
	
	
	public QuestionSet(ExperienceSampler es, int setId){
		this.es = es;
		this.setId = setId;
	}

	public void addOpenQuestion(String text, int number){
		this.questionNumbers.addElement(new Integer(number));
		this.questionTexts.addElement(text);
		this.questionTypes.addElement(new Integer(1));
		this.questionChoices.addElement(null);
		this.numberOfQuestions++;
	}
	
	public void addChoiceQuestion(String text, int number, String[] choices, boolean exclusive){
		this.questionNumbers.addElement(new Integer(number));
		this.questionTexts.addElement(text);
		if(exclusive)
			this.questionTypes.addElement(new Integer(2));
		else
			this.questionTypes.addElement(new Integer(3));
		this.questionChoices.addElement(choices);
		this.numberOfQuestions++;
	}
	
	public Form getQuestionForm(){
		if(this.questionForm != null)
			return this.questionForm;
		Form questions = new Form("Questions("+this.setId+")");
		for(int i=0; i<this.numberOfQuestions; i++){
			int type = ((Integer)(questionTypes.elementAt(i))).intValue();
			String questionText  = (String)this.questionTexts.elementAt(i);
			if(type == 1){
				//Open question
				questions.append(i + ". " + questionText + "\n");
				TextField field = new TextField(null, "", 256, TextField.ANY);
				questions.append(field);
				questions.append("\n");
				this.questionElements.addElement(field);
			}else if(type == 2 || type == 3){
				//Multiple choice question
				String[] choices = (String[])this.questionChoices.elementAt(i);
				questions.append(i + ". " + questionText + "\n");
				ChoiceGroup group = null;
				if(type == 2)
					group = new ChoiceGroup(null, Choice.EXCLUSIVE);
				else
					group = new ChoiceGroup(null, Choice.MULTIPLE);

				int numberOfChoices = choices.length;					
				for(int j=0; j<numberOfChoices; j++){
					group.append(choices[j], null);
				}
				questions.append(group);
				questions.append("\n");
				this.questionElements.addElement(group);
			}
		}
		questions.addCommand(new Command("Back", Command.BACK, 1));
		questions.addCommand(new Command("Send", Command.ITEM, 1));
		questions.setCommandListener(this.es);
		this.questionForm = questions;
		return questions;
	}
	
	public String getAnswerMessage(){
		/*Answer message:
		 *
		 *---------------------------------------------------------------------------
		 *| CMD | LENGTH | DEVICE ID | QUESTION SET ID | NUMBER OF QUESTIONS | XXXX |
		 *---------------------------------------------------------------------------
   	 	 *   1      3          2              1                   1              n
		 *
		 *	CMD = "A" or "a"   
		 *
		 *	Questions:
		 *
		 *	--------------------------------------------
		 *	| QUESTION NUMBER | ANSWER LENGTH | ANSWER |
		 *	--------------------------------------------
	     *   	 	1                  2           n
		 *
		 *	Answer format:
		 *		1) "yadayadayada"
		 *		2) "12"
		 *		3) "1;5;7;9"
		 */

		ByteArrayOutputStream bout = new ByteArrayOutputStream();
		int numberOfAnswers = 0;
		
		for(int i=0; i<this.numberOfQuestions; i++){
			String answer = null;
			int type = ((Integer)(questionTypes.elementAt(i))).intValue();
			if(type == 1){
				answer = ((TextField)this.questionElements.elementAt(i)).getString();
				if(answer == null || answer.length() == 0)
					continue;	
			}else if(type == 2){
				ChoiceGroup group = ((ChoiceGroup)this.questionElements.elementAt(i));
				if(group != null)
					answer = "" + (group.getSelectedIndex()+1);
				if(answer == null || answer.length() == 0)
					continue;
			}else if(type == 3){
				ChoiceGroup group = ((ChoiceGroup)this.questionElements.elementAt(i));
				boolean[] choices = new boolean[group.size()];
				group.getSelectedFlags(choices);
				answer = "";
				for(int j=0; j<choices.length; j++){
					if(choices[j])
						answer += (j+1)+";";
				}
				if(answer == null || answer.length() == 0)
					continue;
			}
			numberOfAnswers++;
			int number = ((Integer)this.questionNumbers.elementAt(i)).intValue();
			intToBytes(number,1,bout);
			intToBytes(answer.length(),2,bout);
			try {
				bout.write(answer.getBytes());
			} catch (IOException e) {
				// nothing to do
			}
		}
		byte[] data = bout.toByteArray();
		intToBytes(this.setId, 1, bout);
		intToBytes(numberOfAnswers, 1 ,bout);
		try {
			bout.write(data);
		} catch (IOException e) {
			// Nothing to do
		}		
		return bout.toString();
	}
	
	private void intToBytes(int number, int size, ByteArrayOutputStream out){
		byte[] array = new byte[size];
		
		for (int i = size-1; i >= 0; i--){
			array[i] = (byte)(number % 256);
			number = number / 256;
		}
		try{
			out.write(array);
		}catch (IOException e){
			// TODO
		}
	}
}
