/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.plugin.sampler;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import javax.microedition.io.Connector;
import javax.microedition.io.SocketConnection;

/**
 * This class is used by the DataTransmitter class to transmit sensor data
 * from BeTelGeuse to a server. The DataTransmitter uses the BeTelGeuse data
 * transfer protocol (see the manual).
 * @author Joonas Kukkonen, Eemil Lagerspetz
 *
 */

public class SamplerConnection {

	private String url = "";
	private SocketConnection conn = null;
	private InputStream in = null;
	private OutputStream out = null;
	
	
	/**
	 * Constructs a new DataTransmitterConnection
	 * @param url - the url of the server or object the DataTransmiterConnection connects to
	 * @throws ConnectionException if and error occured during the constuction of the object
	 */
	
	public SamplerConnection(String url){
		this.url = url;
	}
	
	/**
	 * Returns the address where data is to be sent.
	 * @return address of the server
	 */
	
	public String getAddress(){
		return url;
	}
	
	public synchronized int openConnection(int deviceID, int type, String userName) throws ConnectionException{
		ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
		
		/*
		 * Construct the connection initialization message
		 * 
		 * -------------------------------------------------------------------------------------------
		 * | CMD | LENGTH | DEVICE ID | TYPE OF THE DEVICE | LENGTH OF THE DESCRIPTION | DESCRIPTION |
		 * -------------------------------------------------------------------------------------------
		 *    1      3          2               1                         1                   n
		 *    
		 *    CMD = 'C'
		 *    LENGTH = Length of the message
		 *    DEVICE ID = Unique device identifier
		 *    TYPE OF THE DEVICE = Quess what...
		 *    LENGTH OF THE DESCRIPTION = Length of the DESCRIPTION field
		 *    DESCRIPTION = The description of the device
		 */
		
		try {
			byteOut.write("C".getBytes());
			intToBytes(4 + userName.getBytes().length, 3, byteOut);
			intToBytes(deviceID, 2, byteOut);
			intToBytes(type, 1, byteOut);
			intToBytes(userName.getBytes().length, 1, byteOut);
			byteOut.write(userName.getBytes());
		} catch (IOException e) {
			// Should not happen
		}
		
		try {
			this.conn = (SocketConnection)Connector.open("socket://" + url, Connector.READ_WRITE, true);
			this.in = this.conn.openInputStream();
			this.out = this.conn.openOutputStream();
		} catch (IOException e) {
			throw new ConnectionException("Failed to create a socket connection. Check the url.");
		}
				
		send(byteOut.toByteArray());
		
		/*
		 * Read the response
		 *     1             3                   2
		 *  --------------------------------------------
		 *  | Cmd | Length of the message | Device ID  |
		 *  --------------------------------------------
		 */
		
		
		int[] response = receive();
			
		return (response[4]*256 + response[5]);
	}
	
	/**
	 * Closes the connection.
	 * @throws ConnectionException if the connection did not close properly
	 */
	
	public synchronized void closeConnection() throws ConnectionException{
		
		try {
		    if (this.conn != null){
			this.conn.close();	
		    }
		    if (this.in != null){
			this.in.close();	
		    }
		    if (this.out != null){
			this.out.close();
		    }
		} catch (IOException e) {
			throw new ConnectionException("[DTC] Unable to close the connection.");
		}
	}
	
	/**
	 * Sends a byte array throug the connection.
	 * @param message the message to be sent
	 * @throws ConnectionException if an error occured during the sending of the message
	 */
	
	public synchronized void send(byte[] message) throws ConnectionException{
	    try {
		//out = conn.openOutputStream();
		out.write(message);
		out.flush();
		//out.close();
	    } catch (IOException e) {
		throw new ConnectionException("Couldn't send a message through a socket.");
	    }
	}
	
	/**
	 * Reads a message from the connection. The ammount read depends on the length
	 * field of the received message (2nd, 3rd and 4th byte).
	 * @return the received message
	 * @throws ConnectionException if an error occured during the reading
	 */
    
	public synchronized int[] receive() throws ConnectionException{
		/*
		 * PROBLEM! Java byte's range is -128 to 127. However unsigned bytes are
		 * used in the protocol. Thus we only use the read() method in InputStream
		 * that maps bytes to ints with the range of 0 to 255.
		 */
		try {
			int[] head = new int[4];
			//in = conn.openInputStream();
			int  length;
			
			//Read the command.
			head[0] = in.read();
			//Read the 3 byte length field
			head[1] = in.read();
			length = head[1] * 256 * 256;
			head[2] = in.read();
			length += head[2] * 256;
			head[3] = in.read();
			length += head[3];
			
			int [] msg = new int[length+4];
			
			for (int i = 4; i < length+4; i++){
				msg[i] = in.read();
			}
			
			msg[0] = head[0];
			msg[1] = head[1];
			msg[2] = head[2];
			msg[3] = head[3];
			//in.close();
			return msg;
		} catch (IOException e) {
			throw new ConnectionException("Could not read a message from the socket.");
		}
	}
	
	private void intToBytes(int number, int size, ByteArrayOutputStream out){
		byte[] array = new byte[size];
		
		for (int i = size-1; i >= 0; i--){
			array[i] = (byte)(number % 256);
			number = number / 256;
		}
		try{
			out.write(array);
		}catch (IOException e){
			// This should not happen
		}
	}

}
