/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.plugin.transmitter;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;

import javax.microedition.io.SocketConnection;
import javax.microedition.lcdui.Alert;

import hiit.BetelgeuseMIDP;
import hiit.blackboard.BlackboardFactory;
import hiit.blackboard.Blackboard;
import hiit.blackboard.Customer;
import hiit.blackboard.CustomerData;
import hiit.blackboard.CustomerInfo;
import hiit.blackboard.ProducerCache;
import hiit.blackboard.event.EventData;
import hiit.blackboard.event.GSMEvent;
import hiit.blackboard.filter.CombinatorFilter;
import hiit.blackboard.filter.ConstraintFilter;
import hiit.blackboard.filter.DelayFilter;
import hiit.blackboard.filter.DynamicCompareFilter;
import hiit.blackboard.filter.Filter;
import hiit.blackboard.filter.CompareFilter;
import hiit.blackboard.filter.GPSVelocityFilter;
import hiit.datagatherer.DataGatherer;
import hiit.logger.Logger;
import hiit.mapper.Properties;
import hiit.reader.GPSReader;
import hiit.reader.GSMReader;
import hiit.reader.LocalDeviceReader;
import hiit.reader.LocalGSMReader;

/**
 * Transmitter sends data from Readers to a server using a TransmitterConnection
 * object. The Transmitter comes with a GUI that implements the Plugin
 * interface. The GUI (TransmitterGUI class) is attached by the Transmitter to
 * Betelgeuse's GUI. After a Transmitter object is contructed, a thread needs to
 * be created from it. Should the Transmitter be switched off in the
 * constructor, it needs to be put online by calling switchOn().
 *
 * @author Joonas Kukkonen and Eemil Lagerspetz
 *
 */

public class BoardTransmitter implements Runnable, Customer, SomeBoardTransmitter {

    // Setting this to false will terminate the transmitter
    private boolean isRunning = true;

    // Is the transmitter paused or not?
    private boolean switchedOn = false;

    private boolean tryConnect = false;

    private boolean newaddress = false;

    private boolean addressChanged = false;

    private BoardTransmitterConnection conn = null;

    private Logger logger = null;

    private int maxSampleTimeout = 60000; // delay in sending to the server

    //private int distanceTimeout = 50;

    private Vector newProducers;

    private Vector allProducers;

    private Hashtable producerID;

    private int nextReaderId = 0;

    private Hashtable producerParameters;

    //private Customer GPSReader = null;

    // Stream blocks...
    private Vector streamData = new Vector();

    private Vector subscriptions = new Vector();

    // Session ID starts at 0. We only want to request new session ID
    // when we start the transmitter. Otherwise we try to retrain the old ID.
    private int sessionID = 0;

    // The compression used... default = 0 <-> no compression is used
    private int compression = 0;

    private DataGatherer dgat;

    private boolean init;

    private BoardTransmitterMIDPGUI gui;

    private BetelgeuseMIDP bet;

    private int transmissionNumber = 0;

    private Vector ldr = null;

    private Blackboard bb;

    private boolean mode;

    private Vector ignoredProducerTypes = new Vector();    // updates from these are handled with events
	private Vector eventsToSave = new Vector();            // events waiting to be saved
	private Customer gsmReader = null;
	private Customer gpsReader = null;
	private boolean filterInited = false;

	public long lastSendTime = -1;

	// producer-cache
	private ProducerCache producerCache = null;

	private int[] gpsReaderIncludeParamPositions = null;
	private String[] gpsReaderIncludeParamNames = null;
	private int[] gpsReaderIncludeParamTypes = null;


	private int[] gsmReaderIncludeParamPositions = null;
	private String[] gsmReaderIncludeParamNames = null;
	private int[] gsmReaderIncludeParamTypes = null;



	//

	private boolean closeConnectionBetweenSends;           // currently not in use

	// kbs
	private double lastSendKb = -1;
	private double lastSendKbs = -1;
	private double lastUploadTime = -1;

	private Timer pingWaitTimer = null;
	private final static int PING_DELAY_MINUTES = 15;


	private long lastGPSUpdate = -1;

	/*
	// adaptive send-interval
	private boolean useAdaptiveSendInterval = true;
	private double optimizeSendIntervalAmount = 1;        // 0-1
	private double bufferSizePerc = 2.0;                  // 1 = kbs
	private boolean adaptiveSendCalibrated = false;
	private int adaptiveSendCalibrateWaitPerc = 2;
	private double lastSendTime = -1;
	private double lastUploadTime = -1;
	private int currentBufferSize = 0;
	private double optimalBufferSize = -1;
	//
	*/


    /**
     * Creates a new DataTransmitter object.
     *
     * @param bet
     *            the current BeTelGeuseMIDP object
     * @param dgat
     *            the current DataGatherer object
     * @param url
     *            the URL of the server data is to be sent to
     * @param logger
     *            the logger object used to log error messages
     */

    public BoardTransmitter(BetelgeuseMIDP bet, DataGatherer dgat, String url,
            Logger logger, boolean closeConnectionBetweenSends) {
        this.init = true;
        this.bet = bet;
        this.newProducers = new Vector();
        this.allProducers = new Vector();
        this.producerID = new Hashtable();
        this.producerParameters = new Hashtable();
        this.dgat = dgat;
        this.logger = logger;
        this.closeConnectionBetweenSends = closeConnectionBetweenSends;

        this.mode = ((Boolean)(Properties.getIdMode().get(new Integer(Properties.GPS_READER)))).booleanValue();

        try {
            this.conn = new BoardTransmitterConnection(this, url, logger);
        } catch (ConnectionException e) {
            if (this.logger != null)
                this.logger.logMessage("[BT] " + e.toString());
        }

        try {
            this.gui = new BoardTransmitterMIDPGUI(this, logger);
        } catch (Exception e) {
            if (this.logger != null)
                logger.logMessage("Debug102: " + e.toString());
        }

        try {
            this.bet.addPlugin(this.gui);
        } catch (Exception e) {
            if (this.logger != null)
                logger.logMessage("Debug103: " + e.toString());
        }

        // handle gps/gsm with events
        //TODO this should be configurable using the xml file...
        //this.ignoredProducerTypes.addElement(new Integer(Properties.GPS_READER));
        //this.ignoredProducerTypes.addElement(new Integer(Properties.LOCAL_GSM_READER));
        // do not init the filters (cheat and tell the system that they are already initialized)
        filterInited = true;

    	producerCache = new ProducerCache(logger);


        // The new subscriber interface.
        this.bb = BlackboardFactory.getBlackboard(logger);
        this.bb.subscribeConsumer(this);  // receive notifications of new producers
        // If readers were created before me?






        // This sends a dummy-packet to the server in 'PING_DELAY_MINUTES' intervals.
        // Used to force the socket-connection to stay open, even if filtering keeps the connection idle.
        if (BetelgeuseMIDP.pingSocketToServer == true)
        	initPingTimer();
    }

    /*
     * The caller of this method needs to handle any changes to the session ID.
     * Meaning if the session ID changes, the sensor configurations need to be
     * retransmitted.
     */

    private boolean openConnection() {
        String friendlyName = this.dgat.getFriendlyName(), bluetoothAddress = this.dgat
        .getBluetoothAddress();
        if (bluetoothAddress == null) {
            if (this.logger != null)
                this.logger.logMessage("[BT] "
                        + "The bluetooth address of Betelgeuse is null!");
            return false;
        }

        try {
            // FIXME
            int id = conn.openConnection(this.sessionID, 0,
                    bluetoothAddress, this.compression, friendlyName, this.bet
                    .getUserName(), this.bet.getPassword());
        	if (this.sessionID == 0){
        		this.sessionID = id;
                this.bb.setSessionId(this.sessionID);
        	}
        } catch (ConnectionException e) {
            if (this.logger != null)
                this.logger.logMessage("[BT] "
                        + "Could not send connection initialization message.");
            Alert a = new Alert("Could not send connection init message!");
            a.setTimeout(Alert.FOREVER);
            this.bet.showAlert(a);
            return false;
        }
        return true;
    }

    /**
     *
     */

    /*
     * The data message format is:
     *
     * 1 3 2 2
     * ---------------------------------------------------------------------------------------------- |
     * Cmd | Length of the msg | Session ID | Number of blocks | Data | ... |
     * Data |
     * ----------------------------------------------------------------------------------------------
     *
     * in which each "data" field is:
     *
     * 2 1 1 2 x 1 2 x 1 2 x
     * ------------------------------------------------------------------------------------------------------------- |
     * Sensor ID | NOP | Param Icons.updateProducerList(this, this.producers);D |
     * Length | Value | Param ID | Length | Value | ... | Param ID | Length |
     * Value |
     * -------------------------------------------------------------------------------------------------------------
     *
     * NOP = number of parameters
     */

    /**
     * Starts the transmitter. If the transmitter has not been turned on, run
     * method will wait until switchOn() has been called.
     */

    public synchronized void run() {
        if (this.logger != null)
            logger.logMessage("[BT] " + "DataTransmitter plugin started.");
        long time1 = 0, time2, waitTime = maxSampleTimeout;
        // long lastCheck = 0;
        double oldLatMin = 0, oldLongMin = 0;
        int oldLatDeg = 0, oldLongDeg = 0;
        ByteArrayOutputStream out1 = new ByteArrayOutputStream();
        ByteArrayOutputStream out2 = new ByteArrayOutputStream();

        // Little hack... connector only updates the readers every 1 min, so
        // lets "get" an update!
        // This speeds up the startup...

        if (this.init) {
            try {
                this.wait();
            } catch (InterruptedException e) {
                // woken up, got readers
            }
        }

        while (isRunning) {

            if (newaddress && !addressChanged) {
                close(false);
                tryConnect = true;
                switchOff();
                addressChanged = true;
            }

            if (!switchedOn) {
                try {
                    this.wait(10000);
                } catch (InterruptedException e) {
                    // woken up, no matter...
                }
            } else {
                if (tryConnect) {
                    if (openConnection()) {
                        this.gui.showSwitchedOn(true);
                        tryConnect = false;
                    } else {
                        // unable to open connection. let's try again.
                    	BetelgeuseMIDP.logger.logMessage("BT: openConn failed...");
                        continue;
                    }
                }


                // This needs to be done before waiting, to prevent a user is
                // too quick - bug

                // Do we need to sleep for a while? You don't want to send too
                // much data...
                if ((time2 = System.currentTimeMillis()) - time1 < maxSampleTimeout) {
                    try {
                        this.wait(maxSampleTimeout - (time2 - time1));
                    } catch (InterruptedException e) {
                        // Woken up, nothing to do...
                    }
                }

                /*
                if (this.useAdaptiveSendInterval && this.lastUploadTime != -1) {
                	BetelgeuseMIDP.logger.logMessage("BT: cal?: " + adaptiveSendCalibrated + ", lastUploadTime: " + lastUploadTime );

                	if (!adaptiveSendCalibrated) {
	                	if (lastUploadTime < 1) {
	                		try {
	                			adaptiveSendCalibrateWaitPerc *= 2;
	                        	int calibrateWait = adaptiveSendCalibrateWaitPerc*1000;
	                			BetelgeuseMIDP.logger.logMessage("BT: calWait: " + calibrateWait);

	                        	this.wait(calibrateWait);
	                		} catch (InterruptedException e) {
	                         }
	                	} else {
	    	                this.optimalBufferSize = this.bufferSizePerc*this.lastSendKbs;
	    	                this.adaptiveSendCalibrated = true;
	                	}
                	} else {
                		double bufferPerc = this.optimizeSendIntervalAmount*(this.currentBufferSize/this.optimalBufferSize);
                    	long now = time2;
                    	double delayPerc = (1-this.optimizeSendIntervalAmount)*( (now-this.lastSendTime)/this.maxSampleTimeout );

                    	BetelgeuseMIDP.logger.logMessage("BT: bufP: " + bufferPerc + ", delP: " + delayPerc + ", skip?: " + ((bufferPerc+delayPerc) < 1) + ", opt: " + optimalBufferSize + ", optAmount: " + this.optimizeSendIntervalAmount);

                    	if ((bufferPerc+delayPerc) < 1) {
                    		continue;
                    	}
                    	this.lastSendTime = now;
                	}
                }
                BetelgeuseMIDP.logger.logMessage("BT: adapt AFTER");
                */

                if (newaddress) {
                    if (addressChanged) {
                        sendConfigurations(true);
                        this.newaddress = false;
                    } else {
                        continue;
                    }
                } else {
                    // The transmitter is up and running. Check first if new
                    // Readers have been found.

                    if (this.newProducers.size() > 0) {
                        // There are new readers. Construct sensor configuration
                        // messages and send them.
                        sendConfigurations(false);
                    }

                }


                // Close method might have been called while the thread was
                // sleeping
                if (!isRunning)
                    return;

                if (allProducers.size() > 0) {
                    if (this.ldr == null)
                        this.ldr = this.dgat.getConnector().getLDR();

                    int dataSize = 0;

                    long startTime = System.currentTimeMillis();
                    // Loop through all the Reader objects and construct data
                    // messages.
                    Enumeration enumer = producerID.keys();
                    int sensorId, paramId, blocks = 0;
                    out1.reset();
                    out2.reset();

                    /*
                     * Write each sensor block to out1 ByteArrayStream!
                     */
                    while (enumer.hasMoreElements()) {
                        Customer cust = (Customer) enumer.nextElement();
                        CustomerInfo ci = cust.getInfo();
                        Integer custType = new Integer(ci.getType());

                        if (!ci.isAlive()){
                            continue;
                        } else if (this.ignoredProducerTypes.contains(custType)){
                        	// producer handled with events, have to check for events waiting to be saved...
                        	if (!this.eventsToSave.contains(custType)){
                        	//	continue;
                        	} else {
                        		this.eventsToSave.removeElement(custType);
                        	}
                        	continue;
                        }

                        // Get the parameter mappings for the Reader
                        Hashtable parameterID = (Hashtable) this.producerParameters
                        .get(cust);

                        try{
                            switch (ci.getType()) {

                            case Properties.ECG_ACC_BYTE_READER:
                            case Properties.ICUBEX_READER:
                            //case Properties.LOCAL_GSM_READER:
                            case Properties.GPS_READER:
                                continue;

                            default:
                                if(!mode && (ci.getType() == Properties.LOCAL_READER
                                        || ci.getType() == Properties.GSM_READER
                                        || ci.getType() == Properties.GPS_READER

                                        || ignoredProducerTypes.contains(new Integer(ci.getType())) // handled with events, do not get data from BB. added 22.6.09
                                       // || ci.getType() == Properties.LOCAL_GSM_READER  // added 8.6.09
                                        ))
                                    continue;



                            	CustomerData cd = this.bb.getData(cust);
                            	//

                            	/*
                            	// extract params
                            	if (cust.getInfo().getType() == Properties.GPS_READER) {
                            		cd = ProducerCache.extractParameters(cd, new String[] { "Latitude", "Longitude"}, new int[] { Properties.DATATYPE_STRING, Properties.DATATYPE_STRING } );
                            		ci = cd.getInfo();
                            		// BetelgeuseMIDP.logger.logMessage("RUN extracted: " + cd);
                            	}


                            	if (ci.getType() == Properties.LOCAL_GSM_READER) {
	                            	// check cache
	                            	 *
	                            	 cd = ProducerCache.extractParameters(cd, new String[] { "BTDevices"}, new int[] { Properties.DATATYPE_STRING } );
	                            	 cd = producerCache.getUpdated(cust, cd);
	                            	 ci = cd.getInfo();
	                            	//
                            	}
                            */
                            	/*
                            	if (ci.getType() == Properties.LOCAL_GSM_READER) {
                               		cd = ProducerCache.extractParametersByPosition(
                               				cd,
                               				gsmReaderIncludeParamNames,
                               				gsmReaderIncludeParamPositions,
                               				gsmReaderIncludeParamTypes
                               		);
                               		ci = cd.getInfo();
                            	}*/



                            	if (cd == null){
                                	continue;

                            }




                            blocks++;
                            String[] params = ci.getParameterNames();
                            sensorId = ((Integer) this.producerID.get(cust))
                            .intValue();
                            // write the sensor ID
                            intToBytes(sensorId, 2, out1);
                            // Write the timestamp
                            String ts = "" + (System.currentTimeMillis()-this.conn.getTimeDifference());
                            intToBytes(ts.length(), 1, out1);
                            try {
                                out1.write(ts.getBytes());
                            } catch (IOException e1) {
                                // TODO Auto-generated catch block
                                e1.printStackTrace();
                            }
                            // write the number of parameters
                            intToBytes(params.length, 1, out1);

                            Vector[] data = cd.getData();
                            int[] types = this.bb.getData(cust)
                            .getParameterTypes();
                            for (int i = 0; i < params.length; i++) {
                                paramId = ((Integer) parameterID.get(params[i]))
                                .intValue();
                                // Write the parameter's ID
                                intToBytes(paramId, 1, out1);
                                // Write the parameter's length

                                String value = "";
                                int type = types[i];
                                if (type == 0) {
                                    value = (String) data[i].elementAt(0);
                                } else if (type == 1) {
                                    value += (Integer) data[i].elementAt(0);
                                } else if (type == 2) {
                                    value += (Double) data[i].elementAt(0);
                                } else if (type == 3) {
                                    value += (Float) data[i].elementAt(0);
                                }

                                intToBytes(value.length(), 2, out1);

                                // Write the parameter
                                try {
                                    out1.write(value.getBytes());
                                } catch (IOException e) {
                                    // This should not happen
                                }
                            }
                            break;
                            } // End of switch
                        }catch (Throwable tb){
                            logger.logMessage("[BT] switch catch: " + tb);
                        }
                    } // End of while
                    if (blocks == 0 && streamData.size() == 0) {
                        time1 = System.currentTimeMillis();
                        continue;
                    }

                    // Now we are sure that we will send something.
                    transmissionNumber++;

                    // Write the command byte
                    int length = 4;
                    byte[] data = null, data2 = null;
                    if (blocks > 0) {
                        data = out1.toByteArray();
                        length += data.length;
                    }

                    StringBuffer strb = new StringBuffer();
                    int streamBlocks = streamData.size();

                    for (int i = 0; i < streamBlocks; i++) {
                        strb.append((String) streamData.elementAt(i));
                    }

                    this.streamData = new Vector();

                    data2 = strb.toString().getBytes();
                    length += data2.length;
                    int allBlocks = blocks;

                    if (data2.length > 0)
                        allBlocks += streamBlocks;

                    try {
                        out2.write("D".getBytes());
                    } catch (IOException e) {
                        // This should not happen
                    }
                    dataSize = length + 4;
                    intToBytes(length, 3, out2);

                    // Write the session ID
                    intToBytes(this.sessionID, 2, out2);
                    intToBytes(allBlocks, 2, out2);


                    long startSendTime = System.currentTimeMillis();

                    try {
                        try {
                            if (data != null)
                                out2.write(data);
                            if (data2 != null && data2.length > 0)
                                out2.write(data2);
                        } catch (IOException ioe) {
                            // Do nothing
                        }

                        this.conn.send(out2.toByteArray());

                    } catch (ConnectionException e) {
                        // Eek! It failed! Save the session ID...
                        int temp = this.sessionID;
                        // Try to reconnect
                        // Note: If the battery is low, this might not give up a
                        // popup immediately
                        if (this.openConnection()) {
                            // Did the session ID change?
                            if (this.sessionID != temp)
                                // It did, retransmit the sensor configurations,
                                // forget about sending the data
                                this.sendConfigurations(true);
                            else {
                                // It did not, retransmit the data
                                try {
                                    this.conn.send(out2.toByteArray());
                                } catch (ConnectionException e1) {
                                    // Bugger! It didn't work! Forget about it
                                    // and move on with your life...
                                    if (this.logger != null)
                                        this.logger
                                        .logMessage("[BT] "
                                                + "Failed to send a data message: "
                                                + e.toString());
                                }
                            }
                        }
                    }
                    lastSendTime = System.currentTimeMillis();


                    //


                    long endTime = lastSendTime;
                    long transmissionTime = endTime - startTime;

                    if (BetelgeuseMIDP.measureKbs == true){
	                    try {
		                    // measure kbs
		                    long totSendTime = endTime - startSendTime;
		                    double totData = 0;
		                    if (data != null) {
		                    	totData += (double)data.length;
		                    }
		                    if (data2 != null) {
		                    	totData += (double)data2.length;
		                    }
		                    this.lastSendKb =  totData/(double)1024 ;
		                    this.lastUploadTime = (double)totSendTime/1000;
		                    this.lastSendKbs = lastSendKb / this.lastUploadTime;

		                    // BetelgeuseMIDP.logger.logMessage("BT: sec: " + this.lastUploadTime + ", kb: " + lastSendKb + ", kbs: " + this.lastSendKbs);
	                    } catch (Exception e){
	                    	BetelgeuseMIDP.logger.logMessage("BT measure fps ERR: " + e);
	                    }
	                    //
                    }


                    if (this.ldr != null) {
                        for (int i = 0; i < ldr.size(); i++) {
                            try {
                                LocalDeviceReader temp = (LocalDeviceReader) ldr
                                .elementAt(i);
                                temp.sendTransmissionParams(transmissionNumber,
                                        transmissionTime, dataSize);
                            } catch (ClassCastException e) {
                                // Ooops... someone screwed up when creating the
                                // LocalDeviceReader objects
                                this.ldr.removeElementAt(i);
                            }
                        }
                    }

                }


                // close socket-connection between sends?
                if (this.closeConnectionBetweenSends) {
                	tryConnect = true;
	                try {
						// Don't call closeConnection() here, since it sends the QUIT-message to the server,
	                	// after we wouldn'tbe able to reopen the connection again with the same sessionID...
	                	// This way, the connection is reopened on next run.
	                	this.conn.closeConnection();
					} catch (ConnectionException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
                }
                //

            }
            time1 = System.currentTimeMillis();
        }

        /*
         * Send the connection termination message:
         *
         * 1 3 2 -------------------------------------------- | Cmd | Length of
         * the message | Session ID |
         * --------------------------------------------
         */

        ByteArrayOutputStream byteout = new ByteArrayOutputStream();

        try {
            byteout.write("Q".getBytes());
            intToBytes(2, 3, byteout);
            intToBytes(sessionID, 2, byteout);
        } catch (IOException e) {
            // Should not happen
        }
        try {
            this.conn.send(byteout.toByteArray());
            int[] message = this.conn.receive();

            if (message[0] == 0x54 || message[0] == 0x74) {
                this.conn.closeConnection();
            } else {
                // TODO The server for some reason doesn't want to close the
                // connection
            }
        } catch (ConnectionException e) {
            // Something went wrong, nothing to do...
        }
    }

    /*
     * This method constructs, sends and saves sensor configurations. Format of
     * the message:
     *
     * 1 3 2 2
     * ------------------------------------------------------------------------------------------------------- |
     * Cmd | Length of the message |Session ID | Number of sensors | Sensor |
     * ... | Sensor |
     * -------------------------------------------------------------------------------------------------------
     *
     * in which each "sensor" field is:
     *
     * 12 2 1 2 m n
     * ---------------------------------------------------------------------- |
     * BT address | Sensor type | LON | LOP | Friendly name | Parameters |
     * ----------------------------------------------------------------------
     */

    private void sendConfigurations(boolean reconfigureAll) {
     	try {
            Vector configure = new Vector();
            Hashtable tempTable = new Hashtable();
            if (reconfigureAll) {
                for (int i = 0; i < allProducers.size(); i++)
                    configure.addElement(allProducers.elementAt(i));
                newProducers = new Vector();
            } else {
                for (int i = 0; i < newProducers.size(); i++)
                    configure.addElement(newProducers.elementAt(i));
                newProducers = new Vector();
            }

            ByteArrayOutputStream out = new ByteArrayOutputStream();

            String address, parameters, name;
            String[] params;
            int type;

            for (int i = 0; i < configure.size(); i++) {
                Customer cust = (Customer) configure.elementAt(i);
                CustomerInfo ci = cust.getInfo();
                address = ci.getAddress();

                name = ci.getName();
                type = ci.getType();

                try {
                    // Write the sensor ID
                    intToBytes(this.nextReaderId, 2, out);
                    producerID.put(cust, new Integer(this.nextReaderId));
                    this.nextReaderId++;
                    // Write the bluetooth address
                    out.write(address.getBytes());
                    // Write the sensor type number
                    intToBytes(type, 2, out);
                    // Write the length of the name field
                    intToBytes(name.length(), 1, out);
                    // Write the sensor name
                    out.write(name.getBytes());
                    // Write the number of parameters
                    String[] parameterNames = ci.getParameterNames();
                    intToBytes(parameterNames.length, 1, out);
                    int nextParameterId = 0;
                    Hashtable parameterID = new Hashtable();
                    for (int j = 0; j < parameterNames.length; j++) {
                        // Write parameter information...
                        intToBytes(nextParameterId, 1, out);
                        intToBytes(parameterNames[j].length(), 2, out);
                        out.write(parameterNames[j].getBytes());

                        // Save the mapping...
                        parameterID.put(parameterNames[j], new Integer(
                                nextParameterId));
                        nextParameterId++;
                    }
                    this.producerParameters.put(cust, parameterID);

                } catch (IOException e) {
                    // Nothing to be done...
                	logger.logMessage("BT: " + e);
                	System.out.println("[BT] " + e);
                } catch (Throwable th){
                    logger.logMessage("[BT] w paramnames caught: " + th);
                }
            }
            byte[] data = out.toByteArray();
            out.reset();

            // Message command ('S' or 's')
            out.write(0x53);
            // Write the message length
            intToBytes(data.length + 4, 3, out);
            // Session ID
            intToBytes(this.sessionID, 2, out);
            // Number of readers configured in this message
            intToBytes(configure.size(), 2, out);
            // Write the data
            try {
                out.write(data);
            } catch (IOException e1) {
                // Should not happen.
            	logger.logMessage("BT ioException: " + e1);
            }

            byte[] msg = out.toByteArray();
            int[] message;
            try {
                conn.send(msg);
                // Read the sensor identification message...
                message = conn.receive();
            } catch (ConnectionException e) {
                if (logger != null)
                    logger
                    .logMessage("[BT] "
                            + "Transmitting sensor identification message failed: "
                            + e.toString());
                close();
                return;
            } catch (Exception e){
            	return;
            }

            // Handle the responce...

            try {
                int i = 0;
                /*
                if (message[0] != 'i' && message[0] != 'I' && logger != null) {
                    logger
                    .logMessage("[BT] "
                            + "Invalid sensor identification reply received from the server - wrong message command. Transmitter closed.");
                    close();
                }
                int sensors = message[6] * 256 + message[7];
                if (sensors != configure.size()) {
                    if (logger != null)
                        logger
                        .logMessage("[BT] "
                                + "Invalid sensor identification reply received from the server - not enough sensor indentifications. Transmitter closed.");
                    close();
                    return;
                }
                */
                // TODO Handle the rest of the message. The server might not
                // want all of the data...
                // TODO This feature hasn't been added to this data transmitter!

                // For now we only use this for the ECG_ACC_BYTE_READER, which
                // is designed for this purpose...
                for (int j = 0; j < configure.size(); j++) {
                	Customer cust = ((Customer) configure.elementAt(j));
                	CustomerInfo pi = cust.getInfo();
                	if (pi.getType() == Properties.ECG_ACC_BYTE_READER
                			|| pi.getType() == Properties.ICUBEX_READER
                			|| pi.getType() == Properties.CALL_READER
                			|| pi.getType() == Properties.GPS_READER
                			//  || pi.getType() == Properties.LOCAL_GSM_READER   // added 5.6.09
                	) {
                		if (!this.subscriptions.contains(cust)
                				&& !this.ignoredProducerTypes.contains(new Integer(cust.getInfo().getType()))) {
                			if (logger != null)
                				logger.logMessage("[BT] Subscribing to: "
                						+ pi.getName());
                			// Get all of the data from the ECG reader...
                			this.bb.subscribeToProducer(this, cust);
                			this.subscriptions.addElement(cust);
                		}
                	}

                    if (!this.ignoredProducerTypes.contains(new Integer(cust.getInfo().getType()))){
	                    if(!mode && !this.subscriptions.contains(cust)){
	                        if(pi.getType() == Properties.GPS_READER){
	                            this.bb.subscribeToProducer(this, cust);
	                            this.subscriptions.addElement(cust);
	                        }
	                        if(pi.getType() == Properties.GSM_READER){
	                            this.bb.subscribeToProducer(this, cust);
	                            this.subscriptions.addElement(cust);
	                        }
	                        if(pi.getType() == Properties.LOCAL_READER){
	                            this.bb.subscribeToProducer(this, cust);
	                            this.subscriptions.addElement(cust);
	                        }
	                        /*
	                        // added 8.6.09
	                        if(pi.getType() == Properties.LOCAL_GSM_READER){
	                            this.bb.subscribeToProducer(this, cust);
	                            this.subscriptions.addElement(cust);
	                        }
	                        //
	                         */
	                    }
                    }
                }


            } catch (IndexOutOfBoundsException e) {
                if (logger != null)
                    logger
                    .logMessage("[BT] "
                            + "Invalid sensor identification reply received from the server - message is too short. Transmitter closed.");
                close();
                return;
            }
        } catch (Throwable t) {
            logger.logMessage("[BT] sendconf caught: " + t);
        }
    }

    /**
     * Returns true if Transmitter is switched on, false otherwise.
     *
     * @return true if transmitter is online, false otherwise.
     */

    public boolean switchedOn() {
        return switchedOn;
    }

    /**
     * Turns the Transmitter on. Causes the TransmitterConnection object to
     * connect to a server.
     */

    public boolean switchOn() {
        this.switchedOn = true;
        tryConnect = true;
        return true;
    }

    /**
     * Turns the Transmitter off. Causes the TransmitterConnection object to
     * disconnect from the server. Calling this method will not cause the thread
     * to die. The Transmitter will continue transmitting the data after
     * switchOn() is called. If you want to destroy the transmitter, call the
     * close() method.
     *
     */

    public void switchOff() {
        this.switchedOn = false;
        this.gui.showSwitchedOn(false);
    }

    /**
     * Closes the Transmitter. Note that calling this method causes the thread
     * to die.
     *
     */

    public void close() {
        close(true);
    }

    private void close(boolean andStop) {
        if (andStop) {
            isRunning = false;
        }
        if (!this.switchedOn) {
            // not on, so nothing to close
            return;
        }

        ByteArrayOutputStream byteout = new ByteArrayOutputStream();

        try {
            byteout.write("Q".getBytes());
            intToBytes(2, 3, byteout);
            intToBytes(sessionID, 2, byteout);
        } catch (IOException e) {
            // Should not happen
        }
        try {
            this.conn.send(byteout.toByteArray());
            int[] message = this.conn.receive();

            if (message[0] == 0x54 || message[0] == 0x74) {
                this.conn.closeConnection();
            } else {
                // TODO The server for some reason doesn't want to close the
                // connection
            }
        } catch (ConnectionException e) {
            // Something went wrong, nothing to do...
        }
    }

    /**
     * Rerturns the TransmitterConnection object.
     *
     * @return the TransmitterConnection object used to send data
     */

    public BoardTransmitterConnection getConnection() {
        return this.conn;
    }

    /**
     * Returns the maximum interval for sending data. Smaller value causes the
     * Transmitter to send data more often.
     *
     * @return the maximum transmission interval in milliseconds
     */

    public int getMaxSampleTimeout() {
        return this.maxSampleTimeout;
    }

    /**
     * Sets the maximum intreval for sending data. Smaller value causes the
     * Transmitter to send data more often.
     *
     * @param timeout
     *            the maximum transmission interval in milliseconds
     */

    public void setMaxSampleTimeout(int timeout) {
        this.maxSampleTimeout = timeout;
    }

    /*
     * A general method for finding Readers. Problems: multiple devices,
     * multiple devices in the same class, etc... Figure out how to use this if
     * you want the transmitter to handle multiple devices
     */

    private void intToBytes(int number, int size, ByteArrayOutputStream out) {
        byte[] array = new byte[size];

        for (int i = size - 1; i >= 0; i--) {
            array[i] = (byte) (number % 256);
            number = number / 256;
        }
        try {
            out.write(array);
        } catch (IOException e) {
            // TODO
        }
    }

    private void longToBytes(long number, int size, ByteArrayOutputStream out) {
        byte[] array = new byte[size];

        for (int i = size - 1; i >= 0; i--) {
            array[i] = (byte) (number % 256);
            number = number / 256;
        }
        try {
            out.write(array);
        } catch (IOException e) {
            // TODO
        }
    }

    /**
     * This method is called when the DataTransmitter should change the address
     * of the server. It does this by creating a new DataTransmitterConnection
     * in the run; the address has already been changed there.
     *
     */

    public void changeAddress() {
        this.newaddress = true;
        this.addressChanged = false;
    }

    public CustomerInfo getInfo() {
        return null;
    }

    public void newData(Customer prod, CustomerData pd) {
    	// BetelgeuseMIDP.logger.logMessage("BT newData raw: " + prod.getInfo().getName() + ", GPS? " + (prod.getInfo().getType() == Properties.GPS_READER) + ", pd: " + pd );


    	/*
    	if (prod.getInfo().getType() == Properties.GPS_READER) {
    		// pd = ProducerCache.extractParameters(pd, gpsReaderIncludeParamNames, gpsReaderIncludeParamTypes);
    		pd = ProducerCache.extractParametersByPosition(pd, gpsReaderIncludeParamNames, gpsReaderIncludeParamPositions, gpsReaderIncludeParamTypes);
    	}
    	*/

        // TODO Auto-generated method stub
        String time = "" + (pd.getTimestamp()-this.conn.getTimeDifference());
        int[] types = pd.getParameterTypes();
        String[] params = pd.getParameterNames();

        Object o = this.producerID.get(prod);
        if (o == null)
            return;
        int id = ((Integer) o).intValue();
        Hashtable paramMappings = (Hashtable) this.producerParameters.get(prod);

        ByteArrayOutputStream message = new ByteArrayOutputStream();
        int paramCount = 0;
        Vector[] data = pd.getData();

        for (int i = 0; i < params.length && i < data.length; i++) {
            Vector v = data[i];

            int paramId = ((Integer) paramMappings.get(params[i])).intValue();
            int type = types[i];

            if (type == 0) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = (String) o;
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {

                    }
                    paramCount++;
                }
            } else if (type == 1) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = "" + ((Integer) o).intValue();
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {

                    }
                    paramCount++;
                }
            } else if (type == 2) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = "" + ((Double) o).doubleValue();
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {

                    }
                    paramCount++;
                }
            } else if (type == 3) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = "" + ((Float) o).floatValue();
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {

                    }
                    paramCount++;
                }
            }
        }

        ByteArrayOutputStream byteout = new ByteArrayOutputStream();
        intToBytes(id, 2, byteout);
        intToBytes(time.length(), 1, byteout);

        try {
            byteout.write(time.getBytes());
            intToBytes(paramCount, 1, byteout);
            byteout.write(message.toByteArray());
        } catch (IOException e) {
            if (logger != null) logger.logMessage("BT: " + e);
        }
        if (byteout.size() > 0)
            this.streamData.addElement(byteout.toString());



        // testing: adjust GPS-request interval according to when the last distanceChange -event was received.
        if (BetelgeuseMIDP.adjustGPSRequestInterval == true) {
	        if (prod.getInfo().getType() == Properties.GPS_READER) {
	        	// GPS has to be in 'request-mode'...
	        	if(((Boolean)(Properties.getIdMode().get(new Integer(Properties.GPS_READER)))).booleanValue() == false) {

	        		long now = System.currentTimeMillis();
		        	if (lastGPSUpdate != -1) {
		            	long dif = now-lastGPSUpdate;
		        		int interval = bet.getGPSSampleInterval();
		        		int newInterval = interval;
		        		if (dif < interval) {
		        			newInterval = (int)dif; // interval/2;
		        		} else if (dif > (interval*2)) {
		        			newInterval = interval*2;
		        		}
		        		// limit interval to 10000-60000
		        		if (newInterval < 10000)
		        			newInterval = 10000;

		        		if (newInterval > 60000)
		        			newInterval = 60000;

	        			bet.setGPSSampleInterval(newInterval);
		        	}
		        	lastGPSUpdate = now;
	        	}
	        }
        }
        //




        // currentBufferSize += byteout.size();
    }

    public void newData(Customer prod, CustomerData pd, EventData eventData){
		//if (logger != null) logger.logMessage("* BoardTrans: newData EV, prod: " + prod.getInfo().getName() + ", " + eventData.getName() + ", fil:" + eventData.getFilter() );
    	Integer prodType = new Integer(prod.getInfo().getType());
    	if (!this.eventsToSave.contains(prodType))
    		this.eventsToSave.addElement(prodType);


    	this.newData(prod, pd);
	}

    public boolean requestData() {
        return false;
    }

    public void setCustomerInfo(CustomerInfo info) {
        // TODO Auto-generated method stub

    }

    public boolean setProducerMode(int mode, long timeout) {
        return false;
    }

    public void updateCustomerInfo(Blackboard bb, Customer c, CustomerInfo pi) {
        this.newProducers.addElement(c);
    }

    public void updateProducerList(Blackboard bb, Vector producers) {
    	// System.out.println("BT: update prod list: " + producers.size());
    	// if (logger != null) logger.logMessage("BT: updateProdList: " + producers.size());

    	for (int i = 0; i < producers.size(); i++) {
            Customer cust = (Customer) producers.elementAt(i);
        	// if (logger != null) logger.logMessage("BT: updateProdList: " + cust + ", cont: " + allProducers.contains(cust) + ", internal: " + cust.getInfo().isInternal() );
            if (!allProducers.contains(cust) && (!cust.getInfo().isInternal())) {

            	//allProducers.addElement(cust);
            	//newProducers.addElement(cust);

        		/*
            	if (cust.getInfo().getType() == Properties.LOCAL_GSM_READER){
            		allProducers.addElement(cust);
            		newProducers.addElement(cust);
            	}
            	*/


                 /*
            	// subscribe to gps-reader
            	if (cust.getInfo().getType() == Properties.GPS_READER && gpsReader == null) {
            		allProducers.addElement(cust);
                    newProducers.addElement(cust);
                    gpsReader = cust;

                    this.bb.subscribeToEvent(this, gpsReader, "distanceChange");

            	/*
            	} else {
            		// and others (LGR)
            	 	allProducers.addElement(cust);
                    newProducers.addElement(cust);
            	*/
            	//}



            	//
            	/*
            	if (cust.getInfo().getType() == Properties.GPS_READER){
            		allProducers.addElement(cust);
            		newProducers.addElement(cust);
            	}
            	*/


                // handle producer with events:
                //if (ignoredProducerTypes.contains(new Integer(cust.getInfo().getType()))) {
            	int custType = cust.getInfo().getType();
                if (custType == Properties.LOCAL_GSM_READER || custType == Properties.GPS_READER){
            	 //if (custType == Properties.GPS_READER){
            		allProducers.addElement(cust);
            		newProducers.addElement(cust);

                	switch (cust.getInfo().getType()){
                	case Properties.GPS_READER:
                		this.gpsReader = cust;
                		break;

                	case Properties.LOCAL_GSM_READER:
                		this.gsmReader = cust;

                		break;
                	}

                	// if (gpsReader != null)
                	// if (gsmReader != null)
                	if (gpsReader != null)
                		prepareEventFilter();
                }

            }
        }
        if (this.logger != null)
            logger.logMessage("[BT] "
                    + "Data sources received. Number of data sources: "
                    + this.allProducers.size());
        if (this.init) {
            synchronized (this) {
                this.notify();
                init = false;
            }
        }
    }


	 private void prepareEventFilter(){
		 if (!filterInited){
			 filterInited = true;

			 //TODO try to understand the following code:


			 // filtering

			 Filter distanceFilter = new GPSVelocityFilter(gpsReader, this, 0.015, CompareFilter.TYPE_GREATER_THAN, "HDOP");

			 ConstraintFilter c = new ConstraintFilter(distanceFilter, gpsReader, logger);
			 c.setName("const sat: > 3, hdop: < 7");
			 CompareFilter satellitesF = new CompareFilter(gpsReader, null, 3, CompareFilter.TYPE_GREATER_THAN, false);
			 CompareFilter hdopF = new CompareFilter(gpsReader, null, 6.1, CompareFilter.TYPE_LESS_THAN, false);

			 c.addConstraint(satellitesF);
			 c.addConstraint(hdopF);

			 this.bb.subscribeToEvent(satellitesF, gpsReader, "satellitesChange");
			 this.bb.subscribeToEvent(hdopF, gpsReader, "HDOPChange");
			 this.bb.subscribeToEvent(c, gpsReader, "distanceChange");


			 /*FILTERING USED IN 22th-23th Sebtember 2009 test
			 ConstraintFilter c = new ConstraintFilter(this, gpsReader, logger);
			 c.setName("const sat: > 3, hdop: < 7");

			 Filter distanceFilter = new DynamicCompareFilter(gpsReader, c, 0.015, CompareFilter.TYPE_GREATER_THAN, false, "HDOP");


			 CompareFilter satellitesF = new CompareFilter(gpsReader, null, 3, CompareFilter.TYPE_GREATER_THAN, false);
			 CompareFilter hdopF = new CompareFilter(gpsReader, null, 6.1, CompareFilter.TYPE_LESS_THAN, false);

			 c.addConstraint(satellitesF);
			 c.addConstraint(hdopF);

			 this.bb.subscribeToEvent(satellitesF, gpsReader, "satellitesChange");
			 this.bb.subscribeToEvent(hdopF, gpsReader, "HDOPChange");
			 this.bb.subscribeToEvent(distanceFilter, gpsReader, "distanceChange");
			  */




			 /* FILTERING USED IN 18th-20th Sebtember 2009 tests
			 // hdop < 7 & satellites >= 4 && distance 20 m
			ConstraintFilter c = new ConstraintFilter(this, gpsReader, logger);
			c.setName("const sat: > 3, hdop: < 7");

			//Filter distanceFilter = new CompareFilter(gpsReader, c, 0.02, CompareFilter.TYPE_GREATER_THAN, true);


			CompareFilter satellitesF = new CompareFilter(gpsReader, null, 3, CompareFilter.TYPE_GREATER_THAN, false);
			CompareFilter hdopF = new CompareFilter(gpsReader, null, 6.1, CompareFilter.TYPE_LESS_THAN, false);

			c.addConstraint(satellitesF);
			c.addConstraint(hdopF);

			this.bb.subscribeToEvent(satellitesF, gpsReader, "satellitesChange");
			this.bb.subscribeToEvent(hdopF, gpsReader, "HDOPChange");

			//this.bb.subscribeToEvent(distanceFilter, gpsReader, "distanceChange");
			this.bb.subscribeToEvent(c, gpsReader, "satHDOPConstraint");
			//
			*/




			/* SAMULI'S OLD CODE, DO NOT ERASE
			 // filtering
			 // hdop < 7 & satellites >= 4 && distance 20 m
			ConstraintFilter c = new ConstraintFilter(this, gpsReader, logger);
			c.setName("const sat: > 3, hdop: < 7");

			Filter distanceFilter = new CompareFilter(gpsReader, c, 0.02, CompareFilter.TYPE_GREATER_THAN, true);


			CompareFilter satellitesF = new CompareFilter(gpsReader, null, 3, CompareFilter.TYPE_GREATER_THAN, false);
			CompareFilter hdopF = new CompareFilter(gpsReader, null, 6.1, CompareFilter.TYPE_LESS_THAN, false);

			c.addConstraint(satellitesF);
			c.addConstraint(hdopF);

			this.bb.subscribeToEvent(satellitesF, gpsReader, "satellitesChange");
			this.bb.subscribeToEvent(hdopF, gpsReader, "HDOPChange");

			this.bb.subscribeToEvent(distanceFilter, gpsReader, "distanceChange");
			//
			 */



			 /*
			 if (this.gpsReader != null)
				 this.bb.subscribeToProducer(this, gpsReader);
			  */



			 /*
			if (this.gsmReader != null)
				this.bb.subscribeToProducer(this, gsmReader);
			*/




			// BTDevices
			// BetelgeuseMIDP.logger.logMessage("BT: prepareEventFilter");
			// this.bb.subscribeToEvent(this, this.gsmReader, "BTDevicesChange");



			/*
			// Send latitude/longitude only
			// ProducerCache can extract specified fields from CustomerData:
			// ProducerCache.extractParameters(cd, includeParameterNames, includeParameterTypes), can be called from newData -method.
			//
			// currently, must register for some event to use ProducerCache...
			//
			this.bb.subscribeToEvent(this, this.gpsReader, "distanceChange");
			this.gpsReaderIncludeParamNames = new String[] { "Latitude", "Longitude"};
			this.gpsReaderIncludeParamTypes = new int[] { Properties.DATATYPE_STRING, Properties.DATATYPE_STRING };
			this.gpsReaderIncludeParamPositions = ProducerCache.getParameterPositionsByName( ((GPSReader)this.gpsReader).getParameterNames(), this.gpsReaderIncludeParamNames);
	    	// this.gpsReaderIncludeParamPositions = ((GPSReader)this.gpsReader).getParameterPositionsByName(this.gpsReaderIncludeParamNames);
			//
			*/

	        // used when extracting attributes from incoming data
			// ProducerCache.extractParameters(cd, includeParameterNames, includeParameterTypes)
			if (this.gsmReader != null) {
				 this.gsmReaderIncludeParamNames = new String[] { "BTDevices"};
				this.gsmReaderIncludeParamTypes = new int[] { Properties.DATATYPE_STRING };
				this.gsmReaderIncludeParamPositions = ProducerCache.getParameterPositionsByName( ((LocalGSMReader)this.gsmReader).getParameterNames(), this.gsmReaderIncludeParamNames);
			}

			if (this.gpsReader != null) {
				this.gpsReaderIncludeParamNames = new String[] { "Latitude", "Longitude"};
				this.gpsReaderIncludeParamTypes = new int[] { Properties.DATATYPE_STRING, Properties.DATATYPE_STRING };
				this.gpsReaderIncludeParamPositions = ProducerCache.getParameterPositionsByName( ((GPSReader)this.gpsReader).getParameterNames(), this.gpsReaderIncludeParamNames);
				//
			}


			/*
			// gps 20 m change
			Filter distanceFilter = new CompareFilter(gpsReader, this, 0.02, CompareFilter.TYPE_GREATER_THAN, true);
			distanceFilter.setLogger(logger);
			this.bb.subscribeToEvent(distanceFilter, gpsReader, "distanceChange");
			*/



			/*
			// delaying updates
			Filter gpsDelay = new DelayFilter(gpsReader, this, 10000);
			this.bb.subscribeToProducer(gpsDelay, gpsReader);

			Filter gsmDelay = new DelayFilter(gsmReader, this, 10000);
			this.bb.subscribeToProducer(gsmDelay, gsmReader);
			*/



		 }
	  }


	 /*
	 public void toggleAdaptiveSendInterval(boolean mode) {
		 this.useAdaptiveSendInterval = mode;
	 }
	 public void setAdaptiveSendIntervalValue(double value) {
		 this.optimizeSendIntervalAmount = value;
	 }
	 */

	 private void initPingTimer(){
        pingWaitTimer = new Timer();
		PingWait task = new PingWait(this, conn, sessionID, PING_DELAY_MINUTES*60000);
		long pingCheckDelay = 30000;
		pingWaitTimer.schedule(task, pingCheckDelay, pingCheckDelay);
	 }


	// ping socket
	class PingWait extends TimerTask {
		private BoardTransmitterConnection conn;
		private int sessionID;
		private BoardTransmitter parent;
		private long delay;
		public PingWait(BoardTransmitter parent, BoardTransmitterConnection conn, int sessionID, long delay) {
			this.conn = conn;
			this.parent = parent;
			this.sessionID = sessionID;
			this.delay = delay;
		}

		public void run() {
			// check if nothing has been sent
			long now = System.currentTimeMillis();
			if (now > (this.parent.lastSendTime+this.delay)) {
				this.sendPing();
				this.parent.lastSendTime = now;
			}
		}

		private void sendPing() {
			ByteArrayOutputStream byteout = new ByteArrayOutputStream();
			try {
				 // dummy packet, server doesn't recognize 'X' -messages
				byteout.write("X".getBytes());
				intToBytes(2, 3, byteout);
				intToBytes(sessionID, 2, byteout);
			} catch (IOException e) {
			}
			try {
				this.conn.send(byteout.toByteArray());
			} catch (ConnectionException e) {
			}
		}
	}
	//


    public void markParameters(){
    	//Not needed for the class...
    }

}
