/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.plugin.transmitter;

import hiit.BetelgeuseMIDP;
import hiit.logger.Logger;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import javax.microedition.io.Connector;

import javax.microedition.io.SocketConnection;

/**
 * This class is used by the DataTransmitter class to transmit sensor data
 * from BeTelGeuse to a server. The DataTransmitter uses the BeTelGeuse data
 * transfer protocol (see the manual).
 * @author Joonas Kukkonen, Eemil Lagerspetz
 *
 */

public class BoardTransmitterConnection {

    private String url = "";
    private SomeBoardTransmitter transmitter = null;
    private SocketConnection conn = null;
    private InputStream in = null;
    private OutputStream out = null;
    private int realId = -1;
    private Logger logger;
    private long timeDifference = 0;
    
    
    /**
     * Constructs a new DataTransmitterConnection
     * @param t the DataTransmitter that uses this connection
     * @param url the url of the server or object the DataTransmiterConnection connects to
     * @throws ConnectionException if and error occurred during the construction of the object
     */
    
    public BoardTransmitterConnection(SomeBoardTransmitter t, String url, Logger logger) throws ConnectionException{
        this.url = url;
        this.transmitter = t;
        this.logger = logger;
    }
    
    /**
     * Sets the address of the DataTransmitterConnection.
     * @param host - the new address
     */

    public void setAddress(String host){
        this.url = host;
        this.transmitter.changeAddress();
    }
    
    /**
     * Returns the address where data is to be sent.
     * @return address of the server
     */
    
    public String getAddress(){
        return url;
    }
    
    /**
     * Opens connection to a server by sending a connection initialization message.
     * The server sends a session ID back.
     * @param sessionID the current session ID, 0 if none
     * @param bluetoothAddress the Bluetooth address of the device
     * @param compression the compression to be used, 0 for none
     * @param friendlyName the friendly name of the device
     * @return the session ID assigned to BeTelGeuse by the server
     * @throws ConnectionException
     */
    
    public synchronized int openConnection(int sessionID, int deviceType, String bluetoothAddress, int compression, String friendlyName, String userName, String password) throws ConnectionException{
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        ByteArrayOutputStream byteOut2 = new ByteArrayOutputStream();
        
        try {
            intToBytes(sessionID, 2, byteOut2);
            intToBytes(deviceType, 1, byteOut2);
            byteOut2.write(compression);
            byteOut2.write(bluetoothAddress.getBytes());
            if (friendlyName != null){
                intToBytes(friendlyName.length(), 1, byteOut2);
                byteOut2.write(friendlyName.getBytes());
            }else
                byteOut2.write(0);
            if (userName != null){
                intToBytes(userName.length(), 1, byteOut2);
                byteOut2.write(userName.getBytes());
            }else
                byteOut2.write(0);
            if (password != null){
                intToBytes(password.length(), 1, byteOut2);
                byteOut2.write(password.getBytes());
            }else
                byteOut2.write(0);
            
            //Write the header
            byteOut.write("C".getBytes());
            byte[] data = byteOut2.toByteArray();
            //Write the length of the message
            intToBytes(data.length, 3, byteOut);
            //Write the data
            byteOut.write(data);
        } catch (IOException e) {
            // Should not happen
        }
        
        try {
            this.conn = (SocketConnection)Connector.open("socket://" + url, Connector.READ_WRITE, true);
          	
          	this.in = this.conn.openInputStream();
            this.out = this.conn.openOutputStream();
            
        } catch (IOException e) {
        	BetelgeuseMIDP.logger.logMessage("BT conn, url ERR: " + e);
            throw new ConnectionException("Failed to create a socket connection. Check the url.");
        }
        
        send(byteOut.toByteArray());
        
        /*
         * Read the response
         *     1             3                   2          1            n
         *  ----------------------------------------------------------------------
         *  | Cmd | Length of the message | Session ID | OK Byte | DB Session ID |
         *  ----------------------------------------------------------------------
         */
        
        int[] response = receive();

        long current = System.currentTimeMillis();
            
        int temp = 0;
        if(response[6] != 0){
        	BetelgeuseMIDP.logger.logMessage("BT conn, Received an error code from the server: " + response[6] );
            throw new ConnectionException("[DT] Received an error code from the server: " + response[6]);
        }
        
        int dbSessionLength = response[7];
        
        for (int i = 8; i < 8+dbSessionLength; i++){
            temp = temp * 256 + response[i]; 
        }
        this.realId = temp;
        
        int offset = 8+dbSessionLength;
        
        int timeLength = response[offset];
        offset++;
        
        //Meh... mobile java doesn't have String constructor for int array...
        char[] timeBytes = new char[timeLength];
        for(int i=0; i<timeLength; i++){
            timeBytes[i] = (char)response[i+offset];
        }
        
        String time = new String(timeBytes);
        
        if(this.logger != null) logger.logMessage("[BTC]Server time: " + time);
        
        this.timeDifference = current - Long.parseLong(time);
        
        if(this.logger != null) logger.logMessage("[BTC]Time difference: " + this.timeDifference);
        
        return (response[4]*256 + response[5]);
    }
    
    /**
     * Returns the real session ID. This ID is inside the server / DB / etc.
     * Note: this method should be called only after connection initialization.
     * @return the real session ID
     */
    
    public int getRealID(){
        return this.realId;
    }
    
    /**
     * Closes the connection.
     * @throws ConnectionException if the connection did not close properly
     */
    
    public synchronized void closeConnection() throws ConnectionException{
        
        try {
            if (this.conn != null){
            this.conn.close();  
            }
            if (this.in != null){
            this.in.close();    
            }
            if (this.out != null){
            this.out.close();
            }
        } catch (IOException e) {
            throw new ConnectionException("[DTC] Unable to close the connection.");
        }
    }
    
    /**
     * Sends a byte array through the connection.
     * @param message the message to be sent
     * @throws ConnectionException if an error occurred during the sending of the message
     */
    
    public synchronized void send(byte[] message) throws ConnectionException{
        try {
        //out = conn.openOutputStream();
        	
        out.write(message);
        out.flush();

        //out.close();
        } catch (IOException e) {
        	BetelgeuseMIDP.logger.logMessage("BTTransConn ERR: " + e);
        throw new ConnectionException("Couldn't send a message through a socket.");
        }
    }
    
    /**
     * Reads a message from the connection. The amount read depends on the length
     * field of the received message (2nd, 3rd and 4th byte).
     * @return the received message
     * @throws ConnectionException if an error occurred during the reading
     */
    
    public synchronized int[] receive() throws ConnectionException{
        /*
         * PROBLEM! Java byte's range is -128 to 127. However unsigned bytes are
         * used in the protocol. Thus we only use the read() method in InputStream
         * that maps bytes to ints with the range of 0 to 255.
         */
        try {
            
            int[] head = new int[4];
            //in = conn.openInputStream();
            int  length;
            
            //Read the command.
            head[0] = in.read();
            //Read the 3 byte length field
            head[1] = in.read();
            length = head[1] * 256 * 256;
            head[2] = in.read();
            length += head[2] * 256;
            head[3] = in.read();
            length += head[3];
            
            int [] msg = new int[length+4];
            
            for (int i = 4; i < length+4; i++){
                msg[i] = in.read();
            }
            
            msg[0] = head[0];
            msg[1] = head[1];
            msg[2] = head[2];
            msg[3] = head[3];
            //in.close();
            return msg;
        } catch (IOException e) {
            logger.logMessage("[DTC] COULDN'T READ A MESSAGE!");
            throw new ConnectionException("Could not read a message from the socket.");
        }
    }
    
    public void setSocketOption(byte option, int value) throws IllegalArgumentException, IOException {	
    	/*
    	http://codeidol.com/java/wireless-java/MIDP-Networking-and-Serial-Communications/SocketConnection/

    	DELAY
    	Small-buffer writing delay. Setting this option to zero disables the TCP Nagle algorithm for small buffer operations. 
    	Setting it to a non-zero value enables it.

    	KEEPALIVE
    	Keep-alive feature. Setting this option to zero disables the keep-alive feature. 
    	Setting it to a non-zero value enables it.

    	LINGER
    	Number of seconds to wait before closing a connection with pending data output. 
    	Setting the linger time to zero disables the linger wait interval.

    	RCVBUF
    	Size of the receiving buffer, in bytes.

    	SNDBUF
    	Size of the sending buffer, in bytes.
    	*/
    	conn.setSocketOption(option, value);
     }
    public int getSocketOption(byte option) throws IllegalArgumentException, IOException {	
    	return conn.getSocketOption(option);
    }


    private void intToBytes(int number, int size, ByteArrayOutputStream out){
        byte[] array = new byte[size];
        
        for (int i = size-1; i >= 0; i--){
            array[i] = (byte)(number % 256);
            number = number / 256;
        }
        try{
            out.write(array);
        }catch (IOException e){
            // This should not happen
        }
    }
    
    public long getTimeDifference(){
        return this.timeDifference;
    }

}