/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.plugin.transmitter;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.io.Connector;
import javax.microedition.io.file.FileConnection;
import javax.microedition.io.file.FileSystemRegistry;

import hiit.BetelgeuseMIDP;
import hiit.blackboard.Blackboard;
import hiit.blackboard.BlackboardFactory;
import hiit.blackboard.Customer;
import hiit.blackboard.CustomerData;
import hiit.blackboard.CustomerInfo;
import hiit.datagatherer.DataGatherer;
import hiit.logger.Logger;
import hiit.mapper.Properties;
import hiit.observer.ConnectorObserver;
import hiit.observer.Observer;
import hiit.blackboard.event.EventData;
import hiit.reader.GSMReader;
import hiit.reader.LocalDeviceReader;
import hiit.reader.LocalGSMReader;
import hiit.reader.Reader;

public class LocalBoardTransmitter implements Runnable, Customer {
    
    //TODO: LDR is not set and thus it does not get updated with each transmission

    private String fileName;
    private Logger logger;
    private DataGatherer dgat;

    //How many "transmissions" have been done; how many times has a data message been writen to the file
    private int transmissionNumber = 0;

    //All readers that have been configured
    private Vector allProducers = new Vector();
    //All new readers that wait configuring
    private Vector newProducers = new Vector();

    //Stream blocks
    private Vector streamData = new Vector();
    //Readers which this class has subscribed to
    private Vector subscriptions = new Vector();

    //Number of configured readers (also the number for next free reader ID)
    private int configuredReaders = 0;
    //Hashtable of Reader - Reader ID pairs (makes it easy to find correct reader ID when constructing the data message)
    private Hashtable customerID = new Hashtable();
    //Hashtable of Reader - Hashtable of Parameter names (makes it easy to find correct parameter mappings for a reader)
    private Hashtable customerParameters = new Hashtable();   

    //Local device reader which is updated by the transmitter
    private LocalDeviceReader ldr = null;

    //Is the thread running? Set to false if you want to kill the thread
    private boolean isRunning = true;
    //Is output stream opened to a file? (Handy if you want to kill the stream in another method than run)
    private boolean streamOpen = false;

    //How often should the transmitter write data to the file?
    private long maxSampleTimeout = 10000;


    private Vector readerStatus = new Vector();
    private Vector readerDataBatchNumber = new Vector();
    private Vector readerFailureTime = new Vector();

    private final int READER_OK = 1;
    private final int READER_DISCONNECTED = 2;
    private final int READER_FAILURE = 3;
    private final int READER_DEAD = 4;
    private final int CONNECTION_FAILURE_TIME = 120000;
    private final int READER_FAILURE_TIME = 120000;

    private OutputStream out;

    private String userName = "";

    private FileConnection fc = null;
    private BetelgeuseMIDP bet;

    private boolean init = true;
    private boolean mode = true;
    private GSMReader gsmReader = null;
    private LocalGSMReader gsmReader2 = null;
    private int oldId = 0;
    
    private Blackboard bb;

    public LocalBoardTransmitter(String fileName, BetelgeuseMIDP bet, DataGatherer dgat, Logger logger){
        try{
            this.bb = BlackboardFactory.getBlackboard(logger); 
            this.fileName = fileName;
            this.logger = logger;
            this.dgat = dgat;
            this.bet = bet;
            this.mode = ((Boolean)(Properties.getIdMode().get(new Integer(Properties.GPS_READER)))).booleanValue();

            try{
                //Ok... The file system root might be "/" or "C:/". Find out which one.
                Enumeration rootEnum = FileSystemRegistry.listRoots();
                String root = "";
                while (rootEnum.hasMoreElements()) {
                    String r = (String) rootEnum.nextElement();
                    if(r.equalsIgnoreCase("e:/")){
                        root = r;
                        break;
                    }if(r.equals("c:/"))
                        root = r;
                }

                String url = "file:///"+root+fileName;
                if(this.logger != null) this.logger.logMessage("[LT] Using url: " + url);

                //Open the file
                this.fc = (FileConnection)Connector.open(url);
                //Ok... Now we have a connection, but the file might not exists (doh!). Check the existence and create a file if needed
                if(!this.fc.exists())
                    this.fc.create();
                //Yikes... There isn't a write method that appends data to the end of the file, so we need to open the
                //outputstream to a specific location inside the file
                long size = this.fc.fileSize();
                if(size == -1)
                    this.out = this.fc.openOutputStream();
                else
                    this.out = this.fc.openOutputStream(size);

                if(out == null){
                    if(this.logger != null) logger.logMessage("[LT]Output stream is null.");
                }else
                    this.streamOpen = true;
            }catch (IOException e){
                if(this.logger != null) logger.logMessage("[LT]Could not open the output file: " + this.fileName + "\n" + e.toString());
            }
            
            this.bb.subscribeConsumer(this);          
        }catch (Throwable t){
            if(this.logger != null) logger.logMessage("[LT] Throwable (constructor): " + t);
        }
    }

    private void sendConfigurations(){
        try{
            int sensors = 0;
            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            for (int i=0; i<this.newProducers.size();i++){
                Customer cust = (Customer)newProducers.elementAt(i);
                Hashtable table = new Hashtable();
                CustomerInfo ci = cust.getInfo();
                try {
                    bout.write((ci.getAddress()+";").getBytes());
                    bout.write((ci.getName()+";").getBytes());
                    bout.write((ci.getType()+";").getBytes());
                    bout.write((this.configuredReaders + ";").getBytes());

                    String[] params = ci.getParameterNames();

                    bout.write((params.length + ";").getBytes());

                    for(int j = 0; j<params.length; j++){
                        table.put(params[j], new Integer(j));
                        bout.write((params[j]+";").getBytes());
                    }
                    this.customerID.put(cust, new Integer(this.configuredReaders));
                    this.configuredReaders++;
                    sensors++;
                    this.customerParameters.put(cust, table);
                } catch (IOException e) {

                }
                this.newProducers.removeElement(cust);
                this.allProducers.addElement(cust);

                //Set reader status as ok, no failure time and data batch number as 0
                this.readerStatus.addElement(new Integer(1));
                this.readerFailureTime.addElement(new Long(0));
                this.readerDataBatchNumber.addElement(new Integer(1));

            }
            //Write the sensor configurations
            String str = sensors + ";" + bout.toString() + "\n";
            if(this.streamOpen){ 
                try{
                    this.out.write(("B;" + str.length() + ";" + str).getBytes());
                    this.out.flush();
                }catch(IOException ioe){
                    if(this.logger != null) logger.logMessage("[LT] Could not write to the file: " + ioe.toString());
                }
            }
        }catch (Throwable t){
            if(this.logger != null) logger.logMessage("[LT] Throwable (configuration): " + t);
        }
    }

    public synchronized void run(){
        try{

            if (this.logger != null) logger.logMessage("[LT] " +"LocalTransmitter plugin started.");
            long time1 = System.currentTimeMillis(), time2, waitTime = maxSampleTimeout;
            //long lastCheck = 0;
            ByteArrayOutputStream out1 = new ByteArrayOutputStream();
            ByteArrayOutputStream out2 = new ByteArrayOutputStream();

            String str = System.currentTimeMillis() + ";" + this.dgat.getFriendlyName() + ";" + this.dgat.getBluetoothAddress() + ";" + this.bet.getUserName() + ";\n";

            if(this.streamOpen){ 
                try{
                    this.out.write(("A;" + str.length() + ";" + str).getBytes());
                    this.out.flush();
                }catch(IOException ioe){
                    if(this.logger != null) logger.logMessage("[LT] Could not write to the file: " + ioe.toString());
                }
            }

            if (this.init) {
                try {
                    this.wait();
                } catch (InterruptedException e) {
                    // woken up, got readers
                }
            }


            while(isRunning){

                //Do we need to sleep for a while? You don't want to send too much data...
                if ((time2 = System.currentTimeMillis()) - time1 < waitTime){
                    try {
                        long timeToWait = waitTime - (time2-time1);
                        if(timeToWait > maxSampleTimeout)
                            timeToWait = this.maxSampleTimeout;
                        this.wait(timeToWait);
                    } catch (InterruptedException e) {
                        // Woken up, nothing to do...
                    }
                }

                //Close method might have been called while the thread was sleeping
                if (!isRunning)
                    return;

                //The transmitter is up and running. Check first if new Readers have been found.
                if (this.newProducers.size() > 0){
                    //There are new readers. Construct sensor configuration messages and send them.
                    sendConfigurations();
                }

                if (allProducers.size() > 0){

                    int dataSize = 0;

                    long startTime = System.currentTimeMillis();
                    //Loop throug all the Reader objects and construct data messages.
                    Enumeration enumer = customerID.keys();
                    int sensorId, paramId, blocks = 0;
                    out1 = new ByteArrayOutputStream();
                    out2 = new ByteArrayOutputStream();

                    /*
                     * Write each sensor block to out1 ByteArrayStream!
                     */
                    String time = System.currentTimeMillis() + "";

                    while(enumer.hasMoreElements()){
                        Customer cust = (Customer)enumer.nextElement();
                        if (!cust.getInfo().isAlive())
                            continue;
                        
                        CustomerInfo ci = cust.getInfo();

                        //Get the parameter mappings for the Reader
                        Hashtable parameterID = (Hashtable)this.customerParameters.get(cust);

                        int type = ci.getType();
                        switch(type){

                        //Readers that are set to output all the available data...
                        case Properties.ECG_ACC_BYTE_READER:
                        case Properties.ICUBEX_READER:
                        case Properties.MUSIC_LOCAL_READER:
                        case Properties.LOCAL_READER:
                            continue;                           
                        default:
                            if(!mode && (type == Properties.LOCAL_GSM_READER || type == Properties.GSM_READER || type == Properties.GPS_READER))
                                continue;

                        blocks++;
                        String[] params = ci.getParameterNames();
                        sensorId = ((Integer)this.customerID.get(cust)).intValue();
                        try{
                            //write the sensor ID
                            intToBytes(sensorId, 2, out1);
                            intToBytes(time.length(), 1, out1);
                            out1.write(time.getBytes());
                            //write the number of parameters
                            intToBytes(params.length, 3, out1);
                        }catch(IOException e){

                        }
                        
                        CustomerData cd = this.bb.getData(cust);
                        Vector[] data = cd.getData();
                        int[] types = cd.getParameterTypes();

                        for (int i = 0; i < params.length; i++){
                            paramId = ((Integer)parameterID.get(params[i])).intValue();
                            //Write the parameter's ID
                            intToBytes(paramId, 1, out1);
                            //Write the parameter's length

                                String value = "";
                                if (types[i] == 0) {
                                    value = (String) data[i].elementAt(0);
                                } else if (types[i] == 1) {
                                    value += (Integer) data[i].elementAt(0);
                                } else if (types[i] == 2) {
                                    value += (Double) data[i].elementAt(0);
                                } else if (types[i] == 3) {
                                    value += (Float) data[i].elementAt(0);
                                }

                            //Write the parameter
                            intToBytes(value.length(), 2, out1);
                            try {
                                out1.write(value.getBytes());
                            } catch (IOException e) {
                                // This should not happen
                            }                               
                        }
                        break;
                        } //End of switch

                    } //End of while

                    if (blocks == 0 && streamData.size() == 0){
                        time1 = System.currentTimeMillis();
                        continue;
                    }

                    // Now we are sure that we will send something.
                    transmissionNumber++;


                    //Write the command byte
                    int length = 4;
                    byte[] data = null, data2 = null;
                    if(blocks > 0){
                        data = out1.toByteArray();
                    }

                    String tempStr = "";
                    int streamBlocks = streamData.size();


                    for(int i=0; i<streamBlocks; i++){
                        tempStr += (String)streamData.elementAt(i);
                    }

                    this.streamData = new Vector();

                    ByteArrayOutputStream bout = new ByteArrayOutputStream();

                    data2 = tempStr.getBytes();
                    int allBlocks = blocks;

                    if (data2.length > 0)
                        allBlocks += streamBlocks;

                    intToBytes(allBlocks, 2, bout);

                    try{
                        if(data != null)
                            bout.write(data);
                        if(data2 != null && data2.length > 0){
                            bout.write(data2);
                        }
                        bout.write("\n".getBytes());

                        data = bout.toByteArray();
                        bout = new ByteArrayOutputStream();

                        bout.write("C".getBytes());
                        intToBytes(data.length, 3, bout);
                        bout.write(data);
                    }catch(IOException ioe){

                    }

                    if(this.streamOpen){
                        try{
                            this.out.write(bout.toByteArray());
                            this.out.flush();
                        }catch(IOException ioe){
                            if(this.logger != null) logger.logMessage("[LT] Could not write to the file: " + ioe.toString());
                        }
                    }

                    long endTime = System.currentTimeMillis();
                    long transmissionTime = endTime - startTime;
                    if (this.ldr != null){
                        ldr.sendTransmissionParams(transmissionNumber, transmissionTime, dataSize);
                    }
                }
                time1 = System.currentTimeMillis();

            }
        }catch (Throwable t){
            if(this.logger != null) logger.logMessage("[LT] Throwable (run): " + t);
        }
    }

    public void updateProducerList(Blackboard bb, Vector producers) {
        try{
            for (int i = 0; i < producers.size(); i++){
                Customer cust = (Customer)producers.elementAt(i);
                
                if(!cust.getInfo().isInternal())
                    continue;
                if (allProducers.contains(cust))
                    continue;
                if(newProducers.contains(cust))
                    continue;

                newProducers.addElement(cust);
                int type = cust.getInfo().getType();

                if(type == Properties.MUSIC_LOCAL_READER || type == Properties.ICUBEX_READER || type == Properties.ECG_ACC_BYTE_READER){
                    this.bb.subscribeToProducer(this, cust);
                    this.subscriptions.addElement(cust);
                }

                //If GPS Reader doesn't stream...
                if(!mode){
                    if(type == Properties.GPS_READER){
                        this.bb.subscribeToProducer(this, cust);
                        this.subscriptions.addElement(cust);
                    }
                    if(type == Properties.GSM_READER){
                        this.gsmReader = (GSMReader)cust;
                        this.bb.subscribeToProducer(this, cust);
                        this.subscriptions.addElement(cust);
                    }
                    if(type == Properties.LOCAL_GSM_READER){
                        this.gsmReader2 = (LocalGSMReader)cust;
                        this.bb.subscribeToProducer(this, cust);
                        this.subscriptions.addElement(cust);
                    }
                    if(type == Properties.LOCAL_READER){
                        this.bb.subscribeToProducer(this, cust);
                        this.subscriptions.addElement(cust);
                    }
                }

            }
            if (this.logger != null) logger.logMessage("[DT] " +"Producers received. Number of Producers: " +producers.size());
            if(this.init){
                synchronized(this){
                    init = false;
                    this.notify();
                }
            }
        }catch (Throwable t){
            if(this.logger != null) logger.logMessage("[LT] Throwable (device update): " + t);
        }
    }
 
    public void newData(Customer prod, CustomerData pd) {
        Vector[] data = pd.getData();
        CustomerInfo ci = prod.getInfo();
        try{
            if(ci.getType() == Properties.GSM_READER){
                try{
                    if(data[1] != null && data[1].size() > 0){
                        int newId = Integer.parseInt(((String)data[1].elementAt(0)));
                        if(newId == this.oldId)
                            return;
                        else 
                            this.oldId = newId;
                    }
                }catch (Exception e){
                    return;
                }
            }else if(ci.getType() == Properties.LOCAL_GSM_READER){
                try{
                    if(data[11] != null && data[11].size() > 0){
                        int newId = Integer.parseInt(((String)data[11].elementAt(0)));
                        if(newId == this.oldId)
                            return;
                        else 
                            this.oldId = newId;
                    }
                }catch (Exception e){
                    return;
                }
            }
            int[] types = ci.getParameterTypes();
            String[] params = ci.getParameterNames();

            Object o = this.customerID.get(prod);
            if(o == null)
                return;
            int id = ((Integer)o).intValue();
            Hashtable paramMappings = (Hashtable)this.customerParameters.get(prod);

            ByteArrayOutputStream message = new ByteArrayOutputStream();
            int paramCount = 0;

            for (int i=0; i < params.length && i < data.length; i++){
                Vector v = data[i];
                int paramId = ((Integer)paramMappings.get(params[i])).intValue();
                int type = types[i];
                for (int j = 0; j < v.size(); j++){
                    o = v.elementAt(j);
                    String temp = null;
                    if(o == null)
                        continue;

                    if(type == 0)
                        temp = (String)o;
                    else if (type == 1)
                        temp = "" + ((Integer)o).intValue();
                    else if (type == 2)
                        temp = "" + ((Double)o).doubleValue();
                    else if (type == 3)
                        temp = "" + ((Float)o).floatValue();

                    try{
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    }catch (IOException e){

                    }
                    paramCount++;
                }
            }

            ByteArrayOutputStream byteout = new ByteArrayOutputStream();
            try{
                intToBytes(id, 2, byteout);
                String time = "" + System.currentTimeMillis();
                intToBytes(time.length(), 1, byteout);
                byteout.write(time.getBytes());
                intToBytes(paramCount, 3, byteout);
                byteout.write(message.toByteArray());
            }catch (IOException e){

            }
            if(byteout.size() > 0)
                this.streamData.addElement(byteout.toString());
        }catch (Throwable t){
            if(this.logger != null) logger.logMessage("[LT] Throwable (customer update): " + t);
        }
    }

    public void close(){
        this.isRunning = false;
        try{
            this.out.flush();
            this.streamOpen = false;
            this.out.close();
            this.fc.close();
        }catch (IOException ioe){
            if(this.logger != null) logger.logMessage("[LT] Could not close the file: " + ioe.toString());
        }
    }

    /* A general method for finding Readers. Problems: multiple devices,
     * multiple devices in the same class, etc...
     * Figure out how to use this if you want the transmitter to handle
     * multiple devices
     * */

    private void intToBytes(int number, int size, ByteArrayOutputStream out){
        byte[] array = new byte[size];

        for (int i = size-1; i >= 0; i--){
            array[i] = (byte)(number % 256);
            number = number / 256;
        }
        try{
            out.write(array);
        }catch (IOException e){
            // TODO
        }
    }

    private void longToBytes(long number, int size, ByteArrayOutputStream out){
        byte[] array = new byte[size];

        for (int i = size-1; i >= 0; i--){
            array[i] = (byte)(number % 256);
            number = number / 256;
        }
        try{
            out.write(array);
        }catch (IOException e){
            // TODO
        }
    }

    public CustomerInfo getInfo() {
        // TODO Auto-generated method stub
        return null;
    }

    public boolean requestData() {
        // TODO Auto-generated method stub
        return false;
    }

    public boolean setProducerMode(int mode, long timeout) {
        // TODO Auto-generated method stub
        return false;
    }

    public void updateCustomerInfo(Blackboard bb, Customer c, CustomerInfo pi) {
        // TODO Auto-generated method stub
        
    }

	public void newData(Customer prod, CustomerData pd, EventData eventData){
		// TODO Auto-generated method stub
		
	}
	
	public void markParameters(){
    	//Not needed for the class...
    }


}
