/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * http://betelgeuse.hiit.fi/
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz, Mikael Anderson
 * and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.plugin.transmitter;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
//import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.io.Connection;
import javax.microedition.io.Connector;
import javax.microedition.io.SocketConnection;
import javax.microedition.io.file.FileConnection;
import javax.microedition.io.file.FileSystemRegistry;

import hiit.BetelgeuseMIDP;
import hiit.blackboard.Blackboard;
import hiit.blackboard.BlackboardFactory;
import hiit.blackboard.Customer;
import hiit.blackboard.CustomerData;
import hiit.blackboard.CustomerInfo;
import hiit.datagatherer.DataGatherer;
import hiit.logger.Logger;
import hiit.mapper.Properties;
import hiit.observer.ConnectorObserver;
import hiit.observer.Observer;
import hiit.blackboard.event.EventData;
import hiit.reader.GSMReader;
import hiit.reader.LocalDeviceReader;
import hiit.reader.LocalGSMReader;
import hiit.reader.Reader;
import hiit.util.FileUtil;
import hiit.blackboard.event.GPSEvent;
import hiit.blackboard.filter.CombinatorFilter;
import hiit.blackboard.filter.DelayFilter;
import hiit.blackboard.filter.Filter;
import hiit.blackboard.filter.CompareFilter;
import hiit.plugin.transmitter.DebugOutputStream;

/**
 * LocalTransmitter is used to save data from sensors to a file. Three differet
 * kinds of messages are used: description of the platform, sensor and parameter
 * description and data message. Each message ends with \n.
 * 
 * Format of the messages:
 * 
 * Platform description (A) is writen each time new LocalTransmitter is created.
 * The format of the platform description message is:
 * 
 * CMD ; Length ; Timestamp ; Bluetooth name ; Bluetooth address ; \n
 * 
 * Fields:
 * <ul><li>CMD - "A"</li>
 * <li>Length - the length of the rest of the message</li>
 * <li>Timestamp - normal Java timestamp</li>
 * <li>Bluetooth name - Bluetooth name of the platform</li>
 * <li>Bluetooth address - Bluetooth address of the platform</li></ul>
 * 
 * Example message: "A;000123...1345;Test phone;001122334455;\n"
 * 
 * Sensor and parameter description (B) message is writen each time new sensors and
 * parametes need to be configured. Sensors and individual parameters are assigned ID
 * numbers which are used in the data messages to identify the source of data. The
 * format of the sensor and parameter description message is:
 * 
 * CMD ; Length ; Timestamp ; Number of sensors ; BT Address of sensor ; Name of sensor ; Type of sensor ; Sensor ID ; Number of parameters ; Param ; Param ; Param ; ... ; BT address of sensor ; ...  ; \n
 * 
 * Fields:
 * <ul><li>CMD - "B"</li>
 * <li>Length - the length of the rest of the message</li>
 * <li>Timestamp - normal Java timestamp</li>
 * <li>Number of sensors - number of sensor configurations in this message</li>
 * <li>BT Address of sensor - Bluetooth address of the sensor</li>
 * <li>Name of sensor - Bluetooth or other name assigned to the sensor</li>
 * <li>Type of sensor - Type identification number assigned to the sensor in BeTelGeuse</li>
 * <li>Sensor ID - Sensor ID number assigned to the sensor</li>
 * <li>Number of parameters - Number of parameters the sensor has and that are named in the following fields</li>
 * <li>Param - Parameter name</li></ul>
 * 
 * Identification numbers are used also for parameters in data messages. Parameter 
 * ID 0 corresponds to the first configured / named parameter of the sensor.
 * Parameter ID 1 corresponds to the second configured / named parameter of the sensor.
 * 
 * Example message: "B;length;000123...1345;1;0123456789AB;BlueGPS;256;4;2;Longitude;Latitude;\n
 * 
 * The example message configures one sensor which bluetooth address is 0123456789AB and the name BlueGPS.
 * The sensor is a GPS device (256 is assigned to GPS devices). The sensor ID assigned to it is 4.
 * It has two parameters: logitude and latitude. Longitude has a parameter ID 0 and Latitude has parameter ID 1.
 * Latitude parameter can be referenced in the data message with values 4,1 (sensor, parameter).
 * 
 * Data message (C) is used to save a set of data to the file. The format of the
 * data message is as the data messages in the BeTelGeuse data transfer protocol (see the website).
 * 
 * 
 * @author Joonas Kukkonen
 *
 */

public class LocalTransmitter implements Runnable, Customer {

	private String fileName;
	private Logger logger;
	private DataGatherer dgat;
	
	//How many "transmissions" have been done; how many times has a data message been writen to the file
	private int transmissionNumber = 0;

	//Stream blocks
	private Vector streamData = new Vector();
	//Readers which this class has subscribed to
	private Vector subscriptions = new Vector();

	//Number of configured readers (also the number for next free reader ID)
	private int configuredProducers = 0;

	//Local device reader which is updated by the transmitter
	private LocalDeviceReader ldr = null;

	//Is the thread running? Set to false if you want to kill the thread
	private boolean isRunning = true;
	//Is output stream opened to a file? (Handy if you want to kill the stream in another method than run)
	private boolean streamOpen = false;
	
	//How often should the transmitter write data to the file?
	private long maxSampleTimeout = 10000;
	
	
	private Vector customerStatus = new Vector();
	private Vector customerDataBatchNumber = new Vector();
	private Vector customerFailureTime = new Vector();
	private OutputStream out;

	private String userName = "";

	private FileConnection fc = null;
	private BetelgeuseMIDP bet;

	private boolean init = true;
	private boolean mode = true;
	private Customer gsmCustomer = null;
	private int oldId = 0;
	
	
	private Blackboard bb;
	private Vector newProducers = new Vector();
	private Vector allProducers = new Vector();

	private Vector ignoredProducerTypes = new Vector();    // updates from these are handled with events
	private Vector eventsToSave = new Vector();            // events waiting to be saved 
	
	private Hashtable customerID = new Hashtable();
	private Hashtable customerParameters = new Hashtable();
	
	private long timeDifference = 0;
	private boolean syncTime = false;
	
	private Customer gsmReader = null;
	private Customer gpsReader = null;
	private boolean filterInited = false;
	
	public LocalTransmitter(String fileName, BetelgeuseMIDP bet, DataGatherer dgat, Logger logger, boolean sync){
		try{
		    this.syncTime = sync;
			this.fileName = fileName;
			this.logger = logger;
			this.dgat = dgat;
			this.bet = bet;
			this.mode = ((Boolean)(Properties.getIdMode().get(new Integer(Properties.GPS_READER)))).booleanValue();

			
			try {
				String url = FileUtil.composeFilePath(fileName); //"file:///"+root+fileName;
				if(this.logger != null) this.logger.logMessage("[LT] Using url: " + url);
				
				//Open the file
				this.fc = (FileConnection)Connector.open(url);
				//Ok... Now we have a connection, but the file might not exists (doh!). Check the existence and create a file if needed
				if(!this.fc.exists())
					this.fc.create();
				//Yikes... There isn't a write method that appends data to the end of the file, so we need to open the
				//outputstream to a specific location inside the file
				long size = this.fc.fileSize();
				if(size == -1)
					this.out = this.fc.openOutputStream();
				else
					this.out = this.fc.openOutputStream(size);

				if(out == null){
					if(this.logger != null) logger.logMessage("[LT]Output stream is null.");
				}else
					this.streamOpen = true;
			}catch (IOException e){
				if(this.logger != null) logger.logMessage("[LT]Could not open the output file: " + this.fileName + "\n" + e.toString());
			} 
		} catch (Throwable t){
			if (this.logger != null) logger.logMessage("[LT] Throwable (constructor): " + t);
		}
		
		 // handle gps/gsm with events
		// this.ignoredProducerTypes.addElement(new Integer(Properties.GPS_READER));
		// this.ignoredProducerTypes.addElement(new Integer(Properties.LOCAL_GSM_READER));

		this.bb = BlackboardFactory.getBlackboard(logger);
        this.bb.subscribeConsumer(this);
        
	}

	private void sendConfigurations(){
		try{
			int sensors = 0;
			ByteArrayOutputStream bout = new ByteArrayOutputStream();
			for (int i=0; i<this.newProducers.size();i++){
				Customer cust = (Customer)newProducers.elementAt(i);
				Hashtable table = new Hashtable();
				CustomerInfo info = cust.getInfo();
				
				try {
				    bout.write((info.getAddress()+";").getBytes());
					bout.write((info.getName()+";").getBytes());
					bout.write((info.getType()+";").getBytes());
					bout.write((this.configuredProducers + ";").getBytes());

					String[] params = info.getParameterNames();

					bout.write((params.length + ";").getBytes());

					for(int j = 0; j<params.length; j++){
						table.put(params[j], new Integer(j));
						bout.write((params[j]+";").getBytes());
					}
					this.customerID.put(cust, new Integer(this.configuredProducers));
					this.configuredProducers++;
					sensors++;
					this.customerParameters.put(cust, table);
				} catch (IOException e) {

				}

				this.newProducers.removeElement(cust);
				this.allProducers.addElement(cust);

				//Set reader status as ok, no failure time and data batch number as 0
				this.customerStatus.addElement(new Integer(1));
				this.customerFailureTime.addElement(new Long(0));
				this.customerDataBatchNumber.addElement(new Integer(1));
				
				int type = info.getType();
				
				if (type == Properties.ECG_ACC_BYTE_READER
				        || type == Properties.ICUBEX_READER 
				        || type == Properties.CALL_READER 
				        || type == Properties.GPS_READER 
				        || type == Properties.LOCAL_GSM_READER) {
				    if (!this.subscriptions.contains(cust) 
				    		&& !this.ignoredProducerTypes.contains(new Integer(cust.getInfo().getComponentType()))) {
				        if (logger != null)
				            logger.logMessage("[BT] Subscribing to: "
				                    + type);
				        // Get all of the data from the ECG reader...
				        this.bb.subscribeToProducer(this, cust);
				        this.subscriptions.addElement(cust);
				    }
				}
				if (!this.ignoredProducerTypes.contains(new Integer(cust.getInfo().getType()))){
	                if(!mode && !this.subscriptions.contains(cust)){
	                    if(type == Properties.GPS_READER){
	                        this.bb.subscribeToProducer(this, cust);
	                        this.subscriptions.addElement(cust);
	                    }
	                    if(type == Properties.GSM_READER){
	                        this.gsmCustomer = cust;
	                        this.bb.subscribeToProducer(this, cust);
	                        this.subscriptions.addElement(cust);
	                    }
	                    if(type == Properties.LOCAL_READER){
	                        this.bb.subscribeToProducer(this, cust);
	                        this.subscriptions.addElement(cust);
	                    }
	                }
				}

			}
			//Write the sensor configurations
			String str = sensors + ";" + bout.toString() + "\n";
			if(this.streamOpen){ 
				try{
					this.out.write(("B;" + str.length() + ";" + str).getBytes());
					this.out.flush();
				}catch(IOException ioe){
					if(this.logger != null) logger.logMessage("[LT] Could not write to the file: " + ioe.toString());
				}
			}
		} catch (Throwable t){
			if(this.logger != null) logger.logMessage("[LT] Throwable (configuration): " + t);
			t.printStackTrace();
		}
	}

	public synchronized void run(){
	    //Time synchronization with the BTG server
	    if(this.syncTime){
	        this.timeDifference = this.syncTime();	        
	    }
	    
		try{
			if (this.logger != null) logger.logMessage("[LT] " +"LocalTransmitter plugin started.");
			long time1 = System.currentTimeMillis(), time2, waitTime = maxSampleTimeout;
			//long lastCheck = 0;
			ByteArrayOutputStream out1 = new ByteArrayOutputStream();
			ByteArrayOutputStream out2 = new ByteArrayOutputStream();

			String str = (System.currentTimeMillis()-this.timeDifference) + ";" + this.dgat.getFriendlyName() + ";" + this.dgat.getBluetoothAddress() + ";" + this.bet.getUserName() + ";\n";

			if(this.streamOpen){ 
				try{
					this.out.write(("A;" + str.length() + ";" + str).getBytes());
					this.out.flush();
				}catch(IOException ioe){
					if(this.logger != null) logger.logMessage("[LT] Could not write to the file: " + ioe.toString());
				}
			}

	        if (this.init) {
	            try {
	                this.wait();
	            } catch (InterruptedException e) {
	                // woken up, got readers
	            }
	        }

			while(isRunning){
				//Do we need to sleep for a while? You don't want to send too much data...
				if ((time2 = System.currentTimeMillis()) - time1 < waitTime){
					try {
						long timeToWait = waitTime - (time2-time1);
						if(timeToWait > maxSampleTimeout)
							timeToWait = this.maxSampleTimeout;
						this.wait(timeToWait);
					} catch (InterruptedException e) {
						// Woken up, nothing to do...
					}
				}

				//Close method might have been called while the thread was sleeping
				if (!isRunning)
					return;

				//The transmitter is up and running. Check first if new Readers have been found.
				if (this.newProducers.size() > 0){
					//There are new readers. Construct sensor configuration messages and send them.
					sendConfigurations();
				}
				
				if (allProducers.size() > 0){
					if (this.ldr == null){
						Vector temp = this.dgat.getConnector().getLDR();
						if(temp != null && temp.size() > 0)
							this.ldr = (LocalDeviceReader)temp.elementAt(0);
					}

					int dataSize = 0;

					long startTime = System.currentTimeMillis();
					//Loop throug all the Reader objects and construct data messages.
					Enumeration enumer = customerID.keys();
					int sensorId, paramId, blocks = 0;
					out1 = new ByteArrayOutputStream();
					out2 = new ByteArrayOutputStream();

					/*
					 * Write each sensor block to out1 ByteArrayStream!
					 */
					String time = (System.currentTimeMillis()-this.timeDifference) + "";
					
					while(enumer.hasMoreElements()){
                        Customer cust = (Customer) enumer.nextElement();
                        CustomerInfo ci = cust.getInfo();
                        Integer custType = new Integer(ci.getComponentType());
                       
                        
                       // logger.logMessage("LocalTrans run: " + customerID.contains(cust) + ", alive " + ci.isAlive());
                        
                        if (!ci.isAlive()){
                            this.customerID.remove(cust);
                            this.allProducers.removeElement(cust);
                           
                            continue;
                        } else if (this.ignoredProducerTypes.contains(custType)){
                        	// producer handled with events, have to check for events waiting to be saved...
                        	if (!this.eventsToSave.contains(custType)){
                            	//System.out.println("LocalTransmitter: ignored producer: " + ci.getName());
                        		continue;
                        	} else {
                        		this.eventsToSave.removeElement(custType);
                        	}
                        }

                        //System.out.println("LocalTransmitter, write prod: " + ci.getName());
                        
						//Get the parameter mappings for the Reader
                        Hashtable parameterID = (Hashtable) this.customerParameters.get(cust);

						switch (ci.getType()) {

						// Readers that are set to output all the available
						// data...
						case Properties.ECG_ACC_BYTE_READER:
						case Properties.ICUBEX_READER:
						case Properties.LOCAL_GSM_READER:
						case Properties.CALL_READER:
						case Properties.GPS_READER:
							continue;
						default:
							if (!mode
									&& (ci.getType() == Properties.LOCAL_READER
											|| ci.getType() == Properties.GSM_READER || ci
											.getType() == Properties.GPS_READER))
								continue;

							CustomerData cd = this.bb.getData(cust);
							
							if (cd == null) {
								continue;
							}
							blocks++;
							String[] params = ci.getParameterNames();
							sensorId = ((Integer) this.customerID.get(cust))
									.intValue();
							try {
								// write the sensor ID
								intToBytes(sensorId, 2, out1);
								intToBytes(time.length(), 1, out1);
								out1.write(time.getBytes());
								// write the number of parameters
								intToBytes(params.length, 3, out1);
							} catch (IOException e) {

							}
						
						Vector[] data = cd.getData();
						int[] types = this.bb.getData(cust).getParameterTypes();

						for (int i = 0; i < params.length; i++){
							paramId = ((Integer)parameterID.get(params[i])).intValue();
							//Write the parameter's ID
							intToBytes(paramId, 1, out1);
							//Write the parameter's length
							
							String value = "";
                            int type = types[i];
                            if (type == 0) {
                                value = (String) data[i].elementAt(0);
                            } else if (type == 1) {
                                value += (Integer) data[i].elementAt(0);
                            } else if (type == 2) {
                                value += (Double) data[i].elementAt(0);
                            } else if (type == 3) {
                                value += (Float) data[i].elementAt(0);
                            }

                            intToBytes(value.length(), 2, out1);
                            
                            // Write the parameter
                            try {
                                out1.write(value.getBytes());
                            } catch (IOException e) {
                                // This should not happen
                            }								
						}
						break;
						} //End of switch

					} //End of while

					if (blocks == 0 && streamData.size() == 0){
						time1 = System.currentTimeMillis();
						continue;
					}

					
					//if (logger != null) logger.logMessage("LocalTtrans transmissionNum: "+ transmissionNumber);
					
					// Now we are sure that we will send something.
					transmissionNumber++;

					
					
					//Write the command byte
					int length = 4;
					byte[] data = null, data2 = null;
					if(blocks > 0){
						data = out1.toByteArray();
					}

					String tempStr = "";
					int streamBlocks = streamData.size();


					for(int i=0; i<streamBlocks; i++){
						tempStr += (String)streamData.elementAt(i);
					}

					this.streamData = new Vector();

					ByteArrayOutputStream bout = new ByteArrayOutputStream();

					data2 = tempStr.getBytes();
					int allBlocks = blocks;

					if (data2.length > 0)
						allBlocks += streamBlocks;

					intToBytes(allBlocks, 2, bout);

					try{
						if(data != null)
							bout.write(data);
						if(data2 != null && data2.length > 0){
							bout.write(data2);
						}
						bout.write("\n".getBytes());

						data = bout.toByteArray();
						bout = new ByteArrayOutputStream();

						bout.write("C".getBytes());
						intToBytes(data.length, 3, bout);
						bout.write(data);
					}catch(IOException ioe){
						logger.logMessage("[LT] write error: " + ioe);
					}

					if(this.streamOpen){
						try{
							this.out.write(bout.toByteArray());
							this.out.flush();
						}catch(IOException ioe){
							if(this.logger != null) logger.logMessage("[LT] Could not write to the file: " + ioe.toString());
						}
					}
					
					
					long endTime = System.currentTimeMillis();
					long transmissionTime = endTime - startTime;
					if (this.ldr != null){
						ldr.sendTransmissionParams(transmissionNumber, transmissionTime, dataSize);
					}
				}
				time1 = System.currentTimeMillis();

			}
		}catch (Throwable t){
			if(this.logger != null) logger.logMessage("[LT] Throwable (run): " + t);
		}
	}

	public void close(){
		this.isRunning = false;

		this.bb.unsubscribeConsumer(this);
		
		try{
			this.out.flush();
			this.streamOpen = false;
			this.out.close();
			if (this.fc != null)
				this.fc.close();
		}catch (IOException ioe){
			if(this.logger != null) logger.logMessage("[LT] Could not close the file: " + ioe.toString());
		}
	}

	/* A general method for finding Readers. Problems: multiple devices,
	 * multiple devices in the same class, etc...
	 * Figure out how to use this if you want the transmitter to handle
	 * multiple devices
	 * */

	private void intToBytes(int number, int size, ByteArrayOutputStream out){
		byte[] array = new byte[size];

		for (int i = size-1; i >= 0; i--){
			array[i] = (byte)(number % 256);
			number = number / 256;
		}
		try{
			out.write(array);
		}catch (IOException e){
			// TODO
		}
	}

	private void longToBytes(long number, int size, ByteArrayOutputStream out){
		byte[] array = new byte[size];

		for (int i = size-1; i >= 0; i--){
			array[i] = (byte)(number % 256);
			number = number / 256;
		}
		try{
			out.write(array);
		}catch (IOException e){
			// TODO
		}
	}

    public CustomerInfo getInfo() {
        return null;
    }

    public void newData(Customer prod, CustomerData pd) {
    	//if (logger != null) logger.logMessage("LocalTrans newData raw: " + prod.getInfo().getName());
    	
    	/*
        if(prod.getInfo().getType() == Properties.LOCAL_GSM_READER){
            try{
                if(pd.getData()[11] != null && pd.getData()[11].size() > 0){
                    int newId = Integer.parseInt(((String)pd.getData()[11].elementAt(0)));
                    if(newId == this.oldId)
                        return;
                    else 
                        this.oldId = newId;
                }
            }catch (Exception e){
                return;
            }
        }*/
        String time = "" + (pd.getTimestamp() - this.timeDifference);
        int[] types = pd.getParameterTypes();
        String[] params = pd.getParameterNames();
        
        Object o = this.customerID.get(prod);
        
        if (o == null)
            return;
        int id = ((Integer) o).intValue();
        Hashtable paramMappings = (Hashtable) this.customerParameters.get(prod);
        
        ByteArrayOutputStream message = new ByteArrayOutputStream();
        int paramCount = 0;
        Vector[] data = pd.getData();
        
        for (int i = 0; i < params.length && i < data.length; i++) {
            Vector v = data[i];
            int paramId = ((Integer) paramMappings.get(params[i])).intValue();
            int type = types[i];
            
            if (type == 0) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = (String) o;
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {
                    }
                    paramCount++;
                }
            } else if (type == 1) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = "" + ((Integer) o).intValue();
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {
                    }
                    paramCount++;
                }
            } else if (type == 2) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = "" + ((Double) o).doubleValue();
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {
                    }
                    paramCount++;
                }
            } else if (type == 3) {
                for (int j = 0; j < v.size(); j++) {
                    o = v.elementAt(j);
                    String temp = null;
                    if (o == null)
                        continue;
                    temp = "" + ((Float) o).floatValue();
                    try {
                        intToBytes(paramId, 1, message);
                        intToBytes(temp.length(), 2, message);
                        message.write(temp.getBytes());
                    } catch (IOException e) {
                    }
                    paramCount++;
                }
            }
        }
                
        ByteArrayOutputStream byteout = new ByteArrayOutputStream();
        intToBytes(id, 2, byteout);
        intToBytes(time.length(), 1, byteout);
        try {
            byteout.write(time.getBytes());
            intToBytes(paramCount, 3, byteout);
            byteout.write(message.toByteArray());
        } catch (IOException e) {
            
        }
        if(byteout.size() > 0)
            this.streamData.addElement(byteout.toString());
        
    }
    long lastUpdate = 0;
    
	public void newData(Customer prod, CustomerData pd, EventData eventData){
    	Integer prodType = new Integer(prod.getInfo().getComponentType());
    	
    	if (!this.eventsToSave.contains(prodType))
    			this.eventsToSave.addElement(prodType); 
    	this.newData(prod, pd);
	}
	

    public boolean requestData() {
        return false;
    }
    
    public void setCustomerInfo(CustomerInfo info) {
    }
    
    public boolean setProducerMode(int mode, long timeout) {
        return false;
    }
    
    public void updateCustomerInfo(Blackboard bb, Customer c, CustomerInfo pi) {
        this.newProducers.addElement(c);
    }
    
    public void updateProducerList(Blackboard bb, Vector producers) {
    	
        for (int i = 0; i < producers.size(); i++) {
            Customer cust = (Customer) producers.elementAt(i);
            
            // logger.logMessage("updateprodlist: " + cust + ", " + cust.getInfo().getComponentType() + ", " + ignoredProducerTypes.contains(new Integer(cust.getInfo().getComponentType())) );
            if (!allProducers.contains(cust) && (!cust.getInfo().isInternal())){ 
                allProducers.addElement(cust);
                newProducers.addElement(cust);
                
                // handle producer with events:
                if (ignoredProducerTypes.contains(new Integer(cust.getInfo().getComponentType()))) {
                	switch (cust.getInfo().getComponentType()){
                	case Properties.GPS_READER:
                		this.gpsReader = cust;
                		break;
                		
                	case Properties.LOCAL_GSM_READER:
                		this.gsmReader = cust;
                		break;
                	}
                	if (gpsReader != null)
                	//if (gpsReader != null && gsmReader != null)
                		prepareEventFilter();
                }
            }
        }
        if (this.logger != null)
            logger.logMessage("[BT] "
                    + "Data sources received. Number of data sources: "
                    + this.allProducers.size());
        if (this.init) {
            synchronized (this) {
                this.notify();
                init = false;
            }
        }
    }
    
    public long syncTime(){
        if (this.logger != null) logger.logMessage("[LT] " +"Synchronizing time with the server.");
        
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        ByteArrayOutputStream byteOut2 = new ByteArrayOutputStream();
        
        try {
            intToBytes(0, 2, byteOut2);
            intToBytes(0, 1, byteOut2);
            byteOut2.write(0);
            byteOut2.write(this.dgat.getBluetoothAddress().getBytes());
            if (this.dgat.getFriendlyName() != null){
                intToBytes(this.dgat.getFriendlyName().length(), 1, byteOut2);
                byteOut2.write(this.dgat.getFriendlyName().getBytes());
            }else
                byteOut2.write(0);
            if (this.bet.getUserName() != null){
                intToBytes(this.bet.getUserName().length(), 1, byteOut2);
                byteOut2.write(this.bet.getUserName().getBytes());
            }else
                byteOut2.write(0);
            if (this.bet.getPassword() != null){
                intToBytes(this.bet.getPassword().length(), 1, byteOut2);
                byteOut2.write(this.bet.getPassword().getBytes());
            }else
                byteOut2.write(0);
            
            //Write the header
            byteOut.write("C".getBytes());
            byte[] data = byteOut2.toByteArray();
            //Write the length of the message
            intToBytes(data.length, 3, byteOut);
            //Write the data
            byteOut.write(data);
        } catch (IOException e) {
            // Should not happen
        }
        SocketConnection conn = null;
        InputStream sin = null;
        OutputStream sout = null;
        
        int[] response = null;
        
        try {
            conn = (SocketConnection)Connector.open("socket://" + "universe.hiit.fi:10000", Connector.READ_WRITE, true);
            sin = conn.openInputStream();
            sout = conn.openOutputStream();
            
            sout.write(byteOut.toByteArray());
            sout.flush();
            
            int[] head = new int[4];
            //in = conn.openInputStream();
            int  length;
      
            //Read the command.
            head[0] = sin.read();
            //Read the 3 byte length field
            head[1] = sin.read();
            length = head[1] * 256 * 256;
            head[2] = sin.read();
            length += head[2] * 256;
            head[3] = sin.read();
            length += head[3];
            response = new int[length+4];         
            for (int i = 4; i < length+4; i++){
                response[i] = sin.read();
            }
            response[0] = head[0];
            response[1] = head[1];
            response[2] = head[2];
            response[3] = head[3];
            
        } catch (IOException e) {
            if(this.logger != null) logger.logMessage("[LT]Could not sync time with the BTG server.");
            return 0;
        }
   
        long current = System.currentTimeMillis();
            
        int temp = 0;
        if(response[6] != 0){
            if(this.logger!=null) logger.logMessage("[LT]Could not sync time: received an error code from the server: " + response[6]);
            return 0;
        }
        
        int dbSessionLength = response[7];
        int offset = 8+dbSessionLength;
        
        int timeLength = response[offset];
        offset++;
        
        //Meh... mobile java doesn't have String constructor for int array...
        char[] timeBytes = new char[timeLength];
        for(int i=0; i<timeLength; i++){
            timeBytes[i] = (char)response[i+offset];
        }
        
        String time = new String(timeBytes);  
        if(this.logger != null) logger.logMessage("[LT]Server time: " + time);
        long difference = current - Long.parseLong(time);
        if(this.logger != null) logger.logMessage("[LT]Time difference: " + difference);
        
        //Close the connection to the server
        byteOut = new ByteArrayOutputStream();
        try{
            byteOut.write("Q".getBytes());
            intToBytes(2, 3, byteOut);
            intToBytes(response[4]*256 + response[5], 2, byteOut);
            sout.write(byteOut.toByteArray());
            sout.flush();
            sout.close();
            sin.close();
            conn.close();
        }catch(IOException e){
            //Nothing to do here...
        }
        
        return difference;
    }


    private void prepareEventFilter(){
    	if (!filterInited){
    		filterInited = true;
    		
    		/*
        	// gps 30 m change
    		Filter distanceFilter = new GreaterThanFilter(gpsReader, this, (Object)new Double(0.03), GreaterThanFilter.TYPE_DOUBLE, true);
    		distanceFilter.setLogger(logger);
    		this.bb.subscribeToEvent(distanceFilter, gpsReader, "distanceChange");
    		*/
    		
    		
        	/*
        	// gps 30 m change & gsm cellIDChange
    		CombinatorFilter combinatorFilter = new CombinatorFilter(this, true, logger);

    		Filter distanceFilter = new GreaterThanFilter(gpsReader, combinatorFilter, (Object)new Double(0.03), GreaterThanFilter.TYPE_DOUBLE, true);
    		this.bb.subscribeToEvent(distanceFilter, gpsReader, "distanceChange");
    		
    		Filter forwardFilter = new Filter(gsmReader, combinatorFilter);
    		this.bb.subscribeToEvent(forwardFilter, gsmReader, "cellIDChange");
    		combinatorFilter.setProducers( new Customer[] {forwardFilter, distanceFilter} );
    		*/
        	
    		
    		
    		/*
    		// gsm raw data & gps raw data 10sek delay
        	CombinatorFilter combinatorFilter2 = new CombinatorFilter(this, false, logger);
    		
    		Filter gpsDelayFilter = new DelayFilter(gpsReader, combinatorFilter2, 5000);
    		this.bb.subscribeToProducer(gpsDelayFilter, gpsReader);

    		Filter gsmDelayFilter = new DelayFilter(gsmReader, combinatorFilter2, 5000);
    		this.bb.subscribeToProducer(gsmDelayFilter, gsmReader);

    		combinatorFilter2.setProducers( new Customer[] { gpsDelayFilter, gsmDelayFilter } );
    		*/
    		
    	}
    	
    	
    }
    
    public void markParameters(){
    	//Not needed for the class...
    }

}
