/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.reader;

import hiit.BetelgeuseMIDP;
import hiit.blackboard.BlackboardFactory;
import hiit.blackboard.Customer;
import hiit.blackboard.CustomerData;
import hiit.blackboard.event.EventData;
import hiit.blackboard.event.EventHandler;
import hiit.blackboard.event.GPSEvent;
import hiit.connectorUnit.BTReaderFactory;
import hiit.logger.Logger;
import hiit.mapper.Properties;
import hiit.observer.Observer;

import javax.microedition.io.Connection;

import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;



/**
 * GPSReader is a data reader for GPS Bluetooth devices.
 * @author Petteri Nurmi, Joonas Kukkonen and Eemil Lagerspetz
 *
 */

public class GPSReader extends ReaderSkeleton{
	
	private String lastLong = "";
	private String lastLat = "";
	private long coordinateChangeTime = 0;
	private int useOldCoordinate = 10000;
	private boolean gga = false;
	
	private EventHandler eventHandler;
	/**
	 * Constructs a new GPSReader. Call the initialize(..) method to set variables in the new object.
	 */
	
	public GPSReader(){
	}
	/**
	 * Creates a new GPS reader.
	 * @param connection Connection to the remote Bluetooth device.
	 */
	
	public GPSReader(Connection connection, String url, Logger logger, BTReaderFactory factory, int deviceId, boolean streamMode) {
		super(connection, url, logger, factory, deviceId, streamMode);
		this.delay = 100;

		eventHandler = new GPSEvent();
		eventHandler.setLogger(logger);
		BlackboardFactory.getBlackboard(logger).registerProducer(this, this.info, ((GPSEvent)eventHandler).getEventList());
	}
	
	/**
	 * Returns the name of the device.
	 * @return the device name
	 */
	
	public String getDeviceName() {
		return "GPS";
	}
	
	/**
	 * Initializes the Reader. Mean for cases when the Reader implementation is constructed
	 * with an empty constructor by Class.forName() method.
	 * @param connection the connection to the sensor 
	 * @param url the connection URL
	 * @param logger the Logger object
	 * @param factory the BTDeviceFactory which is used to restore connections
	 * @param deviceId the device ID of the sensor
	 * @param streamMode the mode of the Reader at startup. True for streaming mode, false for request mode.
	 */
	
	public void initialize(Connection connection, String url, Logger logger, BTReaderFactory factory, int deviceId, boolean streamMode) {
		super.initialize(connection, url, logger, factory, deviceId, streamMode);
		this.delay = 100;

		eventHandler = new GPSEvent();
		eventHandler.setLogger(logger);
		BlackboardFactory.getBlackboard(logger).registerProducer(this, this.info, eventHandler.getEventList());
	}
	
	/*
	 * Initializes the Hashtable.
	 */
	
	protected void initializeVariables() {
		this.parameters = new Hashtable(12);
		this.parameters.put("Latitude","00D 00.0000'N");
		this.parameters.put("Longitude","000D 00.0000'E");
		this.parameters.put("GMT","00:00:00");
		this.parameters.put("Altitude","0");
		this.parameters.put("Satellites","0");
		this.parameters.put("Speed","0");
		this.parameters.put("PDOP","50");
		this.parameters.put("HDOP","50");
		this.parameters.put("VDOP","50");
		this.parameters.put("Protocols","");
		this.parameters.put("Ground speed","0");
		this.parameters.put("Course","0");
	}
	
	public String[] getParameterNames() {
		
		// Avoid null pointer exceptions
		if(this.parameters==null)return null;
		
		String[] params = new String[12];
		params[0] = "Latitude";
		params[1] = "Longitude";
		params[2] = "GMT";
		params[3] = "Altitude";
		params[4] = "Satellites";
		params[5] = "Speed";
		params[6] = "Ground speed";
		params[7] = "Course";
		params[8] = "PDOP";
		params[9] = "HDOP";
		params[10] = "VDOP";
		params[11] = "Protocols";
		return params;
	}
	
	public int[] getParameterTypes(){
		int [] types = new int[12];
		types[0] = Properties.DATATYPE_STRING;
		types[1] = Properties.DATATYPE_STRING;
		types[2] = Properties.DATATYPE_STRING;
		types[3] = Properties.DATATYPE_STRING;
		types[4] = Properties.DATATYPE_STRING;
		types[5] = Properties.DATATYPE_STRING;
		types[6] = Properties.DATATYPE_STRING;
		types[7] = Properties.DATATYPE_STRING;
		types[8] = Properties.DATATYPE_STRING;
		types[9] = Properties.DATATYPE_STRING;
		types[10] = Properties.DATATYPE_STRING;
		types[11] = Properties.DATATYPE_STRING;
		return types;
	}
	
	/*
	// this is used to pass param-positions to ProducerCache.extractParametersByPosition
	public int[] getParameterPositionsByName(String names[]) {
		int positions[] = new int[names.length];
		String params[] = this.getParameterNames();
		for (int i=0; i<positions.length; i++){
			String name = names[i];
			for (int j=0; j<params.length; j++){
				String paramName = params[j];

				if (name.equalsIgnoreCase(paramName)) {
					positions[i] = j;
					break;
				}
			}
		}
		 return positions;
	}
	//
	*/
	
	/*
	 * Splits a String into tokens. Same functionality could have been reached with StringTokenizer,
	 * but MIDP does not include the class.
	 */
	
	private String[] splitString(String str, String delimiterStr) {
		String[] s;
		int len = delimiterStr.length();
		int counter = 0;
		int index = 0, index2 = -1;
		while((index=str.indexOf(delimiterStr,index))!=-1 && index<str.length()) {
			index+= len;
			counter++;
		}		
		counter++;
		s = new String[counter];
		index = 0;
		for(int i=0;i<counter;i++) {
			index = str.indexOf(delimiterStr,index);
			if(index!=-1)
				s[i] = str.substring(index2+len,index);
			else
				s[i] = str.substring(index2 + len);
			index2 = index;
			index+= len;
		}
		return s;
	}
	
	/*
	 * Reads the data from the device. 
	 */
	
	protected boolean readData() throws java.io.IOException{

		while(true){
			char[] data = new char[100];
			data[0] = 0;
			char c;
			int index = 0;

			while(this.iStream!=null) {
				c = (char)this.iStream.read();
				if(c=='\n' || index==100){					
					break;
				}
				data[index++] = c;
			}			
			if(data[0]=='$') {				
				String msg = new String(data,0,index-1);
				handleMessage(msg);
				
				//check if the message is fubar
				if(msg.lastIndexOf('$')>0)
				    continue;
				
				//Keep reading until you get an updated GPS position!
				if(!mode){
					String lon = this.getParameter("Longitude");
					String lat = this.getParameter("Latitude");
                    String sat = this.getParameter("Satellites");
                    
                    //Check if we have satellites
                    if(sat == null || sat.compareTo("") == 0 || sat.compareTo("0") == 0)
                        return true;
                    //Check that the coordinates are nonzero
                    if(lon.compareTo("") == 0 || lat.compareTo("") == 0 || (lon.compareTo("000D 00.0000'E") == 0 && lat.compareTo("00D 00.0000'N") == 0))
                        return true;
                    
                    /* This should be handled by distance filtering...
                    //Check that we have new coordinates
                    if(lon.compareTo(lastLong) == 0 && lat.compareTo(lastLat) == 0){
                        if(this.coordinateChangeTime + this.useOldCoordinate > System.currentTimeMillis())
                            continue;
                    }
                    						
					coordinateChangeTime = System.currentTimeMillis();
					lastLong = lon;
					lastLat = lat;
					*/
				}
				return true;
				
			}
			if(!mode)
				return false;
		}
	}
	
	/*
	 * Splits the message up and saves it in the Hashtable.
	 */
	
	private void handleMessage(String message) {
		String command = message.substring(3,6);
		String payload = message.substring(7);
		//String[] values = payload.split(",");	
		String[] values = splitString(payload, ",");		
		String str;
		if (command.equals("GGA")) {
			this.gga = true;
			// UTC -> GMT
			str = values[0].substring(0,2) + ":" + values[0].substring(2,4) + ":" + values[0].substring(4,6);
			this.parameters.put("GMT",str);
			
			// Latitude
			if(values[1] != null && values[1].length()>5 && values[2] != null){
				str = values[1].substring(0,2) + "D " + values[1].substring(2,4) + values[1].substring(4) + "'" +
				values[2];
				this.parameters.put("Latitude",str);
			}

			// Longitude
			if(values[3] != null && values[3].length()>6 && values[4] != null){
				str = values[3].substring(0,3) + "D " + values[3].substring(3,5) + values[3].substring(5) + "'" +
				values[4];
				this.parameters.put("Longitude",str);
			}
			
			// values[5] := Position fix
			
			// Number of satellites in view
			str = values[6];
			this.parameters.put("Satellites",str);
			
			// values[7] := HDOP
			
			if(values[7] != null && values[7].compareTo("") != 0){
				this.parameters.put("HDOP", values[7]);
			}
			
			
			// Altitude
			str = values[8] + " " + values[9];
			this.parameters.put("Altitude",str);									
			
			// values[10] := Geodoidal separation
			// values[11] := Age of last differential GPS reading
			// values[12] := Id of the satellite giving the differential GPS information
			// values[13] := Checksum											
			
		}else{
			this.gga = false;
			if(command.equals("RMB")) {

			}else if(command.equals("RMC")) {

			}else if(command.equals("VTG")) {
				if(values.length > 7){
					str = "";

					if(values[5].equalsIgnoreCase("n"))
						str = values[4]+ " knots.";
					//If values[7] is not empty and the first char is k or K....
					else if (values[7].length() > 0 && (values[7].charAt(0)=='k' || values[7].charAt(0) == 'K'))
						str = values[6] + " km/h.";	
					this.parameters.put("Speed",str);
				}

			}else if(command.equals("RMA")) {

				//Ground speed
				str = values[7] + " knots."; 
				this.parameters.put("Ground speed",str);
				this.parameters.put("Course",values[8]);

			}else if(command.equals("GSA")) {

				this.parameters.put("PDOP",values[14]);
				this.parameters.put("HDOP",values[15]);						
				int index = values[16].indexOf('*');
				if(index!=-1)
					//In older protocols, VDOP is followed by checksum (* and two dec numbers)
					this.parameters.put("VDOP",values[16].substring(0,index));
				else
					//There has been changes to last fields in some messages in newer protocol versions
					//This is here just in case...
					this.parameters.put("VDOP",values[16]);

			}else if(command.equals("GSV")) {

			}else {
				return;
			}
		}
		
		str = (String)this.parameters.get("Protocols");
		if(str.indexOf(command)==-1) {
			if(str.equals("")) 
				str+=command;
			else
				str +=", " + command;
			this.parameters.put("Protocols",str);
		}
		
	}
    
    /**
	 * Notifies all subscribers that new data is available.
	 */
	   protected void notifySubscribers() {
		   //Send an update ONLY if we have gotten a new GGA message (contains long and lat values)
		   if(this.gga){
			   //Read the current data only once
			   Vector[] v = new Vector[this.parameters.size()];
			   Vector temp = null;
			   String[] names = this.getParameterNames();

			   for (int i = 0; i<names.length; i++){
				   temp=new Vector();
				   temp.addElement(this.parameters.get(names[i]));
				   v[i] = temp;
			   }

			   CustomerData d = new CustomerData(v, new Date().getTime(), info);

			   BlackboardFactory.getBlackboard(logger).pushData(this, d, eventHandler.getEvents(d));
		   }        
	    }
	
	   
		public void newData(Customer prod, CustomerData pd, EventData eventData){
			// TODO Auto-generated method stub
			
		}

	    public void markParameters(){
	    	this.eventHandler.markParameters();
	    }

}
