/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.reader;

import hiit.blackboard.event.GSMEvent;
import hiit.connectorUnit.BTReaderFactory;
import hiit.logger.Logger;
import hiit.mapper.Properties;
import hiit.observer.ConnectorObserver;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.io.Connection;
import javax.microedition.io.Connector;
import javax.microedition.io.SocketConnection;

/**
 * GSMReader is a local device reader that reads gsm cell id, network, signla strenght, etc,
 * values from PlaceLab cell id server. If the PlaceLab server is not running
 * on the device, isDead() method returns true after the initialize(..) method has been
 * completed.
 * @author Joonas Kukkonen
 *
 */

public class GSMReader extends LocalDeviceReader implements ConnectorObserver{
	
	private SocketConnection conn = null;
	private InputStream in;
	private OutputStream out;
	private int lastCellId = -1;
	private boolean gsmMode = true;
	private Vector gpsReaders = new Vector();
	
	/**
	 * Constructor for GSMReader that parses data from the local device. 
	 * Connects to a local port to find GSM information. Works with 
	 * for example the PlaceLab GSM information server.
	 */
	
	public GSMReader() {
		this.delay = 5000;
		this.deviceId = Properties.GSM_READER;
	}
	
	public void initialize(Connection connection, String url, Logger logger, BTReaderFactory factory, int deviceId, boolean streamMode){
		super.initialize(connection, url, logger, factory, deviceId, streamMode);
		this.deviceId = Properties.GSM_READER;

		this.gsmMode = ((Boolean)(Properties.getIdMode().get(new Integer(Properties.GPS_READER)))).booleanValue();
		if(!gsmMode){
			factory.getConnector().subscribe(this);
		}
			
		try{
			conn = (SocketConnection)Connector.open(url);
			this.out = conn.openOutputStream();
			this.in = conn.openInputStream();
		} catch (IOException ie){
			this.isRunning = false;
			if (logger != null) logger.logMessage("Unable to open connection to " + url);
		}
		if(!mode){
			disconnect();
		}
	}

	public String getDeviceName() {
		return "GSM";
	}
	
	protected void initializeVariables() {
		this.parameters = new Hashtable(14);
		this.parameters.put("Activity", "");
		this.parameters.put("Cell ID","");
		this.parameters.put("Area ID", "");
		this.parameters.put("Signal strength", "");
		this.parameters.put("MCC", "");
		this.parameters.put("MNC", "");
		this.parameters.put("Network name", "");
		this.parameters.put("Discovery number", "");
		this.parameters.put("Last discovery time", "");
		this.parameters.put("Discovered devices", "");
		this.parameters.put("Transmission number", "");
		this.parameters.put("Transmission time", "");
		this.parameters.put("Transmission size", "");
		this.parameters.put("BTDevices", "");
	}
	
	public int[] getParameterTypes(){
		int[] types = new int[14];
		types[0] = Properties.DATATYPE_STRING;
		types[1] = Properties.DATATYPE_STRING;
		types[2] = Properties.DATATYPE_STRING;
		types[3] = Properties.DATATYPE_STRING;
		types[4] = Properties.DATATYPE_STRING;
		types[5] = Properties.DATATYPE_STRING;
		types[6] = Properties.DATATYPE_STRING;
		types[7] = Properties.DATATYPE_STRING;
		types[8] = Properties.DATATYPE_STRING;
		types[9] = Properties.DATATYPE_STRING;
		types[10] = Properties.DATATYPE_STRING;
		types[11] = Properties.DATATYPE_STRING;
		types[12] = Properties.DATATYPE_STRING;
		types[13] = Properties.DATATYPE_STRING;
		return types;
	}
	
	public String[] getParameterNames() {
		// Avoid null pointer exceptions
		if(this.parameters==null)return null;
		
		String[] params = new String[14];
		params[0] = "Activity";
		params[1] = "Cell ID";
		params[2] = "Area ID";
		params[3] = "Signal strength";
		params[4] = "MCC";
		params[5] = "MNC";
		params[6] = "Network name";
		params[7] = "Discovery number";
		params[8] = "Last discovery time";
		params[9] = "Discovered devices";
		params[10] = "Transmission number";
		params[11] = "Transmission time";
		params[12] = "Transmission size";
		params[13] = "BTDevices";
		return params;
	}
		
	protected synchronized boolean readData() {
		try {
			out.write(1);
			out.flush();
		} catch (IOException e) {
			if (logger != null) logger.logMessage("[GSM] Could not request GSM info from server!");
			// might still work, not returning here
		}
		StringBuffer data = new StringBuffer("");
		int length = 0;
		try {
			int ch = 0;
			while ((ch = in.read()) != -1){
				if (ch == '\n')
					break;
				data = data.append((char) ch);
				length++;
			}
		} catch (IOException e) {
			if (logger != null) logger.logMessage("[GSM] Could not read response from server!");
			return false;
		}
		if (length <= 0){
			if (logger != null) logger.logMessage("[GSM] Unexpected end of stream!");
			return false;
		}
		
		String[] names = getParameterNames();
		int validNames = 7;
		int index = 1;
		String reply = data.toString();
		String temp = null;
		while (reply != null && reply.length() > 0){
			int comma = reply.indexOf(',');
			
			if (comma < 0){
				temp = reply;
				reply = null;
			}
			else{
				temp = reply.substring(0, comma);
				reply = reply.substring(comma + 1);
			}
			this.parameters.put(names[index], temp);
			index++;
			if (index == validNames){ // All read
				break;
			}
		}
		
		if(!gsmMode){
			int newCell = Integer.parseInt((String)(this.parameters.get("Cell ID")));
			if(newCell != this.lastCellId){
				lastCellId = newCell;
				for(int i=0; i < this.gpsReaders.size(); i++){
					((Reader)this.gpsReaders.elementAt(i)).readOne();
				}
			}
		}
		
		
		return true;
	}
	
	public void close() {
		if(this.logger != null) this.logger.logMessage("[GSM] Closing GSM reader");
		this.isRunning = false;
		this.factory.getConnector().unsubscribe(this);
		try{
			out.close();
			in.close();
			conn.close();
		}catch (IOException ioe){
			// Error closing, not too deadly
		}
	}
	
	protected void disconnect(){
		try{
			out.close();
			in.close();
			conn.close();
		}catch (IOException ioe){
			// Error closing, not too deadly
		}
	}
	
	/**
	 * Connects to the data source.
	 */
	
	protected boolean connect(){
		try{
			conn = (SocketConnection)Connector.open(url);
			this.out = conn.openOutputStream();
			this.in = conn.openInputStream();
			return true;
		} catch (IOException ie){
			this.isRunning = false;
			if (logger != null) logger.logMessage("Unable to open connection to " + url);
		}
		return false;
	}
	
	public void setStreamingMode(boolean streaming){
		if(!mode && streaming){
			connect();
			this.notify();
		}
		this.mode = streaming;
	}

	public void update(Vector readers, Object connector){
		if(readers != null){
			for(int i = 0; i < readers.size(); i++){
				if(readers.elementAt(i) != null && ((Reader)readers.elementAt(i)).getDeviceId() == Properties.GPS_READER){
					if(!this.gpsReaders.contains(readers.elementAt(i)))
						this.gpsReaders.addElement(readers.elementAt(i));
				}
			}
		}
	}
}
