/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.reader;

import java.io.IOException;
import java.io.OutputStream;
import java.util.Hashtable;
import java.util.Vector;
//import java.util.Vector;

import hiit.blackboard.Customer;
import hiit.blackboard.CustomerData;
import hiit.blackboard.event.EventData;
import hiit.connectorUnit.BTReaderFactory;
import hiit.logger.Logger;
import hiit.mapper.Properties;
import hiit.observer.Observer;

import javax.bluetooth.BluetoothConnectionException;
import javax.microedition.io.Connection;
import javax.microedition.io.StreamConnection;

/**
 * ICubexWiMiniDigParser is a parser for I-CubeX Wireless miniDig Bluetooth
 * device. Up to 8 sensors can be attached to the miniDig. The miniDig sends
 * the sensor data over a Bluetooth connection to the receiver. Note that I-CubeX
 * sensor requires pairing (default pairing code is 1). This class offers methods
 * for defining which sensors are connected to the miniDig. Adjusting the read
 * delay of this class will also adjust the sample rate of the miniDig. Minimum read delay
 * for this class is 3ms (min for the miniDig is 4ms) and maximum 12 seconds (max for miniDig is
 * 16 seconds).
 * @author Joonas Kukkonen, Eemil Lagerspetz
 *
 */

public class ICubexWiMiniDigParser extends ReaderSkeleton{
	
	// Commands transmitted by the IcubeX
	// in response to programming commands
	private static final int RESET_ACK = 35;
	private static final int STATUS = 37;
	// error codes for the status message:
	// configuration setting error
	private static final int CONFIG_ERROR = 90;
	// MIDI protocol error, a DATA byte is > 127
	private static final int MIDI_ERROR = 92;
	// MDI byte has been scrambled
	private static final int SCRAMBLE_ERROR = 94;
	// buffer full, data may be lost
	private static final int BUFFER_ERROR = 95;
	
	private static final int VERSION = 71;
	private static final int MODE = 91;
	// stand-alone mode responses
	private static final int NAME = 101;
	private static final int CONFIG = 106;
	// host mode responses
	private static final int STREAM_DATA = 0;
	private static final int SAMPLE_DATA = 4;
	
	// the identifier for the start of the message 
	private static final int START = 240;
	// the identifier for the end of the message 
	private static final int END = 247;
	// manufacturer id
	private static final int MAN = 125;
	// This can be changed with SET ID (useful with many Wi-miniDigs)
	private int DEV = 0;
	
	// the mode of the WiMiniDig (host = 0, stand-alone = 1)
	private int mode = -1;
	
	// useful commands:
	private static final int SET_MODE = 90;
	private static final int RES = 2;
	private static final int RESET = 34;
	private static final int INTERVAL = 3;
	private static final int STREAM = 1;
	
	
	private static final char ACCELERATION3D = 'a';  // 3 slots
	private static final char TEMPERATURE_HUMIDITY = 't'; // 2 slots, first temperature
	private static final char LIGHT = 'l';
	private static final char ORIENTATION = 'o';
	private static final char DISTANCE = 'd';
	
	private int[] inputTypes = null;
	
	private static final int NO_INPUT = -1;
	private static final int ACCELERATION_INPUT_X = 1;
	private static final int ACCELERATION_INPUT_Y = 2;
	private static final int ACCELERATION_INPUT_Z = 3;
	private static final int HUMIDITY_INPUT = 5;
	private static final int TEMPERATURE_INPUT = 4;
	private static final int AMBIENT_LIGHT_INPUT = 6;
	private static final int ORIENTATION_INPUT = 7;
	private static final int DISTANCE_INPUT = 8;
	
	private String arrangement = null;
	private int slots_used = 0;
	private boolean configured = false;
	private boolean reset = false;
	
	
	
	private String[] parNames = null;
	boolean[] sensorAt = new boolean[8];
	boolean[] resolutions = new boolean[8];
	boolean[] onLine = new boolean[8];
	boolean on = false;
	
	private boolean hires = false;
	private OutputStream oStream = null;
	private int interval = 0;
	int bytesPerMin = 0;
	
	private int notifyDelay = 500; //How often should subscribers be notified?
	
	/**
	 * Creates a new ICubexWiMiniDigParser. Use the initialize(..) method to set variables.
	 *
	 */
	
	public ICubexWiMiniDigParser(){
	}
	
	private void setHostMode(){
		int[] setMode = {START, MAN, DEV, SET_MODE,	0, END }; // 0 = host mode
		boolean fail = false;
		for (int i = 0; i < setMode.length; ++i){
			try {
				oStream.write(setMode[i]);
				oStream.flush();
			} catch (IOException e) {
				fail = true;
			}
		}
		if (fail) {
			logger.logMessage("[ICBX] Failed to send hostmode request.");
		}
		else{
			try {
				//Read the MODE message...
				readData(MODE);
			} catch (IOException e) {
				logger.logMessage("[ICBX] Failed to read mode reply.");
			}
			if (this.mode != 0)
				if (logger != null)
					logger.logMessage("[ICBX] Failed setting mode. Mode: " + mode);
		}
	}
	
	/**
	 * Initializes the ICubexWiMiniDigParser.
	 * @param connection the conneciton object
	 * @param url the url that the conneciton object uses
	 * @param logger the Logger object used to log error messages
	 * @param factory the BTDeviceFactory object used to restore the connection in case of disconnect
	 * @param deviceId the Reader ID assigned to this Reader
	 */
	public void initialize(Connection connection, String url, Logger logger, BTReaderFactory factory, int deviceId){
		//Set delay to 200ms -> miniDig sends data every 300ms
		//Delay 13ms -> data every 20ms (should be enough for activity recognition)
		this.delay = 200;
		this.url = url;
		this.logger = logger;
		this.factory = factory;
		this.deviceId = deviceId;
		//this.bluetoothAddress = bluetoothAddress;
		this.connection = connection;
		this.isRunning = true;
		
		//HACK!
		// Initialize the hash table
		//initializeVariables();
		this.configure("aotld");
		this.giveUp = false;
	}
	
	/**
	 * Returns the name of the device.
	 * @return the device name
	 */
	public String getDeviceName() {
		return "ICBX";
	}
	
	/**
	 * Initializes the variables. However this method does not
	 * create parameters with correct names.
	 */
	
	protected void initializeVariables() {
		this.parameters = new Hashtable(8);
		for (int i = 1; i <= 8; ++i){
			this.parameters.put("sensor" + i,"");
		}
		this.streamParameters = new Hashtable(8);
		for (int i = 1; i <= 8; ++i){
			this.streamParameters.put("sensor" + i,new Vector());
		}
	}
	
	/**
	 * Initializes the parameters Hashtable with correct parameter names.
	 * The miniDig can be used with different sensor configurations,
	 * so the sensors and thus the parameter names can change.
	 *
	 */
	
	protected void initializeVariablesProperly() {
		this.parameters = new Hashtable(slots_used);
		this.streamParameters = new Hashtable(slots_used);
		this.parNames = new String[slots_used];
		this.inputTypes = new int[slots_used];
		for(int i = 0; i < slots_used; i++){
			inputTypes[i] = -1;
		}
		int slot = 0;
		String temp;
		
		for (int i = 0; i < arrangement.length(); ++i){
			char c = arrangement.charAt(i);
			switch (c){
			case ACCELERATION3D:
				temp = "Acceleration X";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = ACCELERATION_INPUT_X;
				slot++;
				temp = "Acceleration Y";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = ACCELERATION_INPUT_Y;
				slot++;
				temp = "Acceleration Z";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = ACCELERATION_INPUT_Z;
				slot++;
				break;
			case TEMPERATURE_HUMIDITY:
				temp = "Temperature";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = TEMPERATURE_INPUT;
				slot++;
				temp = "Humidity";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = HUMIDITY_INPUT;
				slot++;
				break;
			case LIGHT:
				temp = "Light";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = AMBIENT_LIGHT_INPUT;
				slot++;
				break;
			case ORIENTATION:
				temp = "Orientation";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = ORIENTATION_INPUT;
				slot++;
				break;
			case DISTANCE:
				temp = "Distance";
				this.parameters.put(temp, "");
				this.parNames[slot] = temp;
				this.streamParameters.put(temp, new Vector());
				this.inputTypes[slot] = DISTANCE_INPUT;
				slot++;
				break;
			case ' ': // no sensor in slot
				break;
			default:
				if (this.logger != null) logger.logMessage("[ICBX] Unidentified sensor: " + c + "!");
				this.configured = false;
				return;
			}
		}		
	}
	
	/**
	 * Configures the parser to the sensor configuration used. The current
	 * configuration (sensors connected to miniDig) is given as a String parameter.
	 * The configuration String can express the maximum of 8 different sensors.
	 * However some sensors take more than one slot from the miniDig. Thus, for
	 * example, the 3D acceleration sensor description 'a' counts as 3 sensors.
	 * Different sensors are expressed as letters in the configuration String with
	 * space (' ') as marker for empty slot in the miniDig. The configuration letters are:
	 * a - 3D acceleration, o - Orient v2.0, t - Hot&Humid, l - Light and d - FarReach distance sensor.
	 * @param arrangement the current configuration of the miniDig.
	 */
	
	public void configure(String arrangement){
		// "a tl "
		//  31211
		// "atlod"
		//  32111
		this.configured = false;
		slots_used = 0;
		int account = 0;
		for (int i = 0; i < arrangement.length(); ++i){
			char c = arrangement.charAt(i);
			switch (c){
			case ACCELERATION3D:
				sensorAt[account] = true;
				sensorAt[account+1] = true;
				sensorAt[account+2] = true;
				slots_used += 3;
				account += 3;
				break;
			case TEMPERATURE_HUMIDITY:
				sensorAt[account] = true;
				sensorAt[account+1] = true;
				slots_used += 2;
				account += 2;
				break;
			case LIGHT:
				sensorAt[account] = true;
				slots_used++;
				account++;
				break;
			case ORIENTATION:
				sensorAt[account] = true;
				slots_used++;
				account++;
				break;
			case DISTANCE:
				sensorAt[account] = true;
				slots_used++;
				account++;
				break;
			case ' ': // no sensor in slot
				sensorAt[account] = false;
				slots_used++;
				break;
			default:
				if (this.logger != null) logger.logMessage("[ICBX] Unidentified sensor: " + c + ", configuration failed!");
				this.configured = false;
				return;
			}
			if (account > 8){
				if (this.logger != null) logger.logMessage("[ICBX] Too many slots used: " + account + ", configuration failed!");
				this.configured = false;
				return;
			}
		}
		if (account < 8){
			if (this.logger != null) logger.logMessage("[ICBX] Too few slots used: " + account + ", configuration failed!");
			this.configured = false;
			return;
		}
		this.arrangement = arrangement;
		
		initializeVariablesProperly();
		openInput();
		try {
			this.oStream  = ((StreamConnection)connection).openOutputStream();
		}catch(Exception ex) {
			if (logger != null) logger.logMessage("[ICBX] Opening an output stream failed:\n" + ex.toString());
		}
		
		reset();
		setHostMode();
		setHighResolution();
		setSampleInterval();
		setStreamingOn();
		
		this.configured = true;
	}
	
	/*
	 * Opens an input stream to I-CubeX.
	 */
	private void openInput(){
		// Try to create the connection
		try {
			this.iStream = ((StreamConnection)connection).openInputStream();
		} catch (BluetoothConnectionException e) {
			if(this.logger !=null) logger.logMessage("[ICBX] Opening an input stream failed.");
			int status = e.getStatus();
			connected = false;
			isRunning = false;
			switch(status) {
			case BluetoothConnectionException.FAILED_NOINFO:
				logger.logMessage("[ICBX] FAILURE - The connection to the server failed due to unknown reasons.");
				break;
			case BluetoothConnectionException.NO_RESOURCES:
				logger.logMessage("[ICBX] FAILURE - A lack of resources either on the local device or on the remote device.");
				break;
			case BluetoothConnectionException.SECURITY_BLOCK:
				logger.logMessage("[ICBX] FAILURE - The security settings on the local device or the remote device were incompatible with the request.");
				break;
			case BluetoothConnectionException.TIMEOUT:
				logger.logMessage("[ICBX] FAILURE - Timed out.");
				break;
			case BluetoothConnectionException.UNACCEPTABLE_PARAMS:
				logger.logMessage("[ICBX] FAILURE - The configuration parameters provided were not acceptable to either the remote device or the local device.");
				break;
			case BluetoothConnectionException.UNKNOWN_PSM:
				logger.logMessage("[ICBX] FAILURE - No service for the given PSM was registered.");
				break;
			}
			try{
				connection.close();
				connected = false;
				isRunning = false;
			}catch (IOException ex){
				//Do nothing
			}
		} catch(IOException ex) {
			if (logger!= null) logger.logMessage("[ICBX] Opening an input stream failed: \n" + ex.toString());
			try{
				connection.close();
				connected = false;
				isRunning = false;
			}catch (IOException exx){
				//Do nothing
			}
		}
	}
	
	
	/*
	 * Resets the I-CubeX.
	 */
	private void reset(){
		int[] setReset = {START, MAN, DEV, RESET , END }; // 0 = host mode
		this.reset = false;
		send(setReset);
		
		try {
			readData(RESET_ACK);
			reset = true;
		} catch (IOException e) {
			logger.logMessage("[ICBX]Failed to read reset reply.");
		}
		if (!this.reset){
			if (logger != null)
				logger.logMessage("[ICBX]Failed resetting the device!");
		}
	}
	/*
	 * Sets the sensors to high resolution (10-bit) mode.
	 */
	private void setHighResolution(){
		int hiRes = 64;
		this.hires = true;
		int[] setRes = {START, MAN, DEV, RES,	0 , END };
		for (int i = 0; i < sensorAt.length; ++i){
			if (sensorAt[i]){
				this.resolutions[i] = false;
				setRes[4] = hiRes + i; 
				send(setRes);
				try {
					readData(RES);
				} catch (IOException e) {
					logger.logMessage("[ICBX]Failed to read res reply.");
				}
				if (!this.resolutions[i]){
					if (logger != null)
						logger.logMessage("[ICBX]Failed setting resolution on input " + i);
					this.hires = false;
				}
			}
		}
	}
	
	/*
	 * Sets the sample interval of I-CubeX.
	 */
	private void setSampleInterval(){
		this.interval  = 0;
		this.interval = (int)(delay * 1.5);
		int[] setSample = {START, MAN, DEV, INTERVAL, 0, 0, END}; 
		if (interval > 16000){
			interval = 16000;
		}
		if (interval < 4){
			interval = 4;
		}
		this.bytesPerMin = (int)(1000/(this.interval)*21);
		setSample[4] = interval / 128;
		setSample[5] = interval % 128;
		
		send(setSample);
		try {
			readData(INTERVAL);
		} catch (IOException e) {
			logger.logMessage("[ICBX] Failed to read res reply.");
		}
		if (this.interval != interval){
			if (logger != null)
				logger.logMessage("[ICBX] Failed setting sample interval!");
		}
	}	
	
	/*
	 * Sets streaming on.
	 */
	private void setStreamingOn(){
		int on = 64;
		this.on = false;
		int[] setOn = {START, MAN, DEV, STREAM, 0 , END };
		for (int i = 0; i < sensorAt.length; ++i){
			if (sensorAt[i]){
				
				setOn[4] = on + i; 
				send(setOn);
				this.onLine[i] = true;
				try{
				this.readData(STREAM);
				}catch(IOException e){
					if(this.logger != null) logger.logMessage("[ICBX] Failed to read stream command echo.");
				}
				//FIXME: NO reply sent back???
			}
		}
		// How do we check that it was successful?
		this.on = true;
	}
	
	/**
	 * Reads one data message or multiple data messages until the right
	 * command is read. If the cmd parameter is greater or equal to zero,
	 * messages are discarded until a message with the command represented
	 * by the cmd parameter is read. If the cmd parameter is less than zero
	 * the first message is read and analyzed.
	 * @param cmd the command that needs to be read
	 * @return true if message was read succesfully, false otherwise
	 * @throws IOException
	 */
	
	protected boolean readData(int cmd) throws IOException {
		//Right... since the streaming can make things go FUBAR,
		//we need to read a specific message and ignore all of others
		int[] data;
		int index = 0;
		int numberOfMessages = 0;
		
		int c = 0;
		
		
		if(cmd>-1){
			//Read one specific message
			data = new int[100];
			numberOfMessages = 1;
			while(this.iStream!=null) {

				c = this.iStream.read();
				/*
				if(c<0)
					c = c + 256;
				*/
				data[index] = c;
				
				//Okay... sometimes messages go fubar. Read the first byte
				//until it is 240 (new message). HACK!
				if(index == 0 && c != START)
					continue;
				else
					index++;

				if(c == END || index == 100){

					//Ok, so we need to read a specific one
					//Check the command byte...
					if(index>3 && data[3] == cmd){
						//Correct command byte read...
						break;
					}else{
						//Something went wrong... too short message or the command byte was wrong
						//Read another message (reset the loop)
						index = 0;
						data = new int[100];
					}
				}
			}
		}else{
			//used buffer size depends on the sample rate...
			data = new int[bytesPerMin+50];
			while(this.iStream!=null && index < 1100) {
				c = this.iStream.read();
				/*
				if(c<0)
					c = c + 256;
				*/
				data[index] = c;
				

				//Okay... sometimes messages go fubar. Read the first byte
				//until it is 240 (new message). HACK!
				if(index == 0 && c != START)
					continue;
				else
					index++;
				if(c == END){
					numberOfMessages++;
					if(index > bytesPerMin)
						break;
				}
			}
		}
		//HACK! Send one byte after reading a message. This should light up the
		//red light on I-CubeX controller for 0.25s. If ICBX is seding data
		//(yeallow light) and the red light is not on, something has happened to the parser.
		
		int[] msg = {254};
		
		send(msg);
		
		if (data[0] == START){
			handleMessage(data, numberOfMessages);
			return true;
		}
		return false;
	}
	
	/*
	 * Handles the messages from I-CubeX.
	 */
	
	private void handleMessage(int[] data, int numberOfMessages){
		int index = 0;
		for(int j=0; j < numberOfMessages; j++){
			int status = data[index];
			index++;
			int manufacturer = data[index];
			index++;
			int dev = data[index];
			index++;
			int command = data[index];
			index++;

			//if (logger != null) logger.logMessage("status:" + status + " manufacturer: "+ manufacturer);
			//if (logger != null) logger.logMessage("device:" + dev + " command:" + command);

			int value = 0;

			switch (command){
			case RESET_ACK:
				this.reset = true;
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of RESET ACK message: " + data[index]);
				index++;
				break;
			case RES:
				int res = data[index];
				index++;
				if (res < 64){
					logger.logMessage("[ICBX] Wrong resolution!");
				}
				else{
					res = res - 64;
					this.resolutions[res] = true;
				}
				this.hires = true;
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of RESET ACK message: " + data[index]);
				index++;
				break;	
			case INTERVAL:
				this.interval = data[index]*128 + data[index +1];
				index +=2;
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of RESET ACK message: " + data[index]);
				index++;
				break;	
			case STATUS:
				int errorCode = data[index];
				index++;
				switch (errorCode){
				case CONFIG_ERROR:
					logger.logMessage("[ICBX] Wi-miniDig reports: configuration setting error");
					break;
				case MIDI_ERROR:
					logger.logMessage("[ICBX] Wi-miniDig reports: MIDI protocol error, a byte is greater than 127");
					break;
				case SCRAMBLE_ERROR:
					logger.logMessage("[ICBX] Wi-miniDig reports: MDI byte has been scrambled");
					break;
				case BUFFER_ERROR:
					logger.logMessage("[ICBX] Wi-miniDig reports: Buffer is full, data may be lost (decrease sample rate)");
					break;
				}

				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of STATUS message: " + data[index]);
				index++;
				break;

			case VERSION:
				//FIXME: does nothing
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of VERSION message: " + data[index]);
				index++;
				break;

			case MODE:
				int mode = data[index];
				index++;
				this.mode = mode;
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of MODE message: " + data[index]);
				index++;
				break;
				// stand-alone mode responses
			case NAME:
				//FIXME: does nothing
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of NAME message: " + data[index]);
				index++;
				break;

			case CONFIG:
				//FIXME: does nothing
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of CONFIG message: " + data[index]);
				index++;
				break;

				// host mode responses
			case STREAM_DATA:
				for (int i = 0; i < inputTypes.length; ++i){
					if (hires){
						// high 10-bit (2 bytes in body) resolution
						value = data[index]* 8 + data[index +1];
						index += 2;
					}else{	
						//7-bit (one byte per body) resolution
						value = data[index];
						index++;
					}
					double result = 0;
					switch(inputTypes[i]){
						case ACCELERATION_INPUT_X:
							result = ICubexLib.MIDItoAcceleration(value, 1);
							break;
						case ACCELERATION_INPUT_Y:
							result = ICubexLib.MIDItoAcceleration(value, 2);
							break;
						case ACCELERATION_INPUT_Z:
							result = ICubexLib.MIDItoAcceleration(value, 3);
							break;
						case TEMPERATURE_INPUT:
							result = ICubexLib.MIDItoTemperature(value);
							break;
						case HUMIDITY_INPUT:
							result = ICubexLib.MIDItoHumidity(value);
							break;
						case AMBIENT_LIGHT_INPUT:
							result = ICubexLib.MIDItoLux(value);
							break;
						case ORIENTATION_INPUT:
							result = ICubexLib.MIDItoDegrees(value);
							break;
						case DISTANCE_INPUT:
							result = ICubexLib.MIDItoDistance(value);
							break;
						default:
							break;
					}
					
					String resultString = result + "";
					if (resultString.length() > 6){
						int twodecimal = resultString.indexOf('.') + 2;
						if (twodecimal < 2)
							twodecimal = 6;
						if (twodecimal > resultString.length())
							twodecimal = resultString.length();
						//result = Double.parseDouble(resultString.substring(0, twodecimal));
						resultString = resultString.substring(0, twodecimal);
					}
					this.parameters.put(parNames[i], resultString);
					if(this.subscribers.size() > 0){
						((Vector)this.streamParameters.get(parNames[i])).addElement(resultString);
					}

				}
				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of STREAM DATA message: " + data[index]);
				index++;
				break;

			case SAMPLE_DATA:
				int sensor = data[index];
				index++;
				if (hires){
					// high 10-bit (2 bytes in body) resolution
					value = data[index]* 128 + data[index +1];
					index += 2;
				}else{	
					//7-bit (one byte per body) resolution
					value = data[index];
					index++;
				}
				this.parameters.put("sensor" + sensor, "" + value);

				if (data[index] != END)
					logger.logMessage("[ICBX] Unexpected end of SAMPLE DATA message: " + data[index]);
				index++;
				break;

			default:
				logger.logMessage("[ICBX] Unidentified command: " + command);
			return;
			}
		}
	}
	
	/*
	 * Sends a message to I-CubeX.
	 */
	
	private boolean send(int[] what){ 
		for (int i = 0; i < what.length; ++i){
			try {
				oStream.write(what[i]);
				oStream.flush();
			} catch (IOException e) {
				return false;
			}
		}
		return true;
	}
	
	
	
	/**
	 * Reads data from the device. To implement a reader, extend the ReaderSkeleton class and implement the readData method where 
	 * all processing takes place.
	 */
	public void run() {
		long time_1, time_2, time_3;
		int failures = 0;
		try{
			time_3  = System.currentTimeMillis();
			while(this.isRunning && this.configured) {
				
				// Get first tick
				//time_1 = System.currentTimeMillis();
				
				// Read and process data
				try {
					if(!readData())
						failures++;
					else{
						failures = 0;
						this.dataBatchNumber++;
					}
					
					if (delay != 0 && delay < 30000 && failures > 30000 / delay){
						if (logger != null) logger.logMessage("[ICBX] Disconnected from " + url.substring(0, url.indexOf(';'))  + ", trying to restore connection...");
						connected = false;
						restoreConnection();
					}else if (delay == 0 && failures > 1000){
						if (logger != null) logger.logMessage("[ICBX] Disconnected from " + url.substring(0, url.indexOf(';'))  + ", trying to restore connection...");
						connected = false;
						restoreConnection();
					}

				}catch(java.io.IOException ex) {
					connected = false;
					if(isRunning == true){
						if (logger != null) logger.logMessage("[ICBX] Disconnected from " + url.substring(0, url.indexOf(';'))  + ", trying to restore connection...");
						restoreConnection();
					}
				}catch(Throwable t){
					if (logger != null) logger.logMessage("[ICBX] Throwable caught: " + t);
				}
				// Request the monitor of the reader 
				synchronized(this) {
					
					/*
					// Get second tick
					time_2 = System.currentTimeMillis();
					
					// If the delay has not been long enough, sleep for a while
					if((time_2 - time_1) >= 0 && (time_2 - time_1) < delay)
						this.wait(delay - (time_2 - time_1));
					*/
				}
				if(time_3+this.notifyDelay < System.currentTimeMillis()){
					notifySubscribers();
					time_3 = System.currentTimeMillis();
				}
			}
			
			// Close the input stream
			this.iStream.close();
			
		
		}catch(Exception ex) {
			
			if (logger != null) {
				logger.logMessage("[ICBX]Reader error, reader stopped: " + ex.toString());
				ex.printStackTrace();
			}
		}
	}
	
	/**
	 * Returns the parameter types.
	 * @return the parameter types
	 */

	public int[] getParameterTypes() {
		int[] types = new int[8];
		types[0] = Properties.DATATYPE_STRING;
		types[1] = Properties.DATATYPE_STRING;
		types[2] = Properties.DATATYPE_STRING;
		types[3] = Properties.DATATYPE_STRING;
		types[4] = Properties.DATATYPE_STRING;
		types[5] = Properties.DATATYPE_STRING;
		types[6] = Properties.DATATYPE_STRING;
		types[7] = Properties.DATATYPE_STRING;
		return types;
	}
	
	protected void notifySubscribers() {
		if(this.subscribers.size()==0){
			String[] names = this.getParameterNames();
			for (int i = 0; i<names.length; i++){
				this.streamParameters.put(names[i], new Vector());
			}
			return;
		}
		
		//Read the current data only once
		String[] names = this.getParameterNames();
		Vector[] copy = new Vector[names.length];
		
		for (int i = 0; i<names.length; i++){
			copy[i] = (Vector)this.streamParameters.get(names[i]);
			this.streamParameters.put(names[i], new Vector());
		}
				
		synchronized(this) {
			long time = System.currentTimeMillis();
			for(int i=0,j=this.subscribers.size();i<j;i++) {
				if(((Long)this.lastSubscriberUpdate.elementAt(i)).longValue() + ((Long)this.timeouts.elementAt(i)).longValue() < time){
				//synchronized(this.subscribers.elementAt(i)) {
					//Send data to the observer. Observer decides what it does with the data.
					((Observer)this.subscribers.elementAt(i)).update(copy, this);
					this.lastSubscriberUpdate.removeElementAt(i);
					this.lastSubscriberUpdate.insertElementAt(new Long(time), i);
					//Wake the observer
					//this.subscribers.elementAt(i).notify();
				//}
				}
			}
		}		
	}
	
	/**
	 * Closes the reader. Sets the isRunning flag to false which will stop the reader.
	 */
	public void close() {
		this.isRunning = false;
		this.reset();
		try {
			if (this.iStream != null)
				this.iStream.close();
			if (this.connection != null)
				this.connection.close();
		} catch (IOException e) {
			//Nothing can be done
		}
	}
	
	protected boolean readData() throws IOException{
		return this.readData(-1);
	}
	
	/**
	 * Sets the delay Reader waits before reading data again. Default delay is 100 milliseconds.
	 * @param delay Delay for reading data in milliseconds.
	 */
	
	public void setReadDelay(long delay){
		if (delay < 3)
			this.delay = 3;
		else if(delay > 12000)
			this.delay = 12000;
		else
			this.delay = delay;
		//Change the I-Cubex timeout... (1.5x the read dealay);
		this.setSampleInterval();
	}
	
	/** 
	 *  Tries to restore the connection to the device associated with this reader. 
	 * 
	 * 
	 */
	protected synchronized void restoreConnection(){
		int timeout = firstTimeout;
		
		// stop trying only if we are allowed to give up and the final timeout is reached.
		while (!(timeout >= finalTimeout && giveUp)){
			Connection connection;
			
			
			connection = this.factory.tryRestore(this, url, logger);
			//else if (...) { //Some other Factory than the bluetooth one...}
			if(connection != null){
				
				
				try {
					this.iStream = ((StreamConnection)connection).openInputStream();
					this.connection = connection;
					if (logger != null) logger.logMessage("[ICBX]Connection restored to " + url.substring(0, url.indexOf(';')));
					connected = true;
					reset();
					setHostMode();
					setHighResolution();
					setSampleInterval();
					setStreamingOn();
					return;
				} catch (IOException e) {
					if (logger != null) logger.logMessage("[ICBX]Unable to restore connection.");
				}
			}	
			
			try {
				this.wait(timeout);
			} catch (InterruptedException e) {
				// Woken up, let's just be happy and continue
			}
			
			timeout = timeout * 2;
			if (timeout > finalTimeout)
				timeout = finalTimeout;
		}
		isRunning = false;
		if (logger != null) logger.logMessage("[ICBX]Reader "+ getDeviceName() + " cannot restore connection.");
	}

	public void newData(Customer prod, CustomerData pd, EventData eventData){
		// TODO Auto-generated method stub
		
	}
}
