/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.reader;

import hiit.blackboard.BlackboardFactory;
import hiit.blackboard.Blackboard;
import hiit.blackboard.Customer;
import hiit.blackboard.CustomerData;
import hiit.blackboard.CustomerInfo;
import hiit.blackboard.event.EventData;
import hiit.blackboard.event.GPSEvent;
import hiit.blackboard.event.GSMEvent;
import hiit.connectorUnit.BTDeviceConnector;
import hiit.connectorUnit.BTReaderFactory;
import hiit.logger.Logger;
import hiit.mapper.Properties;
import hiit.observer.Observer;

import java.io.IOException;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.bluetooth.RemoteDevice;
import javax.microedition.io.Connection;

/**
 * A LocalReader providing nearby Bluetooth devices and GSM cell information.
 * GSM information will be available if provided to the BeTelGeuse Blackboard
 * from an external program. (For example, a Python script can send us the GSM
 * info)
 * 
 * @author Joonas Kukkonen and Eemil Lagerspetz
 */

public class LocalGSMReader extends LocalDeviceReader {
    private static final String signalstr = "Signal Strength";
    
    private static final String signalbars = "Signal Bars";
    
    private static final String cellid = "Cell ID";
    
    private static final String areaid = "Area ID";
    
    private static final String mcc = "MCC";
    
    private static final String mnc = "MNC";
    
    private static final String phone_battery = "Phone Battery";
    
    private static final String phone_profile = "Phone Profile";
    
    private static final String USER_INACTIVE = "Inactive Time";
    
    private static final String FREE_RAM = "Free RAM";
    
    private static final String GSM_CHANGE = "GSM Changed";
    
    /**
     * If the final timeout is reached and the connection is still down, should
     * we abandon trying?
     */
    protected boolean giveUp;
    
    /**
     * The logger that prints messages. Useful for mobile devices.
     */
    protected Logger logger;
    
    /**
     * Used to store the names and values of parameters. The name (key) is the
     * parameter name that is visualized whereas the value is the current sensor
     * reading of that parameter.
     */
    protected Hashtable parameters;
    
    /**
     * Boolean flag used for managing when the reader should run.
     */
    protected boolean isRunning = true;
    
    /**
     * A list of objects that have subscribed to events from the reader. Each
     * time the reader reads information from the device, a new event is
     * produced.
     */
    protected Vector subscribers = new Vector();
    
    /**
     * A list of subscriber timeouts (how often they want data).
     */
    
    protected Vector timeouts = new Vector();
    
    /**
     * A vector of timestamps when data was sent last to subscribers.
     */
    
    protected Vector lastSubscriberUpdate = new Vector();
    
    /**
     * Delay, specifies how often messages are read from the sensing device.
     */
    protected long delay = 100;
    
    /**
     * Numerical value is saved to each index describing if there is new data
     * for the poller. 1 means there is new data, 0 no new data. Poller
     * identification number is used as an index.
     */
    protected Vector hasData;
    
    /**
     * Next free poller identification number.
     */
    protected int pullerId = 0;
    
    /**
     * The ReaderFactory used to create the Reader objects. Used in connection
     * restoration.
     */
    
    protected BTReaderFactory factory;
    
    /**
     * The device ID number associated to the sensor to which this Reader is
     * connected to.
     */
    
    protected int deviceId;
    
    /**
     * Describes if the Reader is connected or not.
     */
    
    protected boolean connected = true;
    
    /**
     * The number of last data batch.
     */
    
    protected int dataBatchNumber = 0;
    
    /**
     * The mode of the Reader. True if the Reader is in stream mode, false
     * otherwise.
     */
    
    protected boolean mode = true;
    
    private GSMEvent gsmEvent = null;
    
    /**
     * Constructor for LocalDeviceReader that parses data from the local device.
     * 
     * @param bta
     *            the Bluetooth address of the local device
     * @param btConnector
     *            the BTDeviceConnector object used to inqury for nearby
     *            devices' addresses
     */
    
    public LocalGSMReader(String bta, BTDeviceConnector btConnector) {
        this.delay = 5000;
        this.bta = bta;
        this.btConnector = btConnector;
        this.deviceId = Properties.LOCAL_GSM_READER;
        
        // If it fails, we don't want to show empty cellid variable...
        initializeVariables();
        btConnector.subscribe(this);
       
    }
    
    /**
     * Creates a new LocalDeviceReader. Use the initialize(..) method to set
     * variables.
     * 
     */
    
    public LocalGSMReader() {
        this.delay = 5000;
        this.deviceId = Properties.LOCAL_GSM_READER;
      
    }
    
    public void initialize(Connection connection, String url, Logger logger,
            BTReaderFactory factory, int deviceId, boolean streamMode) {
        this.factory = factory;
        this.btConnector = factory.getConnector();
        this.logger = logger;
        gsmEvent = new GSMEvent(logger);
        this.bta = factory.getConnector().getGatherer().getBluetoothAddress();
        this.deviceId = Properties.LOCAL_GSM_READER;
        initializeVariables();
        btConnector.subscribe(this);
        this.mode = streamMode;
        this.info = new CustomerInfo(getDeviceName(), deviceId, 1, this.bta,
                (streamMode ? 0 : 1), delay, getParameterNames(),
                getParameterTypes(), false);
        BlackboardFactory.getBlackboard(logger).registerProducer(this, info, gsmEvent.getEventList());
		BlackboardFactory.getBlackboard(logger).subscribeConsumer(this);
        
    }
    
    public String getDeviceName() {
        return "LGR";
    }
    
    public String getBluetoothAddress() {
        return this.bta;
    }
    
    protected void initializeVariables() {
        this.parameters = new Hashtable(19);
        this.parameters.put("Activity", "");
        this.parameters.put("Discovery number", "");
        this.parameters.put("Last discovery time", "");
        this.parameters.put("Discovered devices", "");
        this.parameters.put("Transmission number", "");
        this.parameters.put("Transmission time", "");
        this.parameters.put("Transmission size", "");
        this.parameters.put("BTDevices", "");
        this.parameters.put(mcc, "0");
        this.parameters.put(mnc, "0");
        this.parameters.put(areaid, "0");
        this.parameters.put(cellid, "0");
        this.parameters.put(signalstr, "0");
        this.parameters.put(signalbars, "0");
        this.parameters.put(phone_battery, "0");
        this.parameters.put(phone_profile, "");
        this.parameters.put(GSM_CHANGE, "");
        this.parameters.put(USER_INACTIVE, "");
        this.parameters.put(FREE_RAM, "");
    }
    
    public int[] getParameterTypes() {
        int[] types = new int[19];
        types[0] = Properties.DATATYPE_STRING;
        types[1] = Properties.DATATYPE_STRING;
        types[2] = Properties.DATATYPE_STRING;
        types[3] = Properties.DATATYPE_STRING;
        types[4] = Properties.DATATYPE_STRING;
        types[5] = Properties.DATATYPE_STRING;
        types[6] = Properties.DATATYPE_STRING;
        types[7] = Properties.DATATYPE_STRING;
        types[8] = Properties.DATATYPE_STRING;
        types[9] = Properties.DATATYPE_STRING;
        types[10] = Properties.DATATYPE_STRING;
        types[11] = Properties.DATATYPE_STRING;
        types[12] = Properties.DATATYPE_STRING;
        types[13] = Properties.DATATYPE_STRING;
        types[14] = Properties.DATATYPE_STRING;
        types[15] = Properties.DATATYPE_STRING;
        types[16] = Properties.DATATYPE_STRING;
        types[17] = Properties.DATATYPE_STRING;
        types[18] = Properties.DATATYPE_STRING;
        return types;
    }
    
    public String[] getParameterNames() {
        // Avoid null pointer exceptions
        if (this.parameters == null)
            return null;
        
        String[] params = new String[19];
        params[0] = "Activity";
        params[1] = "Discovery number";
        params[2] = "Last discovery time";
        params[3] = "Discovered devices";
        params[4] = "Transmission number";
        params[5] = "Transmission time";
        params[6] = "Transmission size";
        params[7] = "BTDevices";
        params[8] = mcc;
        params[9] = mnc;
        params[10] = areaid;
        params[11] = cellid;
        params[12] = signalstr;
        params[13] = signalbars;
        params[14] = phone_battery;
        params[15] = phone_profile;
        params[16] = GSM_CHANGE;
        params[17] = USER_INACTIVE;
        params[18] = FREE_RAM;
        return params;
    }
    
    public String getParameter(String name) {
        if (this.parameters == null)
            return null;
        return (String) this.parameters.get(name);
    }
    
    public String[] getData() {
        
        // Avoid null pointer exceptions
        if (this.parameters == null)
            return null;
        
        // Reserve memory
        final int max = this.parameters.size();
        String[] params = new String[max];
        
        // Loop though the keys and store the parameter names.
        Enumeration enume = this.parameters.elements();
        int index = 0;
        while (enume.hasMoreElements())
            params[index++] = (String) enume.nextElement();
        
        // Return parameter names.
        return params;
    }
    
    protected boolean readData() {
        return true;
    }
    
    public void close() {
        if (this.logger != null)
            this.logger.logMessage("[LDR] Closing local device reader");
        this.isRunning = false;
        this.factory.getConnector().unsubscribe(this);
    }
    
    /**
     * BTDeviceConnector uses this method to update the readers (and thus
     * announce that a device inquiry has been completed) to the
     * LocalDeviceReader.
     * 
     * @param readers
     *            a Vector of all current readers
     * @param btdc
     *            the object that called the method
     */
    
    public void update(Vector readers, Object btdc) {
        BTDeviceConnector bc = (BTDeviceConnector) btdc;
        Vector devicesC = bc.getConnectedRemoteDevices();
        Vector devices = bc.getLastFound();
        
        long lastDiscoveryTime = bc.getLastDiscoveryTime() / 1000;
        long discoveryNumber = bc.getDiscoveryNumber();
        int numDevices = devices.size();
        
        parameters.put("Discovery number", discoveryNumber + "");
        parameters.put("Last discovery time", lastDiscoveryTime + " s");
        parameters.put("Discovered devices", numDevices + "");
        
        // if (logger != null) logger.logMessage("[LDR] " + "Received " +
        // devices.size() + " devices.");
        
        String name = "";
        for (int i = 0; i < devicesC.size(); i++) {
            RemoteDevice r = (RemoteDevice) devicesC.elementAt(i);
            name += r.getBluetoothAddress() + " ";
            
            try {
                name += r.getFriendlyName(false);
            } catch (IOException e) {
            }
            name += "\n";
        }
        
        if (devices == null || devices.size() == 0) {
            this.parameters.put("BTDevices", name);
            return;
        }
        for (int i = 0; i < devices.size(); i++) {
            RemoteDevice r = (RemoteDevice) devices.elementAt(i);
            if (devicesC.contains(r)) {
                // These have already been added, do not process.
                continue;
            }
            name += r.getBluetoothAddress() + " ";
            
            try {
                name += r.getFriendlyName(false);
            } catch (IOException e) {
            }
            name += "\n";
        }
        this.parameters.put("BTDevices", name);
        notifySubscribers();
    }
    
    public boolean isConnected() {
        return true;
    }
    
    protected void restoreConnection() {
        this.connected = true;
    }
    
    /**
     * Reads data from the device. To implement a reader, extend the
     * ReaderSkeleton class and implement the readData method where all
     * processing takes place.
     */
    public void run() {
        long time_1, time_2;
        int failures = 0;
        try {
            while (this.isRunning) {
                
                if (!mode) {
                    // If in request mode...
                    
                    // check if the reader needs to disconnect...
                    // if(this.iStream != null && this.connection != null)
                    // disconnect();
                    
                    // Wait for something to happen...
                    synchronized (this) {
                        this.wait();
                    }
                    
                    if (!mode) {
                        if (connect()) {
                            // If still in request mode, connect, read one
                            // message and disconnect
                            readData();
                            disconnect();
                            notifySubscribers();
                        }
                        // Jump back and wait...
                        continue;
                    }
                    // else: mode is true, so continue normally... the reader
                    // has already been reconnected...
                }
                
                // Get first tick
                time_1 = System.currentTimeMillis();
                
                // Read and process data
                try {
                    if (!readData())
                        failures++;
                    else {
                        failures = 0;
                        this.dataBatchNumber++;
                    }
                    
                    // Set the hasData Vector elements
                    for (int i = 0; i < pullerId; i++)
                        hasData.setElementAt(new Integer(1), i);
                } catch (Throwable t) {
                    if (logger != null)
                        logger.logMessage("[LDR] Throwable caught: " + t);
                }
                // Request the monitor of the reader
                synchronized (this) {
                    
                    // Get second tick
                    time_2 = System.currentTimeMillis();
                    
                    // If the delay has not been long enough, sleep for a while
                    if ((time_2 - time_1) >= 0 && (time_2 - time_1) < delay)
                        this.wait(delay - (time_2 - time_1));
                }
                notifySubscribers();
            }
            
        } catch (Exception ex) {
            
            if (logger != null) {
                ex.printStackTrace();
                logger.logMessage(ex.toString());
                logger.logMessage("[LDR] Reader error, reader stopped");
            }
        }
    }
    
    /**
     * Subscribes to events from the reader.
     * 
     * @param obs
     *            The Observer that subscribes to events.
     */
    public void subscribe(Observer obs, long timeout) {
        synchronized (this) {
            this.subscribers.addElement(obs);
            this.timeouts.addElement(new Long(timeout));
            this.lastSubscriberUpdate.addElement(new Long(0));
        }
    }
    
    /**
     * Notifies all subscribers that new data is available.
     */
    protected void notifySubscribers() {
        
        // Read the current data only once
        Vector[] v = new Vector[this.parameters.size()];
        Vector temp = null;
        String[] names = this.getParameterNames();
        
        for (int i = 0; i < names.length; i++) {
            temp = new Vector();
            temp.addElement(this.parameters.get(names[i]));
            v[i] = temp;
        }
        
        CustomerData d = new CustomerData(v, new Date().getTime(), info);
        
        EventData[] eventData = this.gsmEvent.getEvents(d);
        BlackboardFactory.getBlackboard(logger).pushData(this, new CustomerData(v, new Date().getTime(), info), eventData);

        /*
        if (this.subscribers.size() == 0)
            return;
        
        synchronized (this) {
            long time = System.currentTimeMillis();
            for (int i = 0, j = this.subscribers.size(); i < j; i++) {
                if (((Long) this.lastSubscriberUpdate.elementAt(i)).longValue()
                        + ((Long) this.timeouts.elementAt(i)).longValue() < time) {
                    // synchronized(this.subscribers.elementAt(i)) {
                    // Send data to the observer. Observer decides what it does
                    // with the data.
                    ((Observer) this.subscribers.elementAt(i)).update(v, this);
                    this.lastSubscriberUpdate.removeElementAt(i);
                    this.lastSubscriberUpdate
                            .insertElementAt(new Long(time), i);
                    // Wake the observer
                    // this.subscribers.elementAt(i).notify();
                    // }
                }
            }
        }*/
    }
    
    /**
     * Unsubscribes the Observer from events from the reader.
     * 
     * @param obs
     *            The object that unsubscribes.
     */
    public void unsubscribe(Observer obs) {
        synchronized (this) {
            int i = this.subscribers.indexOf(obs);
            this.subscribers.removeElement(obs);
            this.timeouts.removeElementAt(i);
            this.lastSubscriberUpdate.removeElementAt(i);
        }
    }
    
    /**
     * Announces a new poller object. The method retuns the identification
     * number for the poller.
     * 
     * @return The identification number.
     */
    public synchronized int announcePuller() {
        pullerId++;
        if (hasData.size() == pullerId - 1)
            hasData.addElement(new Integer(0));
        else
            hasData.setElementAt(new Integer(0), pullerId - 1);
        return pullerId - 1;
    }
    
    /**
     * Pollers can use this method to inqury if the Reader has new data.
     * 
     * @param id
     *            The poller id number
     * @return true if there is new data for the poller, false otherwise
     */
    
    public boolean hasNewData(int id) {
        if (((Integer) hasData.elementAt(id)).intValue() == 1)
            return true;
        
        return false;
    }
    
    /**
     * Sets timeouts for connection restoration. The timeout starts at
     * firstTimeout and it is multiplied with 2 each time connection restoration
     * fails. FinalTimeout defines the maximum lenght for a timeout. If giveUp
     * is true, the reader does not try to restore the connection after the
     * finalTimeout value is reached. If giveUp is false, the Reader tries to
     * restore connection indefinetly. Default firstTimeout is 4000 milliseconds
     * and finalTimeout 128000 milliseconds.
     * 
     * @param firstTimeout
     *            The lenght of first timeout in milliseconds
     * @param finalTimeout
     *            The maximum lenght of a timeout in milliseconds
     * @param giveUp
     *            Defines if the Reader gives up connection restoration after
     *            the final timeout is reached
     */
    
    public void setTimeouts(int firstTimeout, int finalTimeout, boolean giveUp) {
        this.firstTimeout = firstTimeout;
        this.finalTimeout = finalTimeout;
        this.giveUp = giveUp;
    }
    
    /**
     * Getter for the timeouts specified in the method above.
     * 
     * @return a Vector with the elements: firstTimeout (int), finalTimeout
     *         (int), giveUp (boolean).
     */
    public Vector getTimeouts() {
        Vector timeouts = new Vector();
        timeouts.addElement(new Integer(firstTimeout));
        timeouts.addElement(new Integer(finalTimeout));
        timeouts.addElement(new Boolean(giveUp));
        return timeouts;
    }
    
    /**
     * Sets the delay Reader waits before reading data again. Default delay is
     * 100 milliseconds.
     * 
     * @param delay
     *            Delay for reading data in milliseconds.
     */
    
    public void setReadDelay(long delay) {
        this.delay = delay;
    }
    
    /**
     * Getter for the read delay.
     * 
     * @return the delay the Reader waits between subsequent data reads.
     */
    public long getReadDelay() {
        return this.delay;
    }
    
    /**
     * Returns the device ID number associated with the sensor to which this
     * Reader is connected to.
     */
    
    public int getDeviceId() {
        return this.deviceId;
    }
    
    /**
     * Returns true if the Reader is dead and no new data can be expected, false
     * otherwise. Note that a Reader wich is not connected might still be alive,
     * because it is trying to restore the connection to the device.
     */
    
    public boolean isDead() {
        return !this.isRunning;
    }
    
    /**
     * Used by the transmitter plugin to update status parameters.
     * 
     * @param transmissionNumber
     * @param transmissionTime
     * @param dataSize
     */
    
    public void sendTransmissionParams(int transmissionNumber,
            long transmissionTime, int dataSize) {
        this.parameters.put("Transmission number", transmissionNumber + "");
        this.parameters.put("Transmission time", transmissionTime + " ms");
        this.parameters.put("Transmission size", dataSize + " B");
    }
    
    /**
     * Returns the number of the last data batch from the sensor. This method
     * can be used to check if the Reader is getting new data from the sensor.
     */
    
    public int getLastDataNumber() {
        return this.dataBatchNumber;
    }
    
    /**
     * Updates the activity parameter
     */
    
    public void updateActivity(String activity) {
        this.parameters.put("Activity", activity);
    }
    
    /**
     * There are two modes available for Readers: streaming mode and request
     * mode. In streaming mode the Reader will read all available data from the
     * data stream. In request mode the Reader only reads data when readOne() is
     * called. Also, in request mode the Reader connects to the source before
     * reading a message and disconnects after the message has been read.
     * 
     * @param streaming
     *            sets streaming mode on with true value and request mode with
     *            false value
     */
    
    public void setStreamingMode(boolean streaming) {
        if (!mode && streaming) {
            this.restoreConnection();
            this.notify();
        }
        this.mode = streaming;
    }
    
    /**
     * Returns the streaming mode.
     * 
     * @return true, if the Reader is in streaming mode. Returns false if the
     *         Reader is in request mode.
     */
    
    public boolean getStreamingMode() {
        return this.mode;
    }
    
    /**
     * Disconnects from the data source.
     */
    
    protected void disconnect() {
        // data source connects to blackboard, not me!
    }
    
    /**
     * Connects to the data source.
     */
    
    protected boolean connect() {
        return true; // data source connects to blackboard, not me!
    }
    
    public void readOne() {
        if (!mode) {
            this.notify();
        }
    }
    
    public void newData(Customer prod, CustomerData pd) {
        // parse here.
        //StringBuffer buf = new StringBuffer();
        Vector data = pd.getData()[0];
        for (int i = 0; i < data.size(); ++i) {
            // CHANGED, CC, NC, AC, CELLID, SS, SB, BATT, PROFILE
            if (i == 8) {
                this.parameters.put(phone_profile, ""
                        + (String) data.elementAt(i));
                //buf.append(" PF=" + (String) data.elementAt(i));
                continue;
            }
            else if (i == 5) {
                this.parameters.put(signalstr, ""
                        + (String) data.elementAt(i));
                //buf.append(" SS=" + (String) data.elementAt(i));
                continue;
            } else if (i == 6) {
                this.parameters.put(signalbars, ""
                        + (String) data.elementAt(i));
                //buf.append(" SS=" + (String) data.elementAt(i));
                continue;
            }
            Integer ig = (Integer) data.elementAt(i);
            switch (i) {
            case 0:
                this.parameters.put(GSM_CHANGE, "" + ig);
                //buf.append(" CH=" + ig);
                break;
            case 1:
                this.parameters.put(mcc, "" + ig);
                //buf.append(" CC=" + ig);
                break;
            case 2:
                this.parameters.put(mnc, "" + ig);
                //buf.append(" NC=" + ig);
                break;
            case 3:
                this.parameters.put(areaid, "" + ig);
                //buf.append(" AC=" + ig);
                break;
            case 4:
                this.parameters.put(cellid, "" + ig);
                //buf.append(" CID=" + ig);
                break;
            case 7:
                this.parameters.put(phone_battery, "" + ig);
                //buf.append(" BAT=" + ig);
                break;
            case 9:
                this.parameters.put(USER_INACTIVE, "" + ig);
                //buf.append(" BAT=" + ig);
                break;
            case 10:
                this.parameters.put(FREE_RAM, "" + ig);
                //buf.append(" BAT=" + ig);
                break;
            }
        }
        //if (BlackboardFactory.DEBUG)
        //    logger.logMessage("[LGR] got data: " + buf.toString());
    }
    
    public boolean requestData() {
        // Our data comes from an external source,
        // which we cannot regulate, so requestData does not make sense.
        return false;
    }
    
    public void updateProducerList(Blackboard bb, Vector customers) {
        for (int i = 0; i < customers.size(); ++i) {
            Customer cs = (Customer) customers.elementAt(i);
            if (cs == this)
                continue;
            // We only subscribe to newly appeared GSM Servers.
            System.out.println("local gsm, updateprodlist: " + cs);
            
            if (cs.getInfo().getType() == hiit.mapper.Properties.GSM_SERVER)
                bb.subscribeToProducer(this, cs);
        }
    }
}
