/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.reader;

import hiit.connectorUnit.BTReaderFactory;
import hiit.logger.Logger;
import hiit.observer.Observer;
import java.util.Vector;

import javax.microedition.io.Connection;

/**
 * The reader interface defines a common class for reading data from sensing
 * devices.
 * 
 * @author Petteri Nurmi, Joonas Kukkonen and Eemil Lagerspetz
 * 
 */
public interface Reader extends Runnable {
    
    /**
     * Returns the name of the device that the reader reads. Used for
     * visualization of devices.
     * 
     * @return String The name of the device.
     */
    public String getDeviceName();
    
    /**
     * Returns the value of a parameter with a given name. If no parameter with
     * the given name exists, returns null.
     * 
     * @param parameter
     *            The name of the parameter whose value is requested.
     * @return The value of the parameter
     */
    public String getParameter(String parameter);
    
    /**
     * Gets the names of available parameters.
     * 
     * @return String[] A String array that contains the names of the parameters
     *         that the reader provides.
     */
    public String[] getParameterNames();
    
    /**
     * Runs the reader. Typically connects to the device, processes the message
     * and sleeps for a while before reading again data from the device.
     */
    public void run();
    
    /**
     * Stops the reader and closes all connections.
     */
    public void close();
    
    /**
     * Unsubscribes the object from events from the reader.
     * 
     * @param obj
     *            The object that unsubscribes.
     */
    public void unsubscribe(Observer obj);
    
    /**
     * Subscribes to events from the reader.
     * 
     * @param obj
     *            The object that subscribes to events.
     */
    public void subscribe(Observer obj, long timeout);
    
    /**
     * Initializes the Reader. Mean for cases when the Reader implementation is
     * constructed with an empty constructor by Class.forName() method.
     * 
     * @param connection -
     *            the connection to the sensor
     * @param url
     *            the connection URL
     * @param logger
     *            the Logger object
     * @param factory
     *            the BTDeviceFactory which is used to restore connections
     * @param deviceId
     *            the device ID of the sensor
     * @param mode
     *            the mode of the Reader at startup. True for streaming mode,
     *            false for request mode.
     */
    
    public void initialize(Connection connection, String url, Logger logger,
            BTReaderFactory factory, int deviceId, boolean mode);
    
    /**
     * Returns all of the parameters in a String array.
     */
    
    public String[] getData();
    
    /**
     * Sets timeouts for connection restoration. The timeout starts at
     * firstTimeout and it is multiplied with 2 each time connection restoration
     * fails. FinalTimeout defines the maximum lenght for a timeout. If giveUp
     * is true, the reader does not try to restore the connection after the
     * finalTimeout value is reached. If giveUp is false, the Reader tries to
     * restore connection indefinetly.
     * 
     * @param firstTimeout
     *            The lenght of first timeout in milliseconds
     * @param finalTimeout
     *            The maximum lenght of a timeout in milliseconds
     * @param giveUp
     *            Defines if the Reader gives up connection restoration after
     *            the final timeout is reached
     */
    
    public void setTimeouts(int firstTimeout, int finalTimeout, boolean giveUp);
    
    /**
     * Getter for the timeouts specified in the method above.
     * 
     * @return a Vector with the elements: firstTimeout (int), finalTimeout
     *         (int), giveUp (boolean).
     */
    public Vector getTimeouts();
    
    /**
     * Sets the delay Reader waits before reading data again.
     * 
     * @param delay
     *            Delay for reading data in milliseconds.
     */
    
    public void setReadDelay(long delay);
    
    /**
     * Getter for the delay to wait between subsequent reads.
     * 
     * @return the number of milliseconds the reader is to wait between
     *         subsequent data reads.
     */
    public long getReadDelay();
    
    /**
     * Returns the bluetooth address of the sensor device.
     */
    
    public String getBluetoothAddress();
    
    /**
     * Rertuns the device ID which is associated with the sensor device.
     */
    
    public int getDeviceId();
    
    /**
     * Returns true if the Reader is connected to a sensor or other device,
     * false otherwise.
     */
    
    public boolean isConnected();
    
    /**
     * Returns true if the Reader is dead and no new data can be expected, false
     * otherwise. Note that a Reader wich is not connected might still be alive,
     * because it is trying to restore the connection to the device.
     */
    
    public boolean isDead();
    
    /**
     * Returns the types of parameters in getData. The types are in the order of getParameterNames. 
     * The types are defined in {@link hiit.mapper.Properties}.
     * @return the parameter types, as described in Properties. See {@link hiit.mapper.Properties}. 
     */
    
    public int[] getParameterTypes();
    
    /**
     * Returns the number of the last data batch from the sensor. This method
     * can be used to check if the Reader is getting new data from the sensor.
     */
    
    public int getLastDataNumber();
    
    /**
     * There are two modes available for Readers: streaming mode and request
     * mode. In streaming mode the Reader will read all available data from the
     * data stream. In request mode the Reader only reads data when readOne() is
     * called. Also, in request mode the Reader connects to the source before
     * reading a message and disconnects after the message has been read.
     * 
     * @param streaming
     *            sets streaming mode on with true value and request mode with
     *            false value
     */
    
    public void setStreamingMode(boolean streaming);
    
    /**
     * Returns the streaming mode.
     * 
     * @return true, if the Reader is in streaming mode. Returns false if the
     *         Reader is in request mode.
     */
    
    public boolean getStreamingMode();
    
    /**
     * If the Reader is in request mode, calling this method will cause the
     * Reader to connect to the source, read one message and disconnect from the
     * source. Note that calling this method when the Reader is in streaming
     * mode has no effect.
     */
    
    public void readOne();
    
}
