/*
 * BeTelGeuse - A context data gathering tool for Bluetooth devices
 * www.cs.helsinki.fi/group/acs/betelgeuse
 * Copyright (C) 2006 Joonas Kukkonen, Eemil Lagerspetz and Petteri Nurmi
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package hiit.reader;

import hiit.observer.Observer;
import hiit.blackboard.BlackboardFactory;
import hiit.blackboard.Blackboard;
import hiit.blackboard.Customer;
import hiit.blackboard.CustomerData;
import hiit.blackboard.CustomerInfo;
import hiit.blackboard.event.EventHandler;
import hiit.connectorUnit.BTReaderFactory;
import hiit.logger.Logger;
import hiit.mapper.Properties;

import javax.microedition.io.*;
import javax.bluetooth.*;

import java.util.Date;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Vector;
import java.io.IOException;
import java.io.InputStream;

/**
 * The ReaderSkeleton class implements most of the methods described in the
 * Reader interface.
 * 
 * @author Joonas Kukkonen and Eemil Lagerspetz
 * 
 */

public abstract class ReaderSkeleton implements Reader, Customer {
    
    /**
     * Timeout for the 'last try' to raise the connection, milliseconds. If
     * giveUp is set to false, will continue waiting this timeout and trying to
     * raise. If set to true, will give up after waiting this time and trying to
     * raise.
     */
    protected int finalTimeout = 128000;
    
    /**
     * Timeout when waiting to get the connection back up after a connection
     * break the first time. Will continue trying with firstTimeout * 2^n with
     * n=1,2,3,4... where n is the number of tries.
     */
    protected int firstTimeout = 4000;
    
    /**
     * If the final timeout is reached and the connection is still down, should
     * we abandon trying?
     */
    protected boolean giveUp;
    
    /**
     * The connection url to the remote device.
     */
    protected String url;
    
    /**
     * The logger that prints messages. Useful for mobile devices.
     */
    protected Logger logger;
    
    /**
     * Used to store the names and values of parameters. The name (key) is the
     * parameter name that is visualized whereas the value is the current sensor
     * reading of that parameter.
     */
    protected Hashtable parameters;
    
    /**
     * Used to store all of the read data, if an object subscribes for the
     * Reader and asks for all possible data.
     */
    
    protected Hashtable streamParameters;
    
    /**
     * The connection which is used to read data.
     */
    protected InputStream iStream;
    
    /**
     * Boolean flag used for managing when the reader should run.
     */
    protected boolean isRunning = true;
    
    /**
     * A list of objects that have subscribed to events from the reader. Each
     * time the reader reads information from the device, a new event is
     * produced.
     */
    protected Vector subscribers = new Vector();
    
    /**
     * A list of subscriber timeouts (how often they want data).
     */
    
    protected Vector timeouts = new Vector();
    
    /**
     * A vector of timestamps when data was sent last to subscribers.
     */
    
    protected Vector lastSubscriberUpdate = new Vector();
    
    /**
     * Delay, specifies how often messages are read from the sensing device.
     */
    protected long delay = 100;
    
    /**
     * The connection to the remote device.
     */
    
    protected Connection connection;
    
    /**
     * The ReaderFactory used to create the Reader objects. Used in connection
     * restoration.
     */
    
    protected BTReaderFactory factory;
    
    /**
     * The device ID number associated to the sensor to which this Reader is
     * connected to.
     */
    
    protected int deviceId;
    
    /**
     * Describes if the Reader is connected or not.
     */
    
    protected boolean connected = true;
    
    /**
     * The number of last data batch.
     */
    
    protected int dataBatchNumber = 0;
    
    /**
     * The mode of the Reader. True if the Reader is in stream mode, false
     * otherwise.
     */
    
    protected boolean mode;
    
    protected CustomerInfo info;
    
    protected EventHandler eventHandler;
    
    public ReaderSkeleton() {
        
    }
    
    public ReaderSkeleton(Connection connection, String url, Logger logger,
            BTReaderFactory factory, int deviceId, boolean streamMode) {
        this(connection, url, logger, factory, deviceId, false, streamMode);
    }
    
    /**
     * Creates a new BTReader. Initializes the input stream using the Connection
     * object given as input parameter.
     * 
     * @param connection
     *            A reference to an open connection. The connection should be
     *            created using the Connector.open(URL) method.
     */
    public ReaderSkeleton(Connection connection, String url, Logger logger,
            BTReaderFactory factory, int deviceId, boolean streamMode,
            boolean giveUp) {
        this.url = url;
        this.logger = logger;
        this.factory = factory;
        this.deviceId = deviceId;
        // this.bluetoothAddress = bluetoothAddress;
        this.connection = connection;
        this.mode = streamMode;
        
        // Initialize the hash table
        initializeVariables();
        
        this.giveUp = giveUp;
        
        // Try to create the connection
        try {
            if (connection != null && mode)
                this.iStream = ((StreamConnection) connection)
                        .openInputStream();
        } catch (BluetoothConnectionException e) {
            int status = e.getStatus();
            connected = false;
            isRunning = false;
            info.setAlive(false);
            switch (status) {
            case BluetoothConnectionException.FAILED_NOINFO:
                logger
                        .logMessage("FAILURE - The connection to the server failed due to unknown reasons.");
                break;
            case BluetoothConnectionException.NO_RESOURCES:
                logger
                        .logMessage("FAILURE - A lack of resources either on the local device or on the remote device.");
                break;
            case BluetoothConnectionException.SECURITY_BLOCK:
                logger
                        .logMessage("FAILURE - The security settings on the local device or the remote device were incompatible with the request.");
                break;
            case BluetoothConnectionException.TIMEOUT:
                logger.logMessage("FAILURE - Timed out.");
                break;
            case BluetoothConnectionException.UNACCEPTABLE_PARAMS:
                logger
                        .logMessage("FAILURE - The configuration parameters provided were not acceptable to either the remote device or the local device.");
                break;
            case BluetoothConnectionException.UNKNOWN_PSM:
                logger
                        .logMessage("FAILURE - No service for the given PSM was registered.");
                break;
            }
            try {
                connection.close();
            } catch (IOException ex) {
                // Do nothing
            }
        } catch (IOException ex) {
            if (logger != null)
                logger.logMessage(ex.toString());
            connected = false;
            isRunning = false;
            info.setAlive(false);
            try {
                connection.close();
            } catch (IOException exx) {
                // Do nothing
            }
        }
    }
    
    /**
     * Initializes the Reader. Meant for cases when the Reader implementation is
     * constructed with an empty constructor by Class.forName() method.
     * 
     * @param connection
     *            the connection to the sensor
     * @param url
     *            the connection URL
     * @param logger
     *            the Logger object
     * @param factory
     *            the BTDeviceFactory which is used to restore connections
     * @param deviceId
     *            the device ID of the sensor
     * @param streamMode
     *            the mode of the Reader at startup. True for streaming mode,
     *            false for request mode.
     */
    
    public void initialize(Connection connection, String url, Logger logger,
            BTReaderFactory factory, int deviceId, boolean streamMode) {
        this.url = url;
        this.logger = logger;
        this.factory = factory;
        this.deviceId = deviceId;
        // this.bluetoothAddress = bluetoothAddress;
        this.connection = connection;
        this.mode = streamMode;
        
        // Initialize the hash table
        initializeVariables();
        
        this.giveUp = false;
        
        // Try to create the connection
        try {
            if (connection != null && mode)
                this.iStream = ((StreamConnection) connection)
                        .openInputStream();
        } catch (BluetoothConnectionException e) {
            int status = e.getStatus();
            connected = false;
            isRunning = false;
            info.setAlive(false);
            switch (status) {
            case BluetoothConnectionException.FAILED_NOINFO:
                logger
                        .logMessage("FAILURE - The connection to the server failed due to unknown reasons.");
                break;
            case BluetoothConnectionException.NO_RESOURCES:
                logger
                        .logMessage("FAILURE - A lack of resources either on the local device or on the remote device.");
                break;
            case BluetoothConnectionException.SECURITY_BLOCK:
                logger
                        .logMessage("FAILURE - The security settings on the local device or the remote device were incompatible with the request.");
                break;
            case BluetoothConnectionException.TIMEOUT:
                logger.logMessage("FAILURE - Timed out.");
                break;
            case BluetoothConnectionException.UNACCEPTABLE_PARAMS:
                logger
                        .logMessage("FAILURE - The configuration parameters provided were not acceptable to either the remote device or the local device.");
                break;
            case BluetoothConnectionException.UNKNOWN_PSM:
                logger
                        .logMessage("FAILURE - No service for the given PSM was registered.");
                break;
            }
            try {
                connection.close();
                connected = false;
                isRunning = false;
                info.setAlive(false);
            } catch (IOException ex) {
                // Do nothing
            }
        } catch (IOException ex) {
            if (logger != null)
                logger.logMessage(ex.toString());
            try {
                connection.close();
                connected = false;
                isRunning = false;
                info.setAlive(false);
            } catch (IOException exx) {
                // Do nothing
            }
        }
        this.info = new CustomerInfo(getDeviceName(), deviceId, 1, this.url
                .substring(8, 20), (streamMode ? 0 : 1), delay,
                getParameterNames(), getParameterTypes(), false);
        BlackboardFactory.getBlackboard(logger).registerProducer(this, info, null);
        
    }
    
    // Methods that need to be overridden:
    
    /**
     * Get the name of the device whose values the reader reads.
     */
    public abstract String getDeviceName();
    
    /**
     * Initializes the hash table by creating the list of parameters that the
     * reader provides.
     */
    protected abstract void initializeVariables();
    
    /**
     * Abstract method, handles processing and reading of data. The method needs
     * to return true when new data is available and false otherwise.
     */
    protected abstract boolean readData() throws java.io.IOException;
    
    // End of methods to override in subclasses
    
    /**
     * Gets the names of available parameters.
     * 
     * @return String[] A String array that contains the names of the parameters
     *         that the reader provides.
     */
    public String[] getParameterNames() {
        
        // Avoid null pointer exceptions
        if (this.parameters == null)
            return null;
        
        // Reserve memory
        final int i = this.parameters.size();
        String[] params = new String[i];
        
        // Loop though the keys and store the parameter names.
        Enumeration enume = this.parameters.keys();
        int index = 0;
        while (enume.hasMoreElements())
            params[index++] = (String) enume.nextElement();
        
        // Return parameter names.
        return params;
    }
    
    /**
     * Get the value of a specific parameter.
     * 
     * @param name
     *            The name of the parameter whose value is requested.
     * @return String The value of the parameter of null if the reader does not
     *         provide that variable.
     */
    public String getParameter(String name) {
        if (this.parameters == null)
            return null;
        return (String) this.parameters.get(name);
    }
    
    /**
     * Closes the reader and closes all connections. Sets the isRunning flag to
     * false which will stop the reader.
     */
    public void close() {
        this.isRunning = false;
        info.setAlive(false);
        
        try {
            if (this.iStream != null)
                this.iStream.close();
            if (this.connection != null)
                this.connection.close();
        } catch (IOException e) {
            // Nothing can be done
        }
        
    }
    
    /**
     * Reads data from the device. To implement a reader, extend the
     * ReaderSkeleton class and implement the readData method where all
     * processing takes place.
     */
    public synchronized void run() {
        long time_1, time_2;
        int failures = 0;
        try {
            while (this.isRunning) {
                
                if (!mode) {
                    // If in request mode...
                    
                    // check if the reader needs to disconnect...
                    if (this.iStream != null && this.connection != null)
                        disconnect();
                    
                    // Wait for something to happen...
                    synchronized (this) {
                        this.wait();
                    }
                    
                    if (!mode) {
                        if (connect()) {
                            // If still in request mode, connect, read one
                            // message and disconnect
                            readData();
                            disconnect();
                            notifySubscribers();
                        }
                        // Jump back and wait...
                        continue;
                    }
                    // else: mode is true, so continue normally... the reader
                    // has already been reconnected...
                }
                
                // Get first tick
                time_1 = System.currentTimeMillis();
                
                // Read and process data
                try {
                    if (!readData())
                        failures++;
                    else {
                        failures = 0;
                        this.dataBatchNumber++;
                    }
                    
                    if (delay != 0 && delay < 30000 && failures > 30000 / delay) {
                        if (url != null) {
                            if (logger != null)
                                logger.logMessage("Disconnected from "
                                        + url.substring(0, url.indexOf(';'))
                                        + ", trying to restore connection...");
                            connected = false;
                            restoreConnection();
                        }
                    } else if (delay == 0 && failures > 1000) {
                        if (url != null) {
                            if (logger != null)
                                logger.logMessage("Disconnected from "
                                        + url.substring(0, url.indexOf(';'))
                                        + ", trying to restore connection...");
                            connected = false;
                            restoreConnection();
                        }
                    }
                    
                } catch (java.io.IOException ex) {
                    connected = false;
                    if (isRunning) {
                        if (url != null) {
                            if (logger != null)
                                logger.logMessage("Disconnected from "
                                        + url.substring(0, url.indexOf(';'))
                                        + ", trying to restore connection...");
                            restoreConnection();
                        }
                    }
                } catch (Throwable t) {
                    if (logger != null)
                        logger.logMessage("[" + this.getDeviceName()
                                + "]Throwable caught: " + t);
                }
                // Request the monitor of the reader
                synchronized (this) {
                    
                    // Get second tick
                    time_2 = System.currentTimeMillis();
                    
                    // If the delay has not been long enough, sleep for a while
                    if ((time_2 - time_1) >= 0 && (time_2 - time_1) < delay)
                        this.wait(delay - (time_2 - time_1));
                }
                notifySubscribers(); // TSEK
            }
            
            // Close the input stream
            if (iStream != null)
                this.iStream.close();
            
        } catch (Exception ex) {
            if (logger != null) {
                logger.logMessage("[" + this.getDeviceName()
                        + "]Reader error, reader stopped. " + ex.toString());
            }
        }
    }
    
    /**
     * Subscribes to events from the reader.
     * 
     * @param obs
     *            The Observer that subscribes to events.
     */
    public void subscribe(Observer obs, long timeout) {
        synchronized (this) {
            this.subscribers.addElement(obs);
            this.timeouts.addElement(new Long(timeout));
            this.lastSubscriberUpdate.addElement(new Long(0));
        }
    }
    
    /**
     * Notifies all subscribers that new data is available.
     */
    protected void notifySubscribers() {
        // Read the current data only once
        String[] names = this.getParameterNames();
        Vector[] temp = new Vector[names.length];
        
        for (int i = 0; i < names.length; i++)
            temp[i] = (Vector) this.streamParameters.get(names[i]);
        // Push data to the BB.
        BlackboardFactory.getBlackboard(logger).pushData(this,
                new CustomerData(temp, new Date().getTime(), info), null );
        if (this.subscribers.size() == 0)
            return;
        
        synchronized (this) {
            long time = System.currentTimeMillis();
            for (int i = 0, j = this.subscribers.size(); i < j; i++) {
                if (((Long) this.lastSubscriberUpdate.elementAt(i)).longValue()
                        + ((Long) this.timeouts.elementAt(i)).longValue() < time) {
                    // synchronized(this.subscribers.elementAt(i)) {
                    // Send data to the observer. Observer decides what it does
                    // with the data.
                    ((Observer) this.subscribers.elementAt(i)).update(temp,
                            this);
                    this.lastSubscriberUpdate.removeElementAt(i);
                    this.lastSubscriberUpdate
                            .insertElementAt(new Long(time), i);
                    // Wake the observer
                    // this.subscribers.elementAt(i).notify();
                    // }
                }
            }
        }
    }
    
    /**
     * Unsubscribes the Observer from events from the reader.
     * 
     * @param obs
     *            The Observer that unsubscribes.
     */
    public void unsubscribe(Observer obs) {
        synchronized (this) {
            int i = this.subscribers.indexOf(obs);
            this.subscribers.removeElement(obs);
            this.timeouts.removeElementAt(i);
            this.lastSubscriberUpdate.removeElementAt(i);
        }
    }
    
    /**
     * Returns all the parameters from this Reader.
     * 
     * @return String[] A String array that contains the parameters.
     */
    
    public String[] getData() {
        
        // Avoid null pointer exceptions
        if (this.parameters == null)
            return null;
        
        // Reserve memory
        final int max = this.parameters.size();
        String[] params = new String[max];
        
        // Loop though the keys and store the parameter names.
        Enumeration enume = this.parameters.elements();
        int index = 0;
        while (enume.hasMoreElements())
            params[index++] = (String) enume.nextElement();
        
        // Return parameter names.
        return params;
    }
    
    /**
     * Tries to restore the connection to the device associated with this
     * reader.
     */
    
    protected synchronized void restoreConnection() {
        int timeout = firstTimeout;
        
        // stop trying only if we are allowed to give up and the final timeout
        // is reached.
        while (!(timeout >= finalTimeout && giveUp)) {
            Connection connection;
            
            connection = this.factory.tryRestore(this, url, logger);
            // else if (...) { //Some other Factory than the bluetooth one...}
            if (connection != null) {
                
                try {
                    this.iStream = ((StreamConnection) connection)
                            .openInputStream();
                    this.connection = connection;
                    if (logger != null)
                        logger.logMessage("Connection restored to "
                                + url.substring(0, url.indexOf(';')));
                    connected = true;
                    return;
                } catch (IOException e) {
                    if (logger != null)
                        logger.logMessage("Unable to restore connection.");
                }
            }
            
            try {
                this.wait(timeout);
            } catch (InterruptedException e) {
                // Woken up, let's just be happy and continue
            }
            
            timeout = timeout * 2;
            if (timeout > finalTimeout)
                timeout = finalTimeout;
        }
        isRunning = false;
        info.setAlive(false);
        if (logger != null)
            logger.logMessage("Reader " + getDeviceName()
                    + " cannot restore connection.");
    }
    
    /**
     * Sets timeouts for connection restoration. The timeout starts at
     * firstTimeout and it is multiplied with 2 each time connection restoration
     * fails. FinalTimeout defines the maximum lenght for a timeout. If giveUp
     * is true, the reader does not try to restore the connection after the
     * finalTimeout value is reached. If giveUp is false, the Reader tries to
     * restore connection indefinetly. Default firstTimeout is 4000 milliseconds
     * and finalTimeout 128000 milliseconds.
     * 
     * @param firstTimeout
     *            The lenght of first timeout in milliseconds
     * @param finalTimeout
     *            The maximum lenght of a timeout in milliseconds
     * @param giveUp
     *            Defines if the Reader gives up connection restoration after
     *            the final timeout is reached
     */
    
    public void setTimeouts(int firstTimeout, int finalTimeout, boolean giveUp) {
        this.firstTimeout = firstTimeout;
        this.finalTimeout = finalTimeout;
        this.giveUp = giveUp;
    }
    
    /** 
     * Getter for the timeouts specified in the method above.
     * 
     * @return a Vector with the elements: firstTimeout (int), finalTimeout
     *         (int), giveUp (boolean).
     */
    public Vector getTimeouts() {
        Vector timeouts = new Vector();
        timeouts.addElement(new Integer(firstTimeout));
        timeouts.addElement(new Integer(finalTimeout));
        timeouts.addElement(new Boolean(giveUp));
        return timeouts;
    }
    
    /**
     * Sets the delay Reader waits before reading data again. Default delay is
     * 100 milliseconds.
     * 
     * @param delay
     *            Delay for reading data in milliseconds.
     */
    
    public void setReadDelay(long delay) {
        this.delay = delay;
    }
    
    /**
     * Getter for the read delay.
     * 
     * @return the delay the Reader waits between subsequent data reads.
     */
    public long getReadDelay() {
        return this.delay;
    }
    
    /**
     * Returns the bluetooth address of the sensor this Reader is connected to.
     */
    
    public String getBluetoothAddress() {
        return this.url.substring(8, 20);
    }
    
    /**
     * Returns the device ID number associated with the sensor to which this
     * Reader is connected to.
     */
    
    public int getDeviceId() {
        return this.deviceId;
    }
    
    /**
     * Returns true if the Reader is dead and no new data can be expected, false
     * otherwise. Note that a Reader which is not connected might still be
     * alive, because it is trying to restore the connection to the device.
     */
    
    public boolean isDead() {
        return !this.isRunning;
    }
    
    /**
     * Returns true if the Reader is connected to a sensor or other device,
     * false otherwise.
     */
    
    public boolean isConnected() {
        return this.connected;
    }
    
    /**
     * Returns the number of the last data batch from the sensor. This method
     * can be used to check if the Reader is getting new data from the sensor.
     */
    
    public int getLastDataNumber() {
        return this.dataBatchNumber;
    }
    
    /**
     * There are two modes available for Readers: streaming mode and request
     * mode. In streaming mode the Reader will read all available data from the
     * data stream. In request mode the Reader only reads data when readOne() is
     * called. Also, in request mode the Reader connects to the source before
     * reading a message and disconnects after the message has been read.
     * 
     * @param streaming
     *            sets streaming mode on with true value and request mode with
     *            false value
     */
    
    public void setStreamingMode(boolean streaming) {
        if (!mode && streaming) {
            this.restoreConnection();
            this.notify();
        }
        this.mode = streaming;
    }
    
    /**
     * Returns the streaming mode.
     * 
     * @return true, if the Reader is in streaming mode. Returns false if the
     *         Reader is in request mode.
     */
    
    public boolean getStreamingMode() {
        return this.mode;
    }
    
    /**
     * Disconnects from the data source.
     */
    
    protected void disconnect() {
        try {
            this.iStream.close();
            this.connection.close();
            this.iStream = null;
            this.connection = null;
        } catch (IOException e) {
        }
    }
    
    /**
     * Connects to the data source.
     */
    
    protected boolean connect() {
        Connection connection = this.factory.tryRestore(this, url, logger);
        // else if (...) { //Some other Factory than the bluetooth one...}
        if (connection != null) {
            
            try {
                this.iStream = ((StreamConnection) connection)
                        .openInputStream();
                this.connection = connection;
                return true;
            } catch (IOException e) {
                if (logger != null)
                    logger.logMessage("Unable to restore connection.");
            }
        }
        return false;
    }
    
    public void readOne() {
        if (!mode) {
            synchronized (this) {
                this.notify();
            }
        }
    }
    
    public CustomerInfo getInfo() {
        return info;
    }
    
    public void newData(Customer prod, CustomerData pd) {
        // we do not receive data
    }
    
    public boolean requestData() {
        // TODO: What to do here?
        return false;
    }
    
    public boolean setProducerMode(int mode, long timeout) {
        // TODO Auto-generated method stub
        return false;
    }
    
    public void updateCustomerInfo(Blackboard bb, Customer c, CustomerInfo pi) {
        // TODO Auto-generated method stub
    }
    
    public void updateProducerList(Blackboard bb, Vector customers) {
        // Producer only, does not subscribe.
    }
    
    public void markParameters(){
    	//Overwrite if your reader supports events... See GPS reader for example
    }
}
